/////////////////////////////////////////////////////////////////////////////
// Name:        TASOSG.h
// Purpose:     Declares base panel and 3DPanel classes incorporating OSG
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2012, 2013 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef TASOSG_H_
#define TASOSG_H_

#include <wx/defs.h>
#include <wx/app.h>
#include <wx/cursor.h>
#include <wx/glcanvas.h>
#include <wx/fileconf.h>
#include <osgViewer/Viewer>
#include <osgText/Text>
#include <osg/io_utils>
#include <osg/ShapeDrawable>
#include <string>
#include <sstream>
#include "TASTiles.h"
#include "TASApp.h"

#define LNP_PICK_ACTION_NONE      0
#define LNP_PICK_ACTION_IDENTIFY  1
#define LNP_PICK_ACTION_ADD       2
#define LNP_PICK_ACTION_ROTATE    3

#define RADSIN90     1.5708f
#define LNPS_MAX_ARMS 6

#define LNPS_NORTH   0
#define LNPS_EAST    1
#define LNPS_UP      2
#define LNPS_SOUTH   3
#define LNPS_WEST    4
#define LNPS_DOWN    5

int GetOppositeDir(int dir) { return (dir + 3) % LNPS_MAX_ARMS; };

#define LNP_X_AXIS   0
#define LNP_Y_AXIS   1
#define LNP_Z_AXIS   2

osg::Vec3 rotPlane[3] = { osg::Vec3(1,0,0), osg::Vec3(0,1,0), osg::Vec3(0,0,1) };

int arrPlanes[3][4] = {{ LNPS_UP,    LNPS_NORTH, LNPS_DOWN,  LNPS_SOUTH }, 
                       { LNPS_UP,    LNPS_WEST,  LNPS_DOWN,  LNPS_EAST  },
                       { LNPS_NORTH, LNPS_EAST,  LNPS_SOUTH, LNPS_WEST  } };

int arrAxis[3][2] = { { LNPS_EAST, LNPS_WEST }, { LNPS_NORTH, LNPS_SOUTH }, { LNPS_UP, LNPS_DOWN } };

int GetIndexInPlane(int dir, int plane);
int GetOrthogonalDir(int dir1, int dir2);
int GetAxis(int dir);
bool IsOrthogonalAxis(int dir1, int dir2, int axis);

wxString ArrDirNames[LNPS_MAX_ARMS] = { wxT("North"), wxT("East"), wxT("Up"), wxT("South"), wxT("West"), wxT("Down") };
wxString GetDirStringFromNum(int dir) { return ArrDirNames[dir]; };
int GetDirNumFromString(wxString str);

osg::Quat arrAttQuats[LNPS_MAX_ARMS] = {osg::Quat(RADSIN90, osg::Vec3d(1,0,0)), osg::Quat(RADSIN90, osg::Vec3d(0,1,0)),
                                        osg::Quat(0, osg::Vec3d(0,0,0)),        osg::Quat(RADSIN90, osg::Vec3d(1,0,0)),
                                        osg::Quat(RADSIN90, osg::Vec3d(0,1,0)), osg::Quat(0, osg::Vec3d(0,0,0)) };

osg::Vec3 arrPos1Vec[LNPS_MAX_ARMS] =  {osg::Vec3(0,4.5,0),  osg::Vec3(4.5,0,0),
                                        osg::Vec3(0,0,4.5),  osg::Vec3(0,-4.5,0),
                                        osg::Vec3(-4.5,0,0), osg::Vec3(0,0,-4.5)};

osg::Vec3 arrPos2Vec[LNPS_MAX_ARMS] =  {osg::Vec3(0,9.0,0),  osg::Vec3(9.0,0,0),
                                        osg::Vec3(0,0,9.0),  osg::Vec3(0,-9.0,0),
                                        osg::Vec3(-9.0,0,0), osg::Vec3(0,0,-9.0)};


class GraphicsWindowWX;

wxColor LNP_DEFAULT_COLOR = *wxLIGHT_GREY;

class LNP
{
public:
  LNP() { m_bIsValid = false; };
  LNP(wxString sName) { m_sName = sName; m_bIsValid = false; };
  //~LNP() {};

  void Clear();

//  LNP& operator=(const LNP& lnp);
//  bool operator==(const LNP& lnp);
  inline bool operator<(const LNP& lnp) { return (m_sName < lnp.m_sName); };
  inline bool operator<(const LNP& lnp) const { return (m_sName < lnp.m_sName); };

  bool IsValid() { return m_bIsValid; };

  osg::Group* CreateOSGObject();

  wxString  m_sName;
  bool      m_bIsValid;
};

class BoxLoc
{
public:
//  LNPLoc() { m_pLNP = NULL; m_pNode = NULL; };
  BoxLoc(LNP* pLNP, osg::Vec3 vecLoc);

  LNP         m_lnp;
  osg::PositionAttitudeTransform*   m_pat;
  osg::Vec3   m_vecLoc;
};

class LNPAssembly : public std::vector<BoxLoc>
{
public:
};

class OSGCanvas : public wxGLCanvas
{
public:
    OSGCanvas(wxWindow *parent, wxWindowID id = wxID_ANY,
        const wxPoint& pos = wxDefaultPosition,
        const wxSize& size = wxDefaultSize, long style = 0,
        const wxString& name = wxT("LNPCanvas"),
        int *attributes = 0);

    virtual ~OSGCanvas();

    void SetGraphicsWindow(osgViewer::GraphicsWindow *gw)   { _graphics_window = gw; }

    void OnPaint(wxPaintEvent& event);
    void OnSize(wxSizeEvent& event);
    void OnEraseBackground(wxEraseEvent& event);

    void OnChar(wxKeyEvent &event);
    void OnKeyUp(wxKeyEvent &event);

    void OnMouseEnter(wxMouseEvent &event);
    void OnMouseDown(wxMouseEvent &event);
    void OnMouseUp(wxMouseEvent &event);
    void OnMouseMotion(wxMouseEvent &event);

    void UseCursor(bool value);

private:
    DECLARE_EVENT_TABLE()

    osg::ref_ptr<osgViewer::GraphicsWindow> _graphics_window;

    wxCursor _oldCursor;
};

class GraphicsWindowWX : public osgViewer::GraphicsWindow
{
public:
    GraphicsWindowWX(OSGCanvas *canvas);
    ~GraphicsWindowWX();

    void init();

    //
    // GraphicsWindow interface
    //
    void grabFocus();
    void grabFocusIfPointerInWindow();
    void useCursor(bool cursorOn);

    bool makeCurrentImplementation();
    void swapBuffersImplementation();

    // not implemented yet...just use dummy implementation to get working.
    virtual bool valid() const { return true; }
    virtual bool realizeImplementation() { return true; }
    virtual bool isRealizedImplementation() const  { return true; }
    virtual void closeImplementation() {}
    virtual bool releaseContextImplementation() { return true; }

private:
    // XXX need to set _canvas to NULL when the canvas is deleted by
    // its parent. for this, need to add event handler in OSGCanvas
    OSGCanvas*  _canvas;
};

class LNPBaseFrame : public wxFrame
{
public:
  LNPBaseFrame(TASApp* pApp, wxFrame *frame, const wxString& title, const wxPoint& pos,
      const wxSize& size, long style = wxDEFAULT_FRAME_STYLE);

  void SetViewer(osgViewer::Viewer *viewer);
  void OnIdle(wxIdleEvent& event);

//protected:
  TASApp*						  m_pApp;
  wxStatusBar*                    m_pStatusBar;
  osg::ref_ptr<osg::Group>        m_pGroup;
  osg::ref_ptr<osgViewer::Viewer> _viewer;
};

class FractalViewerFrame : public LNPBaseFrame
{
public:
  FractalViewerFrame(TASApp* pApp, wxFrame *frame, const wxString& title, const wxPoint& pos,
      const wxSize& size, long style = wxDEFAULT_FRAME_STYLE);

  bool AddLNP(osg::Vec3& coords);
  bool AddGroup(std::vector<osg::Vec3>* pVectCoords);
  bool FractalViewerFrame::AddCopyAtOffset(osg::ref_ptr<osg::Group> pParentGroup, osg::Vec3 coords);
  bool Grow();

  int             m_stage;
  LNP             m_lnp;
  int             m_scale;
  std::vector<osg::Vec3> m_generator;
  LNPAssembly     m_lnpAssembly;
  wxLogWindow*    m_pLogWin;

private:
  // any class wishing to process wxWidgets events must use this macro
  DECLARE_EVENT_TABLE()
};

// class to handle events with a pick
class PickHandler : public osgGA::GUIEventHandler {
public: 

  PickHandler(osgText::Text* updateText, LNPBaseFrame* pFrame):
      _updateText(updateText) { m_pFrame = pFrame; };
      
  ~PickHandler() {}
  
  bool handle(const osgGA::GUIEventAdapter& ea,osgGA::GUIActionAdapter& aa);

  virtual bool pick(osgViewer::View* view, const osgGA::GUIEventAdapter& ea, int action);

  void setLabel(const std::string& name)
  {
      if (_updateText.get()) _updateText->setText(name);
  }
    
protected:

  LNPBaseFrame* m_pFrame;
  osg::ref_ptr<osgText::Text>  _updateText;
};

#endif // TASOSG_H_
