/////////////////////////////////////////////////////////////////////////////
// Name:        TASTileCanvas.h
// Purpose:     Declares the panel class used for 2-D tile space rendering
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef __TASTileCanvas__
#define __TASTileCanvas__

#include <wx/graphics.h>
#include "TASPanel.h"
#include "TASSimFrm.h"
#include "TASTiles.h"
#include "wxRubberBand.h"


const int INVALID_LOC               = -999999999;
const int TAS_MOUSE_MODE_DRAG       = 0;
const int TAS_MOUSE_MODE_FRONTIER   = 1;

class TASSimulatorFrame;
class TASTilePanel;

class TASRightClickWin : public wxPanel
{
public:
  TASRightClickWin(TASTilePanel* parent);

  void ShowEx(wxPoint& pt, TASCoords& coords);
  void OnSelection(wxCommandEvent& event);

  void UpdateValues();

private:

  TASTilePanel* m_pParent;
  wxListBox*    m_pListBox;
  TASCoords     m_coords;

  DECLARE_CLASS(TASRightClickWin)
  DECLARE_EVENT_TABLE()
};

class TASTilePanel : public TASPanelBase
{
public:
    TASTilePanel(TASSimulatorFrame* parentFrame,
                   TASTilePanel* pMainPanel,
                   wxWindowID id = wxID_ANY,
                   const wxPoint& pos = wxDefaultPosition,
                   const wxSize& size = wxDefaultSize,
                   long style = wxTAB_TRAVERSAL | wxNO_BORDER | wxScrolledWindowStyle,
                   const wxString& name = wxT("TASTilePanel"));    
    ~TASTilePanel();

    virtual void Refresh( bool eraseBackground = true,
                          const wxRect *rect = (const wxRect *) NULL ) { TASPanelBase::Refresh(eraseBackground, rect); };

    void OnPaint(wxPaintEvent& event);
    void OnMouse(wxMouseEvent& event);
    void OnFrontierMouse(wxMouseEvent& event, TASCoords& coordsCurr);
    void OnMouseWheel(wxMouseEvent& event);
    void OnKeyboard(wxKeyEvent& event);
    void OnRtClickMenu(wxCommandEvent& event);
    void OnScrollEx(wxScrollEvent& event);
    void OnScrollWin(wxScrollWinEvent& event);
    void OnSize(wxSizeEvent& event);
    void OnEraseBackground(wxEraseEvent& event);

    wxString GetPositionStatusText();
    void DoCanvasResize(bool bNoRefresh = false, bool bRefreshOverview = true);
    void CheckCanvasSize();

    void UpdateSpace(bool bEntire);
    void PrepareForDrawing(wxDC* pDC);

    bool IsValid();

    bool RenderTileSpace(wxDC* pDC);
    void DrawIteration(size_t epoch);
    void DrawLocation(TASCoords& coords, bool bNeighborhood = false);
    void RefreshTileRect(TASCoords& coords);

    //  Functions for rendering tiles
    void DrawTileEx(wxDC* pDC, TASCoords& coords, TASTileType* pTT, int drawType, bool bOutline = false);
    virtual void GetCoordsFromLocation(int x, int y, TASCoords* coords, bool bUseTilespaceDims = false);
    virtual void GetTileLocation(TASCoords& coords, int* x, int* y);
    void GetCenterViewablePixel(int *x, int *y, bool bFromScaling);
    void UnselectCell(wxDC* pDC);

    void BuildRtClickMenu(wxMenu* pMenu);
    void SetMouseMode(int mode);

    void SetMainPanel(TASTilePanel* pPanel);
    inline bool IsInOverviewWin() { return m_bIsOverview; };   //  Returns true if this instance of a TASTileSpacePanel is embedded in the overview window
    
    virtual float SetScaleFact(float fScaleFact, TASCoords* pCtrCoords = NULL, bool bKeepRelativePosition = false);
    virtual float GetScaleFact();
    float GetScaleStep() { return 25.0f; };
    virtual void DoScaling(wxDC* pDC);

    void Scroll(int x, int y);

    virtual int GetVirtHeight() { return m_virtHeight; };

    void Reset();
    void ResetRubberband();

    TASSimulatorFrame*   m_parentFrame;        //  Pointer to parent frame
    TASSpace*     m_pSpace;             //  Pointer to whichever (type of) space is currently being rendered    

    wxBrush       m_brush;
    wxBrush       m_brushBackground;
    wxPen         m_pen;
    wxPen         m_penBackground;

    wxRubberBand  m_RubberBand;
    wxPoint       m_rubberbandorigin;
    wxPen         m_rubberbandPen;

    float         m_fScaleFact;         //  Scaling factor to control zooming

    int           m_virtWidth;          //  Virtual width of panel
    int           m_virtHeight;         //  Virtual height of panel

    long          m_lBeginX;            //  Position of mouse (x coord)
    long          m_lBeginY;            //  Position of mouse (y coord)

    bool          m_bPanning;           //  Whether or not a grab-drag is occurring
    int           m_drag_start_x;       //  Where the grab-drag started
    int           m_drag_start_y;
    int           m_mouseMode;          //  The current mode specifying what operation to do on a click
                                        //  either grabbing and dragging or modifying the frontier
    wxBitmap      m_bitmap;

    TASRightClickWin* m_pRtClickWin;    //  Pointer to the tile type selection window that appears on right click

protected:

    virtual void DrawOverlay(wxDC* pDC, bool bOnlyDraw);

    TASTilePanel*     m_pMainPanel;       //  Pointer to the main tile panel
    bool              m_bIsOverview;      //  True if this instance of a TASTileSpacePanel is embedded in the overview window
    wxPoint           m_ptMouse;          //  The last mouse position

    wxCursor          m_cursorGrab;
    wxCursor          m_cursorGrabbing;

    wxBrush           m_brushOffspace;    //  Brush and pen used for coloring parts of the canvas outside of the tile space
    wxPen             m_penOffspace;

    //  Variables used for handling mouse-based zooming
    int               m_rotationCount;
    int               m_zoomDirection;
    float             m_fZoomPercentage;
    wxPoint           m_ptZooming;
    int               m_zoomTicks;
    bool              m_bCenterCursorOnDClick;

  DECLARE_CLASS(TASTilePanel)
  DECLARE_EVENT_TABLE()
};

class TAS2HAMTilePanel : public TASTilePanel
{
public:
    TAS2HAMTilePanel(TASSimulatorFrame* parentFrame,
                   TASTilePanel* pMainPanel,
                   wxWindowID id = wxID_ANY,
                   const wxPoint& pos = wxDefaultPosition,
                   const wxSize& size = wxDefaultSize,
                   long style = wxTAB_TRAVERSAL | wxNO_BORDER | wxScrolledWindowStyle,
                   const wxString& name = wxT("TASTilePanel")) : TASTilePanel(parentFrame, pMainPanel, id, pos, size, style, name)
    {
    };
};

#endif /* __TASTileCanvas__ */


