/////////////////////////////////////////////////////////////////////////////
// Name:        TASTiles.cpp
// Purpose:     Defines classes for tiles, tile space, etc.
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------
#include "TASTiles.h"
#include "stdafx.h"

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

// for all others, include the necessary headers (this file is usually all you
// need because it includes almost all "standard" wxWidgets headers)
#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

#include "XPlatformFix.h"
#include TR1(random)

#include <algorithm>
#include <map>

#include <wx/tokenzr.h>
#include <wx/hashmap.h>
#include <wx/filename.h>

#include "TASApp.h"
#include "TASSimFrm.h"
#include "TAS2HAMThread.h"

//  Create the global tile space pointer
TASTileSpace*     g_pTileSpace      = NULL;
TASTileSpace*     g_pTileSpaceBackup= NULL;
TAS2HAMTileSpace* g_p2HAMTileSpace  = NULL;
TASTileTypeSetEx* g_pEditorTileset  = NULL;
TASkTAMTileSpace* g_pkTAMTileSpace  = NULL;

wxString sTempPrefix(_T("Temperature="));
wxString sStagesPrefix(_T("Stages="));
wxString sModePrefix(_T("Mode="));
wxString sGmcPrefix(_T("Gmc="));
wxString sGsePrefix(_T("Gse="));
wxString sIncludeSingletons(_T("IncludeSingletons="));
wxString sBeginSupertile(_T("BeginSupertile"));
wxString sEndSupertile(_T("EndSupertile"));

TASCoords coordsNbrhd[7]    = { TASCoords(0, 0, 0), TASCoords(1, 0, 0), TASCoords(0, 1, 0), TASCoords(-1, 0, 0), TASCoords(0, -1, 0), TASCoords(0, 0, 1), TASCoords(0, 0, -1) };
int TASOppositeDirection[]  = { TAS_SOUTH, TAS_WEST, TAS_NORTH, TAS_EAST, TAS_DOWN, TAS_UP };

TASTileType* g_pEmptyTileType = NULL;
int g_tileSize = 100;

wxString TASGetModeName(int mode)
{
  if (mode == TAS_KTAM_MODE) {
    return TAS_MODE_KTAM_STR;
  } else if (mode == TAS_2HAM_MODE) {
    return TAS_MODE_2HAM_STR;
  } else {
    return TAS_MODE_ATAM_STR;
  }
}

bool TASIsTileType(size_t t)
{
  return (t < TAS_TILE_FRONTIER);
}

wxString TASDirectionNumToString(int dirNum, bool bAllCaps)
{
  wxString retStr = _T("Invalid Direction");

  switch (dirNum) {
    case TAS_NORTH:
      if (bAllCaps) {
        retStr = _T("NORTH");
      } else {
        retStr = _T("North");
      }
      break;
    case TAS_SOUTH:
      if (bAllCaps) {
        retStr = _T("SOUTH");
      } else {
        retStr = _T("South");
      }
      break;
    case TAS_EAST:
      if (bAllCaps) {
        retStr = _T("EAST");
      } else {
        retStr = _T("East");
      }
      break;
    case TAS_WEST:
      if (bAllCaps) {
        retStr = _T("WEST");
      } else {
        retStr = _T("West");
      }
      break;
    case TAS_UP:
      if (bAllCaps) {
        retStr = _T("UP");
      } else {
        retStr = _T("Up");
      }
      break;
    case TAS_DOWN:
      if (bAllCaps) {
        retStr = _T("DOWN");
      } else {
        retStr = _T("Down");
      }
      break;
    default:
      break;
  }
  return retStr;
}

#include <wx/arrimpl.cpp>
WX_DEFINE_OBJARRAY(TASLocationTransitionList);

TASTileEpoch::TASTileEpoch()
{
  ConstructorHelper();
}

TASTileEpoch::TASTileEpoch(int temp)
{
  ConstructorHelper();
  m_temp = temp;
}

void TASTileEpoch::ConstructorHelper()
{
  m_arrBases[TAS_BEGIN_TILES]         =  0;
  m_arrBases[TAS_BEGIN_FRONTIER]      =  0;
  m_arrBases[TAS_BEGIN_EMPTY]         =  0;
  m_temp                              = -1;
}

size_t TASTileEpoch::GetEnd() {
  return m_listTransitions.GetCount();
}

bool TASTileEpoch::AddTransition(TASLocationTransition& trans)
{
  //  See if there is already a transition for the given location
  for (size_t i=0; i<m_listTransitions.GetCount(); i++) {
    if (m_listTransitions[i].m_coords == trans.m_coords) {
      //  If the existing transition results in the same type simply
      //  replace it
      if ((TASIsTileType(m_listTransitions[i].m_typeEnd) && TASIsTileType(trans.m_typeEnd)) ||
          (m_listTransitions[i].m_typeEnd == trans.m_typeEnd))
      {
        m_listTransitions[i] = trans;
        return true;
      } else {  //  If it results in something else, delete the existing
                //  transition and below we'll create a new one
        RemoveTransition(trans);
      }
    }
  }

  //  Create a new transition to the necessary type
  if (trans.m_typeEnd == TAS_TILE_NOT_FOUND) {
    m_listTransitions.Add(trans);
    return true;
  } else if (trans.m_typeEnd == TAS_TILE_FRONTIER) {
    m_listTransitions.Insert(trans, m_arrBases[TAS_BEGIN_EMPTY]);
    m_arrBases[TAS_BEGIN_EMPTY]++;
    return true;
  } else if (TASIsTileType(trans.m_typeEnd)) {
    m_listTransitions.Insert(trans, m_arrBases[TAS_BEGIN_FRONTIER]);
    m_arrBases[TAS_BEGIN_FRONTIER]++;
    m_arrBases[TAS_BEGIN_EMPTY]++;
    return true;
  }

  return false;
}

bool TASTileEpoch::RemoveTransition(TASLocationTransition& trans)
{
  int index = m_listTransitions.Index(trans);

  if (index == wxNOT_FOUND) {
    return false;
  } else {
    m_listTransitions.RemoveAt(index);
    if (trans.m_typeEnd == TAS_TILE_FRONTIER) {
      m_arrBases[TAS_BEGIN_EMPTY]--;
    } else if (TASIsTileType(trans.m_typeEnd)) {
      m_arrBases[TAS_BEGIN_FRONTIER]--;
      m_arrBases[TAS_BEGIN_EMPTY]--;
    }
    return true;
  }
}

/*
bool TASTileEpoch::RemoveTransitionsToType(int baseType)
{
  int rangeEnd = 0;
  if (baseType == TAS_BEGIN_EMPTY) {
    rangeEnd = m_listTransitions.GetCount();
  } else {
    rangeEnd = m_arrBases[baseType+1];
  }

  int num = rangeEnd - m_arrBases[baseType];
  for (size_t i=m_arrBases[baseType]; i<rangeEnd; i++) {
MJP: this is incorrect!!    m_listTransitions.RemoveAt(i);
  }

  for (int j=TAS_BEGIN_TILES; j<=TAS_BEGIN_EMPTY; j++) {
    if (baseType < j) {
      m_arrBases[j] -= num;
    }
  }
}*/

bool TASTileEpoch::RemoveTransitionAtLocation(TASCoords& coords)
{
  for (size_t i=0; i<m_listTransitions.GetCount(); i++) {
    if (m_listTransitions[i].m_coords == coords) {
      TASLocationTransition trans(m_listTransitions[i]);
      return RemoveTransition(trans);
    }
  }
  return false;
}

void TASTileEpoch::ClearTransitions()
{
  m_listTransitions.Empty();
}

bool TASTileSide::operator==(const TASTileSide& right)
{
  if (m_hashNum == right.m_hashNum) {
    return true;
  }
  return false;
}

TASTileSideHash::TASTileSideHash(TASTileSide& tsNorth, TASTileSide& tsEast,
                                 TASTileSide& tsSouth, TASTileSide& tsWest,
                                 TASTileSide& tsUp,    TASTileSide& tsDown)
{
  m_arr[TAS_NORTH] = tsNorth;
  m_arr[TAS_EAST]  = tsEast;
  m_arr[TAS_SOUTH] = tsSouth;
  m_arr[TAS_WEST]  = tsWest;
  m_arr[TAS_UP]    = tsUp;
  m_arr[TAS_DOWN]  = tsDown;

  CreateHash();
}

int orderOptDirs[6] = { TAS_SOUTH, TAS_WEST, TAS_NORTH, TAS_EAST, TAS_DOWN, TAS_UP };
void TASTileSideHash::CreateHash()
{
  m_strHash = _T("");
  for (int i=TAS_NORTH; i<=TAS_DOWN; i++) {
    int d = orderOptDirs[i];
    m_strHash <<  m_arr[d].m_iStrength << _T(":");
    if (m_arr[d].m_hashNum == TAS_TILE_NOT_FOUND) {
      m_strHash << _T("E");
    } else if (m_arr[d].m_hashNum == TAS_TILE_FRONTIER) {
      m_strHash << _T("F");
    } else {
      m_strHash << m_arr[d].m_hashNum;
    }
    if (i != TAS_DOWN) {
      m_strHash += _T(";");
    }
  }
}

bool TASTileSideHash::operator<(const TASTileSideHash& tshRight) const {
  return (m_strHash.Cmp(tshRight.m_strHash) < 0);
}

bool TASTileSideHash::operator==(const TASTileSideHash& tshRight) {
  return (m_strHash.Cmp(tshRight.m_strHash) == 0);
}

TASTileType::TASTileType(int dims)
{
  m_dims          = dims;
  m_concentration = 1;
  m_colTile       = wxNullColour;
  m_colText       = wxNullColour;
}

TASTileType::TASTileType(const TASTileType& tt)
{
  SetEqual(tt);
}

void TASTileType::SetEqual(const TASTileType& tt)
{
  m_dims          = tt.m_dims;
  m_concentration = tt.m_concentration;
  m_sName         = tt.m_sName;
  m_sLabel        = tt.m_sLabel;
  m_sComment      = tt.m_sComment;
  for (int i=TAS_NORTH; i<=TAS_DOWN; i++) {
    m_sides[i]    = tt.m_sides[i];
  }
  m_image         = tt.m_image;
  m_colTile       = tt.m_colTile;
  m_colText       = tt.m_colText;
  m_sImagePath    = tt.m_sImagePath;

  InvalidateBitmaps();
}

TASTileType::TASTileType(const wxString& sTTDefn, size_t* index)
{
  m_dims          = 2;
  m_concentration = 1;
  m_colTile       = wxNullColour;
  m_colText       = wxNullColour;

  size_t len      = sTTDefn.Len();
  size_t newIndex = *index;
  while ((newIndex < len) && (sTTDefn[newIndex] != '\n')) {
    newIndex++;
  }
  wxString sLine = sTTDefn.SubString(*index, newIndex);
  newIndex++;
  *index = newIndex;
  while (sLine.length() > 0) {
    sLine.Trim().Trim(false);

    //  Parse current line and get value
    if (sLine.StartsWith(_T("TILENAME"))) {
      m_sName = sLine.SubString(8, sLine.Len()).Trim(false);
    } else if (sLine.StartsWith(_T("LABEL"))) {
      m_sLabel = sLine.SubString(5, sLine.Len()).Trim(false);
    } else if (sLine.StartsWith(_T("NORTHBIND"))) {
      sLine = sLine.SubString(9, sLine.Len()).Trim(false);
      long strength = 0;
      if (sLine.IsNumber() && sLine.ToLong(&strength)) {
        m_sides[TAS_NORTH].m_iStrength = strength;
      }
    } else if (sLine.StartsWith(_T("EASTBIND"))) {
      sLine = sLine.SubString(8, sLine.Len()).Trim(false);
      long strength = 0;
      if (sLine.IsNumber() && sLine.ToLong(&strength)) {
        m_sides[TAS_EAST].m_iStrength = strength;
      }
    } else if (sLine.StartsWith(_T("SOUTHBIND"))) {
      sLine = sLine.SubString(9, sLine.Len()).Trim(false);
      long strength = 0;
      if (sLine.IsNumber() && sLine.ToLong(&strength)) {
        m_sides[TAS_SOUTH].m_iStrength = strength;
      }
    } else if (sLine.StartsWith(_T("WESTBIND"))) {
      sLine = sLine.SubString(8, sLine.Len()).Trim(false);
      long strength = 0;
      if (sLine.IsNumber() && sLine.ToLong(&strength)) {
        m_sides[TAS_WEST].m_iStrength = strength;
      }
    } else if (sLine.StartsWith(_T("UPBIND"))) {
      m_dims = 3;
      sLine = sLine.SubString(6, sLine.Len()).Trim(false);
      long strength = 0;
      if (sLine.IsNumber() && sLine.ToLong(&strength)) {
        m_sides[TAS_UP].m_iStrength = strength;
      }
    } else if (sLine.StartsWith(_T("DOWNBIND"))) {
      m_dims = 3;
      sLine = sLine.SubString(8, sLine.Len()).Trim(false);
      long strength = 0;
      if (sLine.IsNumber() && sLine.ToLong(&strength)) {
        m_sides[TAS_DOWN].m_iStrength = strength;
      }
    } else if (sLine.StartsWith(_T("NORTHLABEL"))) {
      m_sides[TAS_NORTH].m_sLabel = sLine.SubString(10, sLine.Len()).Trim(false);
    } else if (sLine.StartsWith(_T("EASTLABEL"))) {
      m_sides[TAS_EAST].m_sLabel = sLine.SubString(9, sLine.Len()).Trim(false);
    } else if (sLine.StartsWith(_T("SOUTHLABEL"))) {
      m_sides[TAS_SOUTH].m_sLabel = sLine.SubString(10, sLine.Len()).Trim(false);
    } else if (sLine.StartsWith(_T("WESTLABEL"))) {
      m_sides[TAS_WEST].m_sLabel = sLine.SubString(9, sLine.Len()).Trim(false);
    } else if (sLine.StartsWith(_T("UPLABEL"))) {
      m_dims = 3;
      m_sides[TAS_UP].m_sLabel = sLine.SubString(7, sLine.Len()).Trim(false);
    } else if (sLine.StartsWith(_T("DOWNLABEL"))) {
      m_dims = 3;
      m_sides[TAS_DOWN].m_sLabel = sLine.SubString(9, sLine.Len()).Trim(false);
    } else if (sLine.StartsWith(_T("CONCENTRATION"))) {
      sLine = sLine.SubString(13, sLine.Len()).Trim(false);
      long concentration = 0;
      if (sLine.IsNumber() && sLine.ToLong(&concentration)) {
       m_concentration = concentration;
      }
    } else if (sLine.StartsWith(_T("TILECOLOR"))) {
      wxString sColor = sLine.SubString(9, sLine.Len()).Trim(false);
      if (sColor.Len() > 0) {
        if (sColor.StartsWith(_T("RGB("))) {
          m_colTile.Set(sColor);
        } else {
          m_colTile = sColor;
        }
      }
    } else if (sLine.StartsWith(_T("TEXTCOLOR"))) {
      wxString sColor = sLine.SubString(9, sLine.Len()).Trim(false);
      if (sColor.Len() > 0) {
        if (sColor.StartsWith(_T("RGB("))) {
          m_colText.Set(sColor);
        } else {
          m_colText = sColor;
        }
      }
    } else if (sLine.StartsWith(_T("icon"))) {
      wxFileName fnImg;
      fnImg.AssignDir(g_pContext->m_fnTilesetFile.GetPath());
      m_sImagePath = sLine.SubString(4, sLine.Len()).Trim(false);
      fnImg.SetFullName(m_sImagePath);
      m_image = wxImage(fnImg.GetFullPath());
      if (!m_image.IsOk()) {
        wxString sBadImage = wxString::Format(_T("Failed to load image for tiletype '%s'\n"), m_sName.c_str());
        g_pTASApp->SetEditorOutputText(sBadImage, true, true);
      }
    } else if (sLine.StartsWith(_T("CREATE"))) {
      break;
    }

    while ((newIndex < len) && (sTTDefn[newIndex] != '\n')) {
      newIndex++;
    }
    sLine = sTTDefn.SubString(*index, newIndex);
    newIndex++;
    *index = newIndex;
  }
}

TASTileType::~TASTileType()
{
}

bool TASTileType::IsValid()
{
return (m_sName.Len() > 0);
/*
  if (m_dims == 2) {
    if ((m_iStrengths[4] != -99) || (m_iStrengths[5] != -99) ||
      (m_glues[4].m_strLabel.Len() > 0 || m_glues[5].m_strLabel.Len() >0))
    {
      return false;
    } else {
      return true;
    }
  } else if (m_dims == 3) {
    return true;
  }
  return false;
*/
}

bool TASTileType::IsEqual(TASTileType* pTtype, bool bCompareNameAndLabel, bool bAll)
{
  if (!pTtype || (pTtype->m_dims != m_dims)) {
    return false;
  }

  if (bCompareNameAndLabel || bAll) {
    if ((pTtype->m_sName.CompareTo(m_sName.c_str()) != 0) ||
        (pTtype->m_sLabel.CompareTo(m_sLabel.c_str()) != 0))
    {
      return false;
    }
  }

  if (bAll) {
    if ((pTtype->m_colText != m_colText) || (pTtype->m_colTile != m_colTile))
    {
      return false;
    }
  }

  for (int i=TAS_NORTH; i<TAS_DOWN; i++) {
    if (m_sides[i] != pTtype->m_sides[i]) {
      return false;
    }
  }

  return true;
}

//  Rotation is done in the clockwise direction
bool TASTileType::Rotate(int degrees)
{
  if (degrees != 0 && degrees != 90 && degrees != 180 && degrees != 270 && degrees != 360)
  {
    return false;
  }
  
  TASTileSide ts[4];
  for (int j=0; j<4; j++) {
    ts[j] = m_sides[j];
  }

  int rotCount = degrees / 90;
  for (int i=0; i<4; i++) {
    m_sides[(i+rotCount)%4] = ts[i];
  }
  InvalidateBitmaps();
  return true;
}

wxString TASTileType::Serialize()
{
  wxString outStr;
  
  outStr += _T("TILENAME ") + m_sName + _T("\n");
  outStr += _T("LABEL ") + m_sLabel + _T("\n");
  if (m_concentration > 1) {
    outStr += wxString::Format(_T("CONCENTRATION %d\n"), m_concentration);
  }
  for (int i=TAS_NORTH; i<=(m_dims==2?TAS_WEST:TAS_DOWN); i++) {
    outStr += TASDirectionNumToString(i, true) + _T("BIND ") + wxString::Format(_T("%d"), m_sides[i].m_iStrength) + _T("\n");
  }
  for (int i=TAS_NORTH; i<=(m_dims==2?TAS_WEST:TAS_DOWN); i++) {
    outStr += TASDirectionNumToString(i, true) + _T("LABEL ") + m_sides[i].m_sLabel + _T("\n");
  }
  if (m_colTile != wxNullColour) {
    outStr += _T("TILECOLOR ") + m_colTile.GetAsString() + _T("\n");
  }
  if (m_colText != wxNullColour) {
    outStr += _T("TEXTCOLOR ") + m_colText.GetAsString() + _T("\n");
  }
  if (m_image.IsOk() && (m_sImagePath.Length() > 0)) {
    outStr += _T("icon ") + m_sImagePath + _T("\n");
  }
  outStr += _T("CREATE\n\n");

  return outStr;
}

wxBitmap& TASTileType::GetBitmap(size_t dim)
{
  if (m_vecBitmaps.size() <= dim) {
    m_vecBitmaps.resize(dim+1);
  }
  return m_vecBitmaps[dim].m_bitmap;
}

void TASTileType::InvalidateBitmaps()
{
  m_vecBitmaps.clear();
  m_bitmapFlat3D = wxBitmap();
}


wxFont* g_pFontLabel = NULL;
wxFont* g_pFontLabelSVG = NULL;
wxFont* g_pFontGlues = NULL;
wxFont* g_pFontGluesSVG = NULL;
wxFont* g_pFontGlues3D = NULL;

bool bHighQuality = true;  //  I think this should stay set to 'true', but performance needs to be tested more extensively

//                  TAS_NORTH, TAS_EAST, TAS_SOUTH, TAS_WEST, TAS_UP, TAS_DOWN
int xDirOffset[] = {        1,        2,         1,        0,      1,        1};
int yDirOffset[] = {        1,        1,         3,        1,      0,        2};

void TASTileType::DrawFlat3DSide(wxDC* pDC, wxPoint& pt, int dim, int side)
{
  int strength = m_sides[side].m_iStrength;
  if (strength > 0) {
    m_pen.SetStyle(wxSOLID);
    pDC->SetPen(m_pen);
  } else {
    m_pen.SetStyle(wxDOT);
    pDC->SetPen(m_pen);
  }

  int xOffset = pt.x + (dim * xDirOffset[side]);
  int yOffset = pt.y + (dim * yDirOffset[side]);
  pDC->DrawRectangle(xOffset, yOffset, dim, dim);

  if (strength > 1) {
    int dblBndSep = 4;
    pDC->DrawRectangle(xOffset + dblBndSep,
                       yOffset + dblBndSep,
                       dim - (2*dblBndSep),
                       dim - (2*dblBndSep));
  }

  if (strength > 0) {
    wxCoord w, h;
    pDC->GetTextExtent(m_sides[side].m_sLabel, &w, &h);

    wxCoord xStart = xOffset + (dim / 2) - (w / 2);
    wxCoord yStart = yOffset + (dim / 2) - (h / 2);
    pDC->DrawText(m_sides[side].m_sLabel, xStart, yStart);
  }
}

void TASTileType::DrawFlat3DTile(wxDC* pDC, wxPoint& pt, wxSize& sz)
{
  if (!m_bitmapFlat3D.IsOk()) {
    wxMemoryDC dcMem;
    if (!dcMem.IsOk()) {    //MJP: ERROR condition!
      return;
    }

    if (bHighQuality) {
      double scaleX = (double)sz.GetWidth() / (double)(g_tileSize * 3);
      double scaleY = (double)sz.GetHeight() / (double)(g_tileSize * 4);
      m_bitmapFlat3D.Create(sz.GetWidth(), sz.GetHeight());
      dcMem.SetUserScale(scaleX, scaleY);
    } else {
      m_bitmapFlat3D.Create((g_tileSize * 3), (g_tileSize * 4));
    }
    dcMem.SelectObject(m_bitmapFlat3D);

    wxPoint pt(0, 0);

    m_brush.SetColour(g_pContext->m_colSpaceBkgrnd);
    m_pen.SetColour(g_pContext->m_colSpaceBkgrnd);
    dcMem.SetBrush(m_brush);
    m_pen.SetStyle(wxSOLID);
    dcMem.SetPen(m_pen);

    dcMem.DrawRectangle(pt.x, pt.y, (g_tileSize * 3), (g_tileSize * 4));

    if (m_colTile == wxNullColour) {
      m_brush.SetColour(g_pContext->m_colDefaultTile);
    } else {
      m_brush.SetColour(m_colTile);
    }
    dcMem.SetBrush(m_brush);

    dcMem.SetFont(*g_pFontGlues3D);
    if (m_colText == wxNullColour) {
      m_pen.SetColour(g_pContext->m_colDefaultText);
      dcMem.SetTextForeground(g_pContext->m_colDefaultText);
    } else {
      m_pen.SetColour(m_colText);
      dcMem.SetTextForeground(m_colText);
    }
    dcMem.SetPen(m_pen);

    wxString s(_T("Name:"));
    wxCoord w, h;
    dcMem.DrawText(s, 3, 3);
    dcMem.GetTextExtent(s, &w, &h);
    dcMem.DrawText(m_sName, 3, h + 8);

    s = _T("Label:");
    dcMem.DrawText(s, (2 * g_tileSize) + 3, 3);
    dcMem.DrawText(m_sLabel, (2 * g_tileSize) + 3, h + 8);

    for (int dir=TAS_NORTH; dir<=TAS_DOWN; dir++) {
      DrawFlat3DSide(&dcMem, pt, g_tileSize, dir);
    }
  }

  if (m_bitmapFlat3D.IsOk() && pDC) {
    pDC->DrawBitmap(m_bitmapFlat3D, pt.x, pt.y, false);
  }
}

void TASTileType::DrawTile(wxDC* pDC, wxPoint& pt, size_t width, size_t height, size_t displayMask)
{
  if (m_vecBitmaps.size() <= width) {
    m_vecBitmaps.resize(width+1);
  }

  if (!m_vecBitmaps[width].m_bitmap.IsOk() ||
      (m_vecBitmaps[width].m_height != height) ||
      (m_vecBitmaps[width].m_displayMask != displayMask)) {

    if (!m_image.IsOk() && m_sImagePath.Length() > 0) {
      m_image = wxImage(m_sImagePath);
    }

    if (m_image.IsOk()) {   //  This tiletype has an image designated by the input file
      m_vecBitmaps[width].m_bitmap = wxBitmap(m_image.Scale(width, height));
      m_vecBitmaps[width].m_height = height;
      m_vecBitmaps[width].m_displayMask = displayMask;
    } else {
      m_vecBitmaps[width].m_height = height;
      m_vecBitmaps[width].m_displayMask = displayMask;

      wxMemoryDC dcMem;
      if (bHighQuality) {
        double scaleWidth  = (double)width  / (double)g_tileSize;
        double scaleHeight = (double)height / (double)g_tileSize;
        m_vecBitmaps[width].m_bitmap.Create(width, height);
        dcMem.SelectObject(m_vecBitmaps[width].m_bitmap);
        dcMem.SetUserScale(scaleWidth, scaleHeight);
      } else {
        m_vecBitmaps[width].m_bitmap.Create(g_tileSize, g_tileSize);
        dcMem.SelectObject(m_vecBitmaps[width].m_bitmap);
      }

      if (!dcMem.IsOk()) {    //MJP: ERROR condition!
        return;
      }

      wxPoint pt(0, 0);

      if (m_colTile == wxNullColour) {
        m_brush.SetColour(g_pContext->m_colDefaultTile);
        m_pen.SetColour(g_pContext->m_colDefaultTile);
      } else {
        m_brush.SetColour(m_colTile);
        m_pen.SetColour(m_colTile);
      }

      dcMem.SetBrush(m_brush);
      m_pen.SetStyle(wxSOLID);
      dcMem.SetPen(m_pen);
      dcMem.DrawRectangle(pt.x, pt.y, g_tileSize, g_tileSize);

      m_pen.SetColour(g_pContext->m_colDefaultText);
      dcMem.SetPen(m_pen);

      if (TAS_MASK_ISON(displayMask, TAS_MASK_DISPGLUES) || TAS_MASK_ISON(displayMask, TAS_MASK_DISPSTRENGTHS)) {
        m_pen.SetWidth(3);
        if (m_colText == wxNullColour) {
          m_pen.SetColour(g_pContext->m_colDefaultText);
          dcMem.SetTextForeground(g_pContext->m_colDefaultText);
        } else {
          m_pen.SetColour(m_colText);
          dcMem.SetTextForeground(m_colText);
        }

        dcMem.SetFont(*g_pFontGlues);
        for (int i=TAS_NORTH; i<TAS_UP; i++) {
          DrawBond(&dcMem, pt, i, TAS_MASK_ISON(displayMask, TAS_MASK_DISPGLUES));
        }

        wxCoord w, h;
        if (TAS_MASK_ISON(displayMask, TAS_MASK_DISPLABEL)) {
          dcMem.SetFont(*g_pFontLabel);
          dcMem.GetTextExtent(m_sLabel, &w, &h);
          dcMem.DrawText(m_sLabel, pt.x + (50-(w/2)), pt.y + 40);
        }
        dcMem.SetFont(*wxSWISS_FONT);
        if (TAS_MASK_ISON(displayMask, TAS_MASK_DISPNAME)) {
          dcMem.GetTextExtent(m_sName, &w, &h);
          dcMem.DrawText(m_sName, pt.x + (50-(w/2)), pt.y + 57);
        }
      }

      //  Scale the bitmap appropriately
      if(!bHighQuality && (m_vecBitmaps[width].m_bitmap.GetHeight() != height)) {
        wxImage image = m_vecBitmaps[width].m_bitmap.ConvertToImage();
        m_vecBitmaps[width].m_bitmap = wxBitmap(image.Scale(width, height));
      }
    }
  }

  if (m_vecBitmaps[width].m_bitmap.IsOk() && pDC) {
    pDC->DrawBitmap(m_vecBitmaps[width].m_bitmap, pt.x, pt.y, false);
  }
}

void TASTileType::DrawBond(wxDC* pDC, wxPoint& pt, int dir, bool bShowGlues, bool WXUNUSED(bSVG))
{
  int b = 2;
  int e = g_tileSize - b;
  int g = 5;
  int tg = 10;

  int strength = m_sides[dir].m_iStrength;

  wxCoord w, h;
  w = h = 0;

  if (bShowGlues) {
    pDC->GetTextExtent(m_sides[dir].m_sLabel, &w, &h);
    w = w/2;
  }

  if (strength > 0) {
    m_pen.SetStyle(wxSOLID);
    pDC->SetPen(m_pen);
  } else {
    m_pen.SetStyle(wxDOT);
    pDC->SetPen(m_pen);
  }

  if (dir == TAS_NORTH) {
    pDC->DrawLine(pt.x + b, pt.y + b, pt.x + e, pt.y + b);
    if (strength >= 2) {
      pDC->DrawLine(pt.x + b, pt.y + b+g, pt.x + e, pt.y + b+g);
      if (strength > 2) {
        pDC->DrawLine(pt.x + b, pt.y + b+(g/2), pt.x + e, pt.y + b+(g/2));
      }
    }
    if (strength > 0 && bShowGlues) {
      pDC->DrawText(m_sides[dir].m_sLabel, pt.x + ((e-b)/2)-w, pt.y + b+h-4);
    }
  } else if (dir == TAS_EAST) {
    pDC->DrawLine(pt.x + e, pt.y + b, pt.x + e, pt.y + e);
    if (strength >= 2) {
      pDC->DrawLine(pt.x + e-g, pt.y + b, pt.x + e-g, pt.y + e);
      if (strength > 2) {
        pDC->DrawLine(pt.x + e-(g/2), pt.y + b, pt.x + e-(g/2), pt.y + e);
      }
    }
    if (strength > 0 && bShowGlues) {
      wxCoord cX = pt.x + (g_tileSize - 10);
      wxCoord cY = pt.y + (g_tileSize/2) - w;
      pDC->DrawRotatedText(m_sides[dir].m_sLabel, cX, cY, 270.0);//pt.x + ((e-b)/2)-w, pt.y -e+h-3, 270.0);
    }
  } else if (dir == TAS_SOUTH) {
    pDC->DrawLine(pt.x + b, pt.y + e, pt.x + e, pt.y + e);
    if (strength >= 2) {
      pDC->DrawLine(pt.x + b, pt.y + e-g, pt.x + e, pt.y + e-g);
      if (strength > 2) {
        pDC->DrawLine(pt.x + b, pt.y + e-(g/2), pt.x + e, pt.y + e-(g/2));
      }
    }
    if (strength > 0 && bShowGlues) {
      pDC->DrawText(m_sides[dir].m_sLabel, pt.x + ((e-b)/2)-w, pt.y + e-tg-h);
    }
  } else if (dir == TAS_WEST) {
    pDC->DrawLine(pt.x + b, pt.y + b, pt.x + b, pt.y + e);
    if (strength >= 2) {
      pDC->DrawLine(pt.x + b+g, pt.y + b, pt.x + b+g, pt.y + e);
      if (strength > 2) {
        pDC->DrawLine(pt.x + b+(g/2), pt.y + b, pt.x + b+(g/2), pt.y + e);
      }
    }
    if (strength > 0 && bShowGlues) {
      wxCoord cX = pt.x + h;
      wxCoord cY = pt.y + (g_tileSize/2) + w;
      pDC->DrawRotatedText(m_sides[dir].m_sLabel, cX, cY, 90.0);//pt.x + (b-e)/2-(w+4), pt.y + h, 90.0);//((e-b)/2)-w, -h-tg-2);
    }
  }
}

/*
void TASTileType::DrawSVG(wxDC* pDC, wxPoint& pt, int dim, unsigned displayMask)
{
  if (!m_image.IsOk() && m_sImagePath.Length() > 0) {
    m_image = wxImage(m_sImagePath);
  }

  if (m_image.IsOk()) {   //  This tiletype has an image designated by the input file
//MJP:  not handled
  } else {
    if (!pDC->IsOk()) {    //MJP: ERROR condition!
      return;
    }

//    if (bHighQuality) {
//      double scale = (double)dim / (double)g_tileSize;
//      pDC->SetUserScale(scale, scale);
//    }

    if (m_colTile == wxNullColour) {
      m_brush.SetColour(g_pContext->m_colDefaultTile);
      m_pen.SetColour(g_pContext->m_colDefaultTile);
    } else {
      m_brush.SetColour(m_colTile);
      m_pen.SetColour(m_colTile);
    }

    pDC->SetBrush(m_brush);
    m_pen.SetStyle(wxSOLID);
    pDC->SetPen(m_pen);
    pDC->DrawRectangle(pt.x, pt.y, g_tileSize, g_tileSize);

    m_pen.SetColour(g_pContext->m_colDefaultText);
    pDC->SetPen(m_pen);

    if (TAS_MASK_ISON(displayMask, TAS_MASK_DISPGLUES) || TAS_MASK_ISON(displayMask, TAS_MASK_DISPSTRENGTHS)) {
//      m_pen.SetWidth(3);
      if (m_colText == wxNullColour) {
        m_pen.SetColour(g_pContext->m_colDefaultText);
        pDC->SetTextForeground(g_pContext->m_colDefaultText);
      } else {
        m_pen.SetColour(m_colText);
        pDC->SetTextForeground(m_colText);
      }

      pDC->SetFont(*g_pFontGluesSVG);
      for (int i=TAS_NORTH; i<TAS_UP; i++) {
        DrawBond(pDC, pt, i, TAS_MASK_ISON(displayMask, TAS_MASK_DISPGLUES));
      }

      wxCoord w, h;
      if (TAS_MASK_ISON(displayMask, TAS_MASK_DISPLABEL)) {
        pDC->SetFont(*g_pFontLabelSVG);
        pDC->GetTextExtent(m_sLabel, &w, &h);
        pDC->DrawText(m_sLabel, pt.x + (50-(w/2)), pt.y + 40);
      }
      pDC->SetFont(*wxSWISS_FONT);
      if (TAS_MASK_ISON(displayMask, TAS_MASK_DISPNAME)) {
        pDC->GetTextExtent(m_sName, &w, &h);
        pDC->DrawText(m_sName, pt.x + (50-(w/2)), pt.y + 57);
      }
    }
  }
}
*/

//  An example tile type definition to be read in from inFile:
//LABEL Y
//TILENAME Y2
//EASTBIND 0
//WESTBIND 0
//SOUTHBIND 2
//NORTHBIND 2
//UPBIND 0
//DOWNBIND 0
//WESTLABEL -99
//EASTLABEL -99
//NORTHLABEL y2
//SOUTHLABEL y1
//UPLABEL -99
//DOWNLABEL -99
//TILECOLOR 5693267
//TEXTCOLOR 4103935
//CREATE

TASFrontierLocation* TASFrontierSet::GetElement(size_t ind)
{
  if (ind >= size()) {
    return NULL;
  } else if (ind <= (size()/2)) {
    TASFrontierSet::iterator it = begin();
    while (ind) {
      it++;
      ind--;
    }
    return (TASFrontierLocation*)&(*it);
  } else {
    ind = (size() - ind) - 1;
    TASFrontierSet::reverse_iterator it = rbegin();
    while (ind) {
      it++;
      ind--;
    }
    return (TASFrontierLocation*)&(*it);
  }
}

bool TASFrontierSet::RemoveElement(const TASCoords& coords)
{
  TASFrontierLocation tfl(coords);
  TASFrontierSet::iterator it = find(tfl);
  if (it != end()) {
    erase(it);
    return true;
  }
  return false;
}

bool TASTileSpace::LoadSpaceDefinitionFromFile(wxFFile* inFile)
{
  wxString sFileContents;
  if (inFile->IsOpened() && inFile->ReadAll(&sFileContents)) {
     return LoadSpaceDefinitionFromString(sFileContents);
  }
  return false;
}

bool TASTileSpace::LoadSpaceDefinitionFromString(wxString sSeedDefn)
{
  ResetSpace();
  m_seed.m_iAvgX = 0;
  m_seed.m_iAvgY = 0;
  m_seed.m_iAvgZ = 0;
  bool bFirst = true;
  bool bLoadedStages = false;
  TASCoords coordsZero;
  m_seed.m_coordsLeftmost = coordsZero;

  int sumX, sumY, sumZ, numX, numY, numZ;
  sumX = sumY = sumZ = numX = numY = numZ = 0;

  TASCoords coordsMin(0, 0, 0);
  TASCoords coordsMax(0, 0, 0);

  wxString strTtypeName;

  wxString sLine = sSeedDefn.BeforeFirst('\n');
  sSeedDefn = sSeedDefn.AfterFirst('\n');

  m_seed.clear();

  wxString sMsg = _T("=============================================\n");
  sMsg += _T("Attempting load of the following tile assembly:\n");
  sMsg += sSeedDefn;
  sMsg += _T("\n");
  sMsg += _T("=============================================\n");
  g_pTASApp->SetSimDebugText(sMsg, true, false);

  bool bFailedMsgBox  = false;
  bool bFoundTemp     = false;
  bool bInSupertile   = false;

  while (sSeedDefn.Len() > 0 || sLine.Len() > 0) {
    wxStringTokenizer tok(sLine);
    wxString sTileName;

    long lX, lY, lZ;
    lX = lY = lZ = 0;

    if (tok.CountTokens() == 3) {           //  2-D tile
      sTileName = tok.GetNextToken();
      if (!tok.GetNextToken().ToLong(&lX) ||
          !tok.GetNextToken().ToLong(&lY))
      {
        wxString sMsg;
        if (!bFailedMsgBox) {
          sMsg  = _T("The assembly could not be correctly loaded. Please see\n");
          sMsg += _T("the Output (and Debug) message pane(s) for details.");
          wxMessageBox(sMsg, _T("Error loading assembly"), wxICON_ERROR|wxOK);
          bFailedMsgBox = true;
        }
        sMsg = wxString::Format(_T("Load of assembly failed:  The following line contained an invalid coordinate for the tile location:\n%s\n"), sLine.c_str());
        g_pTASApp->SetSimOutputText(sMsg, true, false);
      } else {
        TASCoords c(lX, lY, lZ);
        TASSeedEntry se(sTileName, c);
        m_seed.push_back(se);

        numX++;
        numY++;
        sumX += (int)lX;
        sumY += (int)lY;
        if ((int)lX < coordsMin.x) coordsMin.x = (int)lX;
        if ((int)lY < coordsMin.y) coordsMin.y = (int)lY;
        if (coordsMax.x < (int)lX) coordsMax.x = (int)lX;
        if (coordsMax.y < (int)lY) coordsMax.y = (int)lY;

        if (bFirst) {
          m_seed.m_coordsLeftmost = c;
          bFirst = false;
        } else if (c.x < m_seed.m_coordsLeftmost.x) {
          m_seed.m_coordsLeftmost = c;
        } else if ((c.x == m_seed.m_coordsLeftmost.x) && (c.y < m_seed.m_coordsLeftmost.y)) {
          m_seed.m_coordsLeftmost = c;
        }

      }
    } else if (tok.CountTokens() == 4) {    //  3-D tile
      sTileName = tok.GetNextToken();
      if (!tok.GetNextToken().ToLong(&lX) ||
          !tok.GetNextToken().ToLong(&lY) ||
          !tok.GetNextToken().ToLong(&lZ))
      {
        wxString sMsg;
        if (!bFailedMsgBox) {
          sMsg  = _T("The assembly could not be correctly loaded. Please see\n");
          sMsg += _T("the Output (and Debug) message pane(s) for details.");
          wxMessageBox(sMsg, _T("Error loading assembly"), wxICON_ERROR|wxOK);
          bFailedMsgBox = true;
        }
        sMsg = wxString::Format(_T("Load of assembly failed:  The following line contained an invalid coordinate for the tile location:\n%s\n"), sLine.c_str());
        g_pTASApp->SetSimOutputText(sMsg, true, false);
      } else {
        TASCoords c(lX, lY, lZ);
        TASSeedEntry se(sTileName, c);
        m_seed.push_back(se);

        numX++;
        numY++;
        numZ++;
        sumX += (int)lX;
        sumY += (int)lY;
        sumZ += (int)lZ;
        if ((int)lX < coordsMin.x) coordsMin.x = (int)lX;
        if ((int)lY < coordsMin.y) coordsMin.y = (int)lY;
        if ((int)lZ < coordsMin.z) coordsMin.z = (int)lZ;
        if (coordsMax.x < (int)lX) coordsMax.x = (int)lX;
        if (coordsMax.y < (int)lY) coordsMax.y = (int)lY;
        if (coordsMax.z < (int)lZ) coordsMax.z = (int)lZ;

        if (bFirst) {
          m_seed.m_coordsLeftmost = c;
          bFirst = false;
        } else if (c.x < m_seed.m_coordsLeftmost.x) {
          m_seed.m_coordsLeftmost = c;
        } else if ((c.x == m_seed.m_coordsLeftmost.x) && (c.y < m_seed.m_coordsLeftmost.y)) {
          m_seed.m_coordsLeftmost = c;
        }
      }
    } else {
      bool bError = false;
      if ((tok.CountTokens() == 1) && (sLine.StartsWith(sTempPrefix))) {
        wxString sTemp = sLine.SubString(sTempPrefix.Len(), sLine.Len());
        long lTemp = 0;
        if (sTemp.Contains(_T(";"))) {
          bFoundTemp = true;
          m_vectTemps.clear();
          m_currTemp = -1;
          wxArrayString arrStrTemps = wxStringTokenize(sTemp, _T(";"));
          if (arrStrTemps.Count() <= 0) {
            bFoundTemp = false;
            bError = true;
            wxString sErr = wxString::Format(_T("An invalid temperature value or sequence was specified: '%s'\n"), sLine.c_str());
            g_pTASApp->SetSimOutputText(sErr, true, false);
          }
          for (size_t i=0; i<arrStrTemps.Count(); i++) {
            wxString sCurr = arrStrTemps[i];
            long lTemp = 0;
            if (sCurr.IsNumber() && sCurr.ToLong(&lTemp)) {
              m_vectTemps.push_back((int)lTemp);
            } else {
              bFoundTemp = false;
              bError = true;
              wxString sErr = wxString::Format(_T("An invalid temperature value or sequence was specified: '%s'\n"), sLine.c_str());
              g_pTASApp->SetSimOutputText(sErr, true, false);
            }
          }
          if (bFoundTemp && !bError) {
            m_currTemp = 0;
            g_pTASApp->SetTemp(m_vectTemps[m_currTemp]);
          }
        } else if (sTemp.IsNumber() && sTemp.ToLong(&lTemp)) {
          bFoundTemp = true;
          g_pTASApp->SetTemp((int)lTemp);
        } else {
          bError = true;
		}
	  } else if ((g_pContext->m_iTASSimMode == TAS_KTAM_MODE) && (tok.CountTokens() == 1) && (sLine.StartsWith(sGmcPrefix))) {
		double lgmc = 0;
		wxString sGmc = sLine.SubString(sGmcPrefix.Len(), sLine.Len());
		if(sGmc.ToDouble(&lgmc)) {
		  g_pkTAMTileSpace->setGmc(lgmc);
		} else {
		  bError = true;
		}
	  } else if ((g_pContext->m_iTASSimMode == TAS_KTAM_MODE) && (tok.CountTokens() == 1) && (sLine.StartsWith(sGsePrefix))) {
		double lgse = 0;
		wxString sGse = sLine.SubString(sGsePrefix.Len(), sLine.Len());
		if(sGse.ToDouble(&lgse)) {
			g_pkTAMTileSpace->setGse(lgse);
		} else {
		  bError = true;
		}
      } else if ((tok.CountTokens() == 1) && (sLine.StartsWith(sStagesPrefix))) {
        wxString sStages = sLine.SubString(sStagesPrefix.Len(), sLine.Len());
        if (g_pTASApp->LoadStageDefinitions(sStages)) {
          bLoadedStages = true;
//          DoRefresh();
        } else {
          wxString sMsg = wxString::Format(_T("Load of stage definitions file '%s' failed.\n"), sStages.c_str());
          g_pTASApp->SetSimOutputText(sMsg, true, false);
        }
      } else if ((tok.CountTokens() == 1) && (sLine.StartsWith(sModePrefix))) {
        // Should be taken care of before we get here
      } else if ((g_pContext->m_iTASSimMode == TAS_2HAM_MODE) && (tok.CountTokens() == 1) && (sLine.StartsWith(sIncludeSingletons))) {
        wxString sInclude = sLine.SubString(sIncludeSingletons.Len(), sLine.Len());
        if (sInclude.Contains(_T("false"))) {
          //  Since this directive should be before any supertile definitions, we can safely
          //  just delete any existing supertile space because they must be the singletons
          g_p2HAMTileSpace->m_bIncludeSingletons = false;
          g_p2HAMTileSpace->ResetSpace();
        }
      } else if ((g_pContext->m_iTASSimMode == TAS_2HAM_MODE) && (tok.CountTokens() == 1) && (sLine.CmpNoCase(sBeginSupertile) == 0)) {
        if (!bInSupertile) {
          bInSupertile = true;
        } else {
          bError = true;
          wxString sErr = wxString::Format(_T("An invalid supertile definition was found.  Each definition should begin with a 'BeginSupertile' and end with an 'EndSupertile' line.\n"));
          g_pTASApp->SetSimOutputText(sErr, true, false);
        }
      } else if ((g_pContext->m_iTASSimMode == TAS_2HAM_MODE) && (tok.CountTokens() == 1) && (sLine.CmpNoCase(sEndSupertile) == 0)) {
        if (bInSupertile) {
//MJP TODO: figure out how to do this from a regular tile space
          //  Create a supertile from what we've got
          TASSupertile* pST = new TASSupertile(&m_seed, &m_tileset, TAS_2HAM_INFINITE_COUNT);  //TODO: fix this - hardcoded to infinite count
          if (pST->m_vectSTC.size() > 0) {
            int max = g_p2HAMTileSpace->m_maxSupertile;
            if (pST->m_vectSTC.size() > g_p2HAMTileSpace->m_maxSupertile) {
              max = pST->m_vectSTC.size();
            }
            if (pST->m_vectSTC[0].size() > g_p2HAMTileSpace->m_maxSupertile) {
              max = pST->m_vectSTC[0].size();
            }
            TASSupertileSpace* pSTS = new TASSupertileSpace(pST, max + (2 * g_p2HAMTileSpace->m_defaultPadding));
            g_p2HAMTileSpace->AddSupertileSpace(pSTS, 0);
          }
          m_seed.clear();
          bInSupertile = false;
        } else {
          bError = true;
          wxString sErr = wxString::Format(_T("An invalid supertile definition was found.  Each definition should begin with a 'BeginSupertile' and end with an 'EndSupertile' line.\n"));
          g_pTASApp->SetSimOutputText(sErr, true, false);
        }
      } else if ((sLine.Trim().Trim(false).Len() != 0) && !tok.GetNextToken().EndsWith(_T(".tds"))) {
        bError = true;
      }
      if (bError /*|| (tok.CountTokens() != 1) || 
          ((!tok.GetNextToken().EndsWith(_T(".tds"))))) && !sLine.StartsWith(sTempPrefix) && !sLine.StartsWith(sStagesPrefix))) */)
      {
        wxString sMsg;
        if (!bFailedMsgBox) {
          sMsg  = _T("The assembly could not be correctly loaded. Please see\n");
          sMsg += _T("the Output (and Debug) message pane(s) for details.");
          wxMessageBox(sMsg, _T("Error loading assembly"), wxICON_ERROR|wxOK);
          bFailedMsgBox = true;
        }
        sMsg = wxString::Format(_T("Load of assembly failed:  The following line was invalid:\n%s\n\n"), sLine.c_str());
        g_pTASApp->SetSimOutputText(sMsg, true, false);
      }
    }

    sLine = sSeedDefn.BeforeFirst(_T('\n'));
    sSeedDefn = sSeedDefn.AfterFirst(_T('\n'));
  }

  if (!bFoundTemp) {        //  If a temperature is not explicitly specified, default to 2
    m_vectTemps.clear();
    m_currTemp = -1;
    g_pTASApp->SetTemp(2);
  }

  if (!bLoadedStages) {
    m_bUseStages = false;
    m_vectStages.clear();
    g_pContext->m_fnStageFile.Clear();
  }

  if (g_pContext->m_iTASSimMode != TAS_2HAM_MODE) {
    GrowSpace(coordsMin);
    GrowSpace(coordsMax);

    if (numX > 0) {
      m_seed.m_iAvgX = sumX / numX;
    }
    if (numY > 0) {
      m_seed.m_iAvgY = sumY / numY;
    }
    if (numZ > 0) {
      m_seed.m_iAvgZ = sumZ / numZ;
    }

    wxString sVal = ValidateSeed();
    if (sVal.Len() > 0) {
      wxString sMsg = wxString::Format(_T("Load of assembly failed:  No tile types matching the following names were found in the current tile set, so they were skipped: %s\n"), sVal.c_str());
      g_pTASApp->SetSimOutputText(sMsg, true, false);

      if (!bFailedMsgBox) {
        sMsg  = _T("The assembly could not be correctly loaded. Please see\n");
        sMsg += _T("the Output (and Debug) message pane(s) for details.");
        wxMessageBox(sMsg, _T("Error loading assembly"), wxICON_ERROR|wxOK);
      }
    }
    LoadSeed();
  }

  return true;
}

bool TASTileSpace::LoadStageDefinitionsFromFile(wxFFile* inFile)
{
  wxString sFileContents;
  if (inFile->IsOpened() && inFile->ReadAll(&sFileContents)) {

    wxString sLine = sFileContents.BeforeFirst('\n');
    sFileContents = sFileContents.AfterFirst('\n');

    long lStageNum = 0;
    std::vector<std::set<wxString> > vectStages;

    while (sFileContents.Len() > 0 || sLine.Len() > 0) {
      wxStringTokenizer tok(sLine);
      wxString sTileName;

      wxString sBegin = tok.GetNextToken();
      wxString sExpected = wxString::Format(_T("Stage%ld:"), lStageNum);

      if (sBegin != sExpected) {
        wxString sMsg = wxString::Format(_T("Load of stage definitions failed:  Invalid format for line '%s'.  Should begin with 'Stage<n>:'.\n"), sLine.c_str());
        g_pTASApp->SetSimOutputText(sMsg, true, false);

        sMsg  = _T("The stage definitions could not be correctly loaded. Please see\n");
        sMsg += _T("the Output (and Debug) message pane(s) for details.");
        wxMessageBox(sMsg, _T("Error loading stage definitions"), wxICON_ERROR|wxOK);
        return false;
      } else {
        std::set<wxString> stage;
        while (tok.HasMoreTokens()) {
          stage.insert(tok.GetNextToken());
        }
        vectStages.push_back(stage);
        lStageNum++;
      }

      sLine = sFileContents.BeforeFirst('\n');
      sFileContents = sFileContents.AfterFirst('\n');
    }

    //  Verify that all named tile types exist in the tile set
    for (size_t i=0; i<m_vectStages.size(); i++) {
      for (std::set<wxString>::iterator iter = vectStages[i].begin(); iter != vectStages[i].end(); iter++) {
        if (m_tileset.IndexOf(*iter) == TAS_TILE_NOT_FOUND) {
          wxString sMsg = wxString::Format(_T("Load of stage definitions failed:  Tile type '%s' for stage %d was not found in the current tile set.\n"), iter->c_str(), i);
          g_pTASApp->SetSimOutputText(sMsg, true, false);

          sMsg  = _T("The stage definitions could not be correctly loaded. Please see\n");
          sMsg += _T("the Output (and Debug) message pane(s) for details.");
          wxMessageBox(sMsg, _T("Error loading stage definitions"), wxICON_ERROR|wxOK);

          return false;
        }
      }
    }

    m_bUseStages = true;
    m_currStage = 0;
    m_vectStages = vectStages;
    return LoadSeed();
  }
  return false;
}

const int ID_TAS_SAVESPACEOPTIONS = 5001;

class TASSpaceSaveOptionsDlg : public wxDialog
{
public:
  TASSpaceSaveOptionsDlg(wxWindow* pWin);

  wxRadioBox* m_pRadioBox;
};

TASSpaceSaveOptionsDlg::TASSpaceSaveOptionsDlg(wxWindow* pWin)
                         : wxDialog(pWin, wxID_ANY, _T("Assembly output options"), wxDefaultPosition, wxDefaultSize)
{
  wxBoxSizer* pSizer = new wxBoxSizer(wxVERTICAL);

  wxString sText;
  sText += _T("Please select whether you want to save just the seed assembly\n");
  sText += _T("or the entire assembly as it currently exists.");

  wxStaticText* pText = new wxStaticText(this, wxID_ANY, sText);
  pSizer->Add(pText, 0, wxTOP | wxLEFT | wxRIGHT | wxBOTTOM, 10);

  wxString arrChoices[] = {_T("Seed only"), _T("Entire assembly") };
  m_pRadioBox = new wxRadioBox(this, ID_TAS_SAVESPACEOPTIONS, _T("Assembly to save"),
                               wxDefaultPosition, wxDefaultSize, 2, arrChoices);

  pSizer->Add(m_pRadioBox, 0, wxLEFT | wxRIGHT | wxBOTTOM, 10);

  wxBoxSizer* pSizerButtons = new wxBoxSizer(wxHORIZONTAL);
  wxButton* pOK = new wxButton(this, wxID_OK);
  pSizerButtons->Add(pOK);
  wxButton* pCancel = new wxButton(this, wxID_CANCEL);
  pSizerButtons->Add(pCancel, 0, wxLEFT | wxRIGHT, 10);
  pSizer->Add(pSizerButtons, 0, wxALIGN_RIGHT);
  pSizer->AddSpacer(5);

  SetAutoLayout(TRUE);
  SetSizer(pSizer);

  pSizer->SetSizeHints(this);
  pSizer->Fit(this);
}

bool TASTileSpace::SaveToFile(wxFFile* outFile)
{
  if (!outFile->IsOpened()) {
    return false;
  }

  //  See if the user wants to save the full current assembly
  //  or just the seed
  TASSpaceSaveOptionsDlg dlg(g_pTASApp->m_pSimFrame);
  if (dlg.ShowModal() == wxID_OK) {

    //  Write out the name of the current tileset file since
    //  this assembly is most likely associated with it
    outFile->Write(wxString::Format(_T("%s\n"), g_pContext->m_fnTilesetFile.GetFullName().c_str()));

    //  Write out the mode
    wxString sModeStr = TASGetModeName(g_pContext->m_iTASSimMode);
    outFile->Write(wxString::Format(_T("Mode=%s\n"), sModeStr.c_str()));

    //  Write out the current temperature value, or the temperature sequence if there is one
    if (m_vectTemps.size() > 0) {
      wxString sTempSeq = _T("Temperature=");
      for (size_t i=0; i<m_vectTemps.size(); i++) {
        if (i > 0) {
          sTempSeq += _T(";");
        }
        sTempSeq += wxString::Format(_T("%d"), m_vectTemps[i]);
      }
      sTempSeq += _T("\n");
      outFile->Write(sTempSeq);
    } else {
      outFile->Write(wxString::Format(_T("Temperature=%d\n"), GetTemp()));
    }

    if (dlg.m_pRadioBox->GetSelection() == 0) {     //  Save only the seed
      SetSeed();
      for (size_t i=0; i<m_seed.size(); i++) {
        outFile->Write(wxString::Format(_T("%s %d %d %d\n"), 
                                        m_seed[i].m_sName.c_str(),
                                        m_seed[i].m_coords.x, m_seed[i].m_coords.y, m_seed[i].m_coords.z));
      }
    } else {  //  Save the entire current assembly

      if (g_pContext->m_iTASSimMode == TAS_2HAM_MODE) {
        for (size_t i=0; i<g_p2HAMTileSpace->m_vectSTS.size(); i++) {
          outFile->Write(wxString::Format(_T("\n%s\n"), sBeginSupertile.c_str()));
          wxString sST;
          g_p2HAMTileSpace->m_vectSTS[i]->m_pST->PrintToString(&sST);
          outFile->Write(sST.c_str());
          outFile->Write(wxString::Format(_T("%s\n"), sEndSupertile.c_str()));
        }
      } else {
        for (size_t i=0; i<m_pLocationIndex->GetSize(); i++) {
          if ((m_pLocationIndex->GetLocation(i) != TAS_TILE_NOT_FOUND) &&
              (m_pLocationIndex->GetLocation(i) != TAS_TILE_FRONTIER))
          {
            TASCoords coords = CoordsFromIndex(i);

            outFile->Write(wxString::Format(_T("%s %d %d %d\n"), 
                                            m_tileset[m_pLocationIndex->GetLocation(i)]->m_sName.c_str(),
                                            coords.x, coords.y, coords.z));
          }
        }
      }
    }
  }
  outFile->Close();

  return true;
}

size_t TASTileSpace::GetRandSeed()
{
  if (g_pTASApp) {
    return g_pTASApp->m_randSeed;
  } else {
    return 0;
  }
}

void TASTileSpace::SetRandSeed(unsigned long seed, bool bRandomSeed)
{
  if (g_pTASApp) {
    g_pTASApp->Srand(seed, bRandomSeed);
  }
}

void TASTileSpace::SetDimNum(int dimNum)
{
  if (m_dimNum != dimNum) {
    m_dimNum  = dimNum;
    m_numNbrs = (m_dimNum == 2) ? TAS_WEST : TAS_DOWN;
    g_pContext->m_iTASDimMode = (dimNum == 2 ? TAS_2D_MODE : TAS_3D_MODE);
    ResetSpace(true);
  }
}

void TASTileSpace::ResetSpace(bool bSaveSeed)
{
  if (!bSaveSeed) {
    m_seed.clear();
  }

  m_epoch         = 0;
  m_numTiles      = 0;
  m_currStage     = 0;
  m_bEpochCached  = false;

  m_cutSets.Reset();

  if (m_vectTemps.size() > 0) {
    m_temp = m_vectTemps[0];
    m_currTemp = 0;
  }

  //  (Re)seed the random number generator
  SetRandSeed(g_pTASApp->m_randSeed);

  wxString sResetMsg = wxString::Format(_T("Reseting tile assembly space.\n"));
  g_pTASApp->SetSimOutputText(sResetMsg, true, false);

  if (m_dimNum == 2) {
    SetDims(TASCoords(-10, -10, 0), TASCoords(20, 20, 0));
  } else {
    //SetDims(TASCoords(-10, -10, -10), TASCoords(20, 20, 20));
    SetDims(TASCoords(-5, -5, -5), TASCoords(10, 10, 10));
  }

  LoadSeed();
}

TASSeed& TASSeed::operator=(TASSeed& tts)
{
  m_iAvgX           = tts.m_iAvgX;
  m_iAvgY           = tts.m_iAvgY;
  m_iAvgZ           = tts.m_iAvgZ;
  m_coordsLeftmost  = tts.m_coordsLeftmost;

  clear();
  for (size_t i=0; i<tts.size(); i++) {
    push_back(tts[i]);
  }

  return *this;
}

wxString TASTileSpace::ValidateSeed()
{
  wxString sRet;

  TASSeed seedTemp;
  seedTemp.m_iAvgX          = m_seed.m_iAvgX;
  seedTemp.m_iAvgY          = m_seed.m_iAvgY;
  seedTemp.m_iAvgZ          = m_seed.m_iAvgZ;
  seedTemp.m_coordsLeftmost = m_seed.m_coordsLeftmost;

  for (size_t i=0; i<m_seed.size(); i++) {
    if (m_tileset.IndexOf(m_seed[i].m_sName) == TAS_TILE_NOT_FOUND) {
      if (sRet.length() > 0) {
        sRet += _T(", ");
      }
      sRet += _T("'") + m_seed[i].m_sName + _T("'");
    } else {
      seedTemp.push_back(m_seed[i]);
    }
  }
  m_seed.clear();
  m_seed = seedTemp;

  return sRet;
}

bool TASTileSpace::LoadSeed()
{
  for (size_t i=0; i<m_seed.size(); i++) {
    if (m_tileset.IndexOf(m_seed[i].m_sName) != TAS_TILE_NOT_FOUND) {
      if (m_bUseStages && (m_vectStages.size() > 0) && (m_vectStages[0].find(m_seed[i].m_sName) == m_vectStages[0].end())) {
        wxString sMsg = wxString::Format(_T("The tile type named '%s' is not\nincluded in stage 0."), m_seed[i].m_sName.c_str());
        wxMessageBox(sMsg, _T("Invalid Seed"), wxICON_ERROR|wxOK);
        return false;
      } else {
        TASTileType* pTT = m_tileset.at(m_seed[i].m_sName);
        AddTile(m_seed[i].m_coords, pTT);
      }
    } else {
      wxString sMsg = wxString::Format(_T("No tile type matching the name '%s' exists\nin the current tile set.  Unable to load seed."), m_seed[i].m_sName.c_str());
      wxMessageBox(sMsg, _T("Invalid Seed"), wxICON_ERROR|wxOK);
      return false;
    }
  }
  m_numTiles = m_seed.size();

  //  Check to see if the seed is tau-stable and warn if it isn't
  bool bTauStable = true;
  if (m_seed.size() > 1) {
    for (size_t i=0; i<m_seed.size(); i++) {
//MJP TODO: this doesn't get the value of the strength between glues!
      if (GetIncidentGlueStrengthSum(m_seed[i].m_coords) < m_temp) {
        bTauStable = false;
        break;
      }
    }
  }
  if (!bTauStable) {
    wxString sMsg = _T("Warning:  this seed assembly is not tau-stable!\n");
    g_pTASApp->SetSimOutputText(sMsg, true, false);
  }

  return true;
}

void TASTileSpace::SetSeed()
{
  m_seed.clear();
  for (size_t i=0; i<GetNumTransitionsToTypeInEpoch(0, TAS_BEGIN_TILES); i++) {
    TASLocationTransition* pTrans = GetTransitionToTypeInEpoch(0, i, TAS_BEGIN_TILES);
    TASTileType* pTT = m_tileset[pTrans->m_typeEnd];
    if (pTT->m_sName.Length() > 0) {
      TASSeedEntry se(pTT->m_sName, pTrans->m_coords);
      m_seed.push_back(se);
    }
  }
}

void TASTileSpace::ClearSeed()
{
  m_seed.clear();
}

size_t TASTileSpace::GetNumTransitionsInEpoch(size_t epoch)
{
  if (epoch >= m_listEpochCache.size()) {
    return 0;
  } else {
    return m_listEpochCache[epoch].m_listTransitions.GetCount();
  }
}

TASLocationTransition* TASTileSpace::GetTransitionInEpoch(size_t epoch, size_t changeNum)
{
  size_t numTransitions = GetNumTransitionsInEpoch(epoch);
  if ((numTransitions == 0) || (changeNum >= numTransitions)) {
    return NULL;
  }

  return &m_listEpochCache[epoch].m_listTransitions[changeNum];
}

size_t TASTileSpace::GetNumTransitionsToTypeInEpoch(size_t epoch, int endType)
{
  if (epoch >= m_listEpochCache.size()) {
    return 0;
  } else if (endType == TAS_BEGIN_EMPTY) {
    return (m_listEpochCache[epoch].m_listTransitions.GetCount() - m_listEpochCache[epoch].m_arrBases[endType]);
  } else {
    return (m_listEpochCache[epoch].m_arrBases[endType+1] - m_listEpochCache[epoch].m_arrBases[endType]);
  }
}

TASLocationTransition* TASTileSpace::GetTransitionToTypeInEpoch(size_t epoch, size_t changeNum, int endType)
{
  size_t numTransitions = GetNumTransitionsToTypeInEpoch(epoch, endType);
  if ((numTransitions == 0) || (changeNum >= numTransitions)) {
    return NULL;
  }

  return &m_listEpochCache[epoch].m_listTransitions[m_listEpochCache[epoch].m_arrBases[endType]+changeNum];
}

bool TASTileSpace::AddNeighborsToFrontierList(const TASCoords& coords)
{
  TASCoords coordsNbr;

  for (int i=TAS_NORTH; i<=m_numNbrs; i++) {
    //  Check for existence of neighbor in location list or frontier list, and add to frontier if necessary
    coordsNbr = m_nbrOffsets[i] + coords;
    size_t locType = m_pLocationIndex->GetLocation(Index(coordsNbr));
    if ((locType == TAS_TILE_NOT_FOUND) || (locType == TAS_TILE_FRONTIER)) {
        if (GetFittingTileTypes(coordsNbr)) {   //  There is at least one tile type that can attach here, add to frontier

        TASFrontierLocation tfl(coordsNbr);
        m_setFrontier.insert(tfl);

        TASLocationTransition trans(coordsNbr, locType, TAS_TILE_FRONTIER);
        m_listEpochCache[m_epoch].AddTransition(trans);

        //  Make sure to recalculate the index from coordsNbr below, because the call to GetIncidentGlueStrengthSum
        //  above can trigger a growth of the space and therefore reindexing and a tricky bug!!!
        m_pLocationIndex->SetLocation(Index(coordsNbr), TAS_TILE_FRONTIER, m_epoch);
      }
    }
  }

  return true;
}

bool TASTileSpace::FixFrontierInNeighborhood(const TASCoords& coords)
{
  size_t num = (m_dimNum == 2) ? 5 : 7;
  for (size_t i(0); i < num; ++i) {
    TASCoords coordsCurr(coordsNbrhd[i] + coords);
    size_t ind = Index(coordsCurr);
    size_t ttn = m_pLocationIndex->GetLocation(ind);
    if ((ttn == TAS_TILE_NOT_FOUND) || (ttn == TAS_TILE_FRONTIER) || m_pLocationIndex->GetEpoch(i) > m_epoch) {
      if (GetIncidentGlueStrengthSum(coordsCurr) >= m_temp) {
        if (ttn != TAS_TILE_FRONTIER) {           //  This location should be added to the frontier
          TASFrontierLocation tfl(coordsCurr);
          m_setFrontier.insert(tfl);

          TASLocationTransition trans(coordsCurr, ttn, TAS_TILE_FRONTIER);
          m_listEpochCache[m_epoch].AddTransition(trans);

          //  Make sure to recalculate the index from coordsNbr below, because the call to GetIncidentGlueStrengthSum
          //  above can trigger a growth of the space and therefore reindexing and a tricky bug!!!
          m_pLocationIndex->SetLocation(Index(coordsCurr), TAS_TILE_FRONTIER, m_epoch);
        }
      } else if (ttn == TAS_TILE_FRONTIER) {      //  This location should be removed from the frontier

        TASFrontierLocation tfl(coordsCurr);
        TASFrontierSet::iterator iter = m_setFrontier.find(tfl);
        if (iter != m_setFrontier.end()) {
          m_setFrontier.erase(iter);
        }

        TASLocationTransition trans(coordsCurr, ttn, TAS_TILE_NOT_FOUND);
        m_listEpochCache[m_epoch].AddTransition(trans);

        m_pLocationIndex->SetLocation(Index(coordsCurr), TAS_TILE_NOT_FOUND, m_epoch);
      }
    }
  }
  return true;
}

void TASTileSpace::CalculateTotalTileConcentrations()
{
  m_iTotalConc = 0;
  for (size_t i=0; i<m_tileset.size(); i++) {
    m_iTotalConc += m_tileset.at(i)->m_concentration;
  }
}

void TASTileSpace::CheckForLocalDeterminism(TASCoords coords, TASTileTypePtrVect* pTTVect)
{
  if (pTTVect->size() > 1) {                    //  Check for situation where multiple tile types
                                                //  could have bound here
    wxString sNames;
    for (size_t n=0; n<pTTVect->size(); n++) {
      sNames += _T("'") + pTTVect->at(n)->m_sName + _T("'");
      if (n < (pTTVect->size()-1)) sNames += _T(", ");
    }
    wxString sMsg = wxString::Format(_T("Violation of local determinism:  %d possible tiletypes fit into location %s: %s\n"),
                                     pTTVect->size(), ((TASCoords)coords).ToString().c_str(), sNames.c_str());
    g_pTASApp->SetSimOutputText(sMsg, true);
  }

  std::vector<TASTileBinding> vectBindingSides;
  GetTileBindings(coords, &vectBindingSides);

  int bindingStrength = 0;
  for (size_t b=0; b<vectBindingSides.size(); b++) {
    if (vectBindingSides[b].m_type == TAS_TILE_INPUT_BINDING) {       //  Check for binding with too much strength
      bindingStrength += vectBindingSides[b].m_strength;
      if (bindingStrength > m_temp) {
        wxString sMsg = wxString::Format(_T("Violation of local determinism:  Tile type '%s' bound to location %s with strength at least %d (which is greater than tau = %d)\n"),
                                         GetTileType(coords)->m_sName.c_str(), ((TASCoords)coords).ToString().c_str(), bindingStrength, m_temp);
        g_pTASApp->SetSimOutputText(sMsg, true);
      }
    } else if (pTTVect->at(0)->m_sides[b].m_iStrength > 0) {          //  Check for case where a neighboring location which
                                                                      //  wasn't an input could now have multiple valid binding
                                                                      //  tile types

      //  Check this location by 'punching out' its output tiles and seeing if it could be refilled by more than one tile type 
      TASCoords coordsNbr = m_nbrOffsets[b] + coords;

      std::vector<TASTileBinding> vectNbrBindingSides;
      GetTileBindings(coordsNbr, &vectNbrBindingSides);

      TASTileTypePtrVect* pTTVectOuts = GetFittingTileTypesRestricted(coordsNbr, &vectNbrBindingSides);
      if (pTTVectOuts && pTTVectOuts->size() > 1) {
        wxString sNames;
        for (size_t n=0; n<pTTVectOuts->size(); n++) {
          sNames += _T("'") + pTTVectOuts->at(n)->m_sName + _T("'");
          if (n < (pTTVectOuts->size()-1)) sNames += _T(", ");
        }
        wxString sMsg = wxString::Format(_T("Violation of local determinism:  %d possible tiletypes fit into location %s: %s\n"),
                                         pTTVectOuts->size(), ((TASCoords)coordsNbr).ToString().c_str(), sNames.c_str());
        g_pTASApp->SetSimOutputText(sMsg, true);
      }
    }
  }
}

bool TASTileSpace::AddTileToLocation(TASCoords coords, int typeInd)
{
  TASTileTypePtrVect* pTTVect = GetFittingTileTypes(coords);
  if (pTTVect) {
    size_t sizeList = pTTVect->size();
    if (sizeList == 1) {                             //  There is exactly one fitting tile type
      bool bRet = AddTile(coords, pTTVect->at(0));
      if (bRet && g_pContext->m_bCheckForLocalDeterminism) {
        CheckForLocalDeterminism(coords, pTTVect);
      }
      return bRet;
    } else if (sizeList > 1) {                       //  There is nondeterminism here

      if (typeInd != -1) {
        //  Use the tile type in the specified index
        bool bRet = AddTile(coords, pTTVect->at(typeInd));
        if (bRet && g_pContext->m_bCheckForLocalDeterminism) {
          CheckForLocalDeterminism(coords, pTTVect);
        }
        return bRet;
      } else if (g_pContext->m_iTASTileConcMode == TAS_TILECONC_MODE_ON) {       //  Utilize tile concentrations

        //  Find the sum of the concentrations of the fitting tile types
        int concSum = 0;
        for (size_t i=0; i<pTTVect->size(); i++) {
          concSum += pTTVect->at(i)->m_concentration;
        }

        //  Generate a random number between 0 and concSum
        int randVal = g_pTASApp->RandInRange(concSum);

        //  Choose the appropriate tile type based on randVal
        int currSum = 0;
        for (size_t j=0; j<pTTVect->size(); j++) {
          currSum += pTTVect->at(j)->m_concentration;
          if (randVal < currSum) {
            bool bRet = AddTile(coords, pTTVect->at(j));
            if (bRet && g_pContext->m_bCheckForLocalDeterminism) {
              CheckForLocalDeterminism(coords, pTTVect);
            }
            return bRet;
          }
        }
      } else {
        //  Randomly pick one of the fitting tiles
        int index = g_pTASApp->RandInRange(sizeList);
        bool bRet = AddTile(coords, pTTVect->at(index));
        if (bRet && g_pContext->m_bCheckForLocalDeterminism) {
          CheckForLocalDeterminism(coords, pTTVect);
        }
        return bRet;
      }
    }
  }
  return false;
}


bool TASTileSpace::TileCanBind(TASTileType* pTtype, const TASCoords& coords)
{
  int strength = 0;

  for (int d2=TAS_NORTH; d2<=m_numNbrs; d2++) {
    TASCoords coordsNbr = m_nbrOffsets[d2] + coords;
    TASTileType* nbrTType = GetTileType(coordsNbr);

    if (nbrTType) {
      strength += m_tileset.GetGlueInteractionStrength(pTtype->m_sides[d2], nbrTType->m_sides[TASOppositeDirection[d2]]);
      if (!m_tileset.m_bUsesNegativeGlues && (strength >= m_temp)) {
        return true;
      }
    }
  }
  if (strength >= m_temp) {
    return true;
  }
  return false;
}

TASTileTypePtrVect* TASTileTypeSet::GetTileTypeListFromHash(TASTileSideHash& tsh, int temp)
{
  TASTileSideHashMap::iterator it = m_mapTileSides.find(tsh);
  if (it != m_mapTileSides.end()) {             //  There is an entry for tile side set, so it's been looked up before
    if (it->second.size() > 0) {                    //  There is at least one fitting tile type
      return &it->second;
    } else {                                        //  There are no fitting tile types
      return NULL;
    }
  } else {                                      //  There is no entry for this tile side set, so do a full search for matches
    TASTileTypePtrVect vectPTT;
    for (size_t i=0; i<size(); i++) {
      int strength = 0;
      for (int d2=TAS_NORTH; d2<=TAS_DOWN; d2++) {
        strength += GetGlueInteractionStrength(at(i)->m_sides[d2], tsh.m_arr[d2]);
        if ((strength >= temp) && !m_bUsesNegativeGlues) {
          break;
        }
      }
      if (strength >= temp) {
        //  A hack to handle staged assembly efficiently
        if (!g_pTileSpace->m_bUseStages ||
          (g_pTileSpace->m_vectStages[g_pTileSpace->m_currStage].find(at(i)->m_sName) != g_pTileSpace->m_vectStages[g_pTileSpace->m_currStage].end())) {
//MJP TODO:  check against current stage before adding.  Also, ensure that all entries are blown away when stages are loaded and between stages
          vectPTT.push_back(at(i));
        }
      }
    }

    m_mapTileSides[tsh] = vectPTT;                  //  Add an entry so this won't be calculated again
    if (vectPTT.size() > 0) {
      return &m_mapTileSides[tsh];
    } else {
      return NULL;
    }
  }
}

TASTileTypePtrVect* TASTileSpace::GetFittingTileTypes(const TASCoords& coords)
{
  TASCoords       coordsNbr;
  TASTileSideHash tsh;
  TASTileSide     tsEmpty;

  for (int d=TAS_NORTH; d<=m_numNbrs; d++) {
    coordsNbr = m_nbrOffsets[d] + coords;
    TASTileType* pTTNbr = GetTileType(coordsNbr);
    if (pTTNbr) {
      tsh.m_arr[d] = pTTNbr->m_sides[TASOppositeDirection[d]];
    } else {
      tsh.m_arr[d] = tsEmpty;
    }
  }
  tsh.CreateHash();
  return m_tileset.GetTileTypeListFromHash(tsh, m_temp);
}

TASTileTypePtrVect* TASTileSpace::GetFittingTileTypesRestricted(const TASCoords& coords, std::vector<TASTileBinding>* pVect)
{
  TASCoords       coordsNbr;
  TASTileSideHash tsh;
  TASTileSide     tsEmpty;

  for (int d=TAS_NORTH; d<=m_numNbrs; d++) {
    if (pVect->at(d).m_type != TAS_TILE_OUTPUT_BINDING) {
      coordsNbr = m_nbrOffsets[d] + coords;
      TASTileType* pTTNbr = GetTileType(coordsNbr);
      if (pTTNbr) {
        tsh.m_arr[d] = pTTNbr->m_sides[TASOppositeDirection[d]];
      } else {
        tsh.m_arr[d] = tsEmpty;
      }
    } else {
      tsh.m_arr[d] = tsEmpty;
    }
  }
  tsh.CreateHash();
  return m_tileset.GetTileTypeListFromHash(tsh, m_temp);
}

int TASTileSpace::GetTileBindings(const TASCoords& coords, std::vector<TASTileBinding>* pVect)
{
  int min = 0;
  bool bInit = false;
  size_t epochAdded = GetEpochAdded(coords);

  TASTileBinding tb;
  pVect->resize(m_numNbrs+1, tb);
  TASTileType* pTtype = GetTileType(coords);
  if (pTtype) {
    for (int d2=TAS_NORTH; d2<=m_numNbrs; d2++) {
      TASCoords coordsNbr = m_nbrOffsets[d2] + coords;
      TASTileType* nbrTType = GetTileType(coordsNbr);
      if (nbrTType) {
        int str = m_tileset.GetGlueInteractionStrength(pTtype->m_sides[d2], nbrTType->m_sides[TASOppositeDirection[d2]]);
        pVect->at(d2).m_strength = str;
        if (str > 0) {
          if (epochAdded > GetEpochAdded(coordsNbr)) {
            pVect->at(d2).m_type = TAS_TILE_INPUT_BINDING;
          } else {
            pVect->at(d2).m_type = TAS_TILE_OUTPUT_BINDING;
          }
        }
        if (!bInit && (str < min)) {
          bInit = true;
          min = str;
        }
      }
    }
  }
  return min;
}

bool TASTileSpace::IsInSeed(const TASCoords& c) {
  for (size_t s=0; s<m_seed.size(); s++) {
    if (m_seed[s].m_coords == c) {
      return true;
    }
  }
  return false;
}

size_t TASTileSpace::AttemptMelting(TileLocationIndex* pPossibleLocsIn)
{
  size_t numMeltedOff = 0;
  bool bDidMelt = true;

  TileLocationIndex* pPossibleLocs = pPossibleLocsIn;
  TileLocationIndex newPossibleLocs;

  while (bDidMelt) {
    size_t currNumMeltedOff = 0;
    bDidMelt = false;
    newPossibleLocs.Clear();
    std::vector<TASCoords> vectCoordsToMelt;
//MJP TODO: using a TileLocationIndex object is not very efficient.  Perhaps this could/should be optimized.
    for (size_t i=0; i<pPossibleLocs->GetSize(); i++) {
      size_t typeNum = pPossibleLocs->GetLocation(i);
      if (TASIsTileType(typeNum)) {
        TASCoords c(CoordsFromIndex(i));
        if (!TileCanBind(m_tileset[typeNum], c)) {

          //  Don't melt off seed tiles
          if (!IsInSeed(c)) {
            vectCoordsToMelt.push_back(c);
            bDidMelt = true;
          }
        }
      }
    }
    if (vectCoordsToMelt.size() > 0) {
      bDidMelt = true;
      for (size_t j=0; j<vectCoordsToMelt.size(); j++) {
        RemoveTile(vectCoordsToMelt[j]);
        currNumMeltedOff++;

        //  Add neighbors of melted locations to newPossibleLocs
        for (int dir=TAS_NORTH; dir<=m_numNbrs; dir++) {
          TASCoords coordsNbr = m_nbrOffsets[dir] + vectCoordsToMelt[j];
          TASTileType* nbrTType = GetTileType(coordsNbr);
          if (nbrTType) {
            newPossibleLocs.SetLocation(Index(coordsNbr), m_tileset.IndexOf(nbrTType->m_sName), m_epoch);
          }
        }
      }
      numMeltedOff  += currNumMeltedOff;
      m_numTiles    -= currNumMeltedOff;

      wxString sMsg = wxString::Format(_T("In step %d, %d tiles melted off the assembly\n"), m_epoch, currNumMeltedOff);
      g_pTASApp->SetSimDebugText(sMsg, true, false);
    }
    pPossibleLocs = &newPossibleLocs;
  }
  return numMeltedOff;
}

bool TASTileSpace::SetTemp(int temp)
{
  if (temp > 0) {

    int origTemp  = m_temp;
    bool bTempUp  = (temp > m_temp);

    m_temp = temp;

    //  Need to clear the hash mapping fitting tile types to neighborhoods because it is based on temperature
    m_tileset.m_mapTileSides.clear();

    //  Always want to increase the epoch number by at least one for a temperature change,
    //  even if there were no changes to the assembly (as long as we're not in epoch 0)
    if (m_epoch > 0) {
      AddEpoch();
    }

    wxString sMsg = wxString::Format(_T("Temperature value (tau) set to %d\n"), m_temp);
    g_pTASApp->SetSimOutputText(sMsg, true, false);

    if (bTempUp) {                    //  Melt off tiles as necessary
      size_t numMeltedOff = AttemptMelting(m_pLocationIndex);
      wxString sMsg = wxString::Format(_T("The temperature change from %d to %d caused %d tiles to melt off of the assembly\n"), origTemp, m_temp, numMeltedOff);
      g_pTASApp->SetSimOutputText(sMsg, true, false);
    } else {    //  The temperature went down, so clear and rebuild the frontier
      RebuildFrontier();
    }

    return true;
  }
  return false;
}

void TASTileSpace::RebuildFrontier()
{
  for (TASFrontierSet::iterator it = m_setFrontier.begin(); it != m_setFrontier.end(); it++) {
    m_pLocationIndex->SetLocation(Index(it->m_coords), TAS_TILE_NOT_FOUND, m_epoch);
  }
  m_setFrontier.clear();

  size_t t = 0;
  TASLocationTransition* pTrans = NULL;
  for (size_t e=0; e<=(size_t)m_epoch; e++) {
    t = 0;
	pTrans = GetTransitionToTypeInEpoch(e, t++, TAS_BEGIN_TILES);
    while (pTrans != NULL) {
      AddNeighborsToFrontierList(pTrans->m_coords);
	  pTrans = GetTransitionToTypeInEpoch(e, t++, TAS_BEGIN_TILES);
    }
  }
}

bool TASTileSpace::IncrementStage()
{
  if (m_currStage < (m_vectStages.size()-1)) {
    m_currStage++;
    wxString sStageChange = wxString::Format(_T("Advancing to stage %d\n"), m_currStage);
    g_pTASApp->SetSimOutputText(sStageChange, true, true);

    //  Need to clear the hash mapping fitting tile types to neighborhoods because
    //  the available tile types have changed
    m_tileset.m_mapTileSides.clear();

    RebuildFrontier();

    return true;
  }
  return false;
}

bool TASTileSpace::UseTimingModelToAddTilesToFrontier()
{
  return AddTileToFrontierHelper(TAS_TILE_TIMING_MODEL);
}

bool TASTileSpace::AddMaxTilesToFrontier()
{
  return AddTileToFrontierHelper(TAS_TILE_ADD_FULL_FRONT);
}

bool TASTileSpace::AddTileToFrontier()
{
  return AddTileToFrontierHelper(TAS_TILE_ADD_ONE);
}

bool TASTileSpace::KillFrontierLocation(TASFrontierLocation* pFrontLoc)
{
  bool bRet = false;

  if (!pFrontLoc) {
    return false;
  }

  TASCoords coords = pFrontLoc->m_coords;

  if (m_setFrontier.RemoveElement(coords)) {
 
    bRet = true;

    size_t indDead = Index(coords);
    m_pLocationIndex->SetLocation(indDead, TAS_TILE_NOT_FOUND, m_epoch);

    if (m_listEpochCache.size() <= (size_t)m_epoch) {
      AddEpoch();
    }

    TASLocationTransition trans(coords, TAS_TILE_FRONTIER, TAS_TILE_NOT_FOUND);
    m_listEpochCache[m_epoch].AddTransition(trans);

    wxString sDeadLoc = wxString::Format(_T("ERROR: Location (%d,%d) being removed from frontier - no fitting tile type (this location shouldn't have been added to the frontier)\n"), coords.x, coords.y);
    g_pTASApp->SetSimDebugText(sDeadLoc, true, false);
  }

  return bRet;
}

bool TASTileSpace::AddTileToFrontierHelper(int numTilesMode)
{
  bool bAdded = false;

  //  For now, the size of the frontier space is limited to RAND_MAX (~= 2^15).
  //  This may need to be increased by generating a random number over a larger range if it
  //  is not sufficient.
  if (m_setFrontier.size() >= RAND_MAX) {
    wxMessageBox(_T("The frontier space is too large to be processed!"), _T("Frontier space exceeded limit"));
  } else if (m_setFrontier.size() == 0) {   //  There are no more frontier locations
    bool b;
    std::vector<wxString> vectName;
    vectName.push_back(_T("termination_criteria"));
    vectName.push_back(_T("terminal_assembly"));
    if (g_pContext->m_bTASBatchMode && g_pContext->m_settingsBatchMode.GetSetting(&vectName, &b)) {
      wxString sRunDone = wxString::Format(_T("Batch run number %d complete.\n"), ++g_pContext->m_lCurrNumRuns);
      g_pTASApp->SetSimOutputText(sRunDone, true);
      g_pContext->m_bDoReset = true;
    } else {
      wxString sEmptyFrontier = wxString::Format(_T("No more tiles can be added:  the frontier is empty!\n"));
      g_pTASApp->SetSimOutputText(sEmptyFrontier, true, true);
    }
  } else {
    if (numTilesMode == TAS_TILE_ADD_ONE) {
      //  Randomly select a location from the frontier list
      int randInd = g_pTASApp->RandInRange(m_setFrontier.size());
      randInd--;  //  This is simply because the first line of the loop will increment it

      bool bContinue = true;
      while (bContinue) {
        randInd = (randInd + 1) % m_setFrontier.size();

        TASFrontierLocation* frontLoc = m_setFrontier.GetElement(randInd);// *it;//*(it + randInd%m_setFrontier.size());// m_listFrontier.at((randInd+i)%m_listFrontier.size());
        if (AddTileToLocation(frontLoc->m_coords)) {  //  Increment iter by one, signifying a successful tile addition
          bAdded = true;
          break;
        } else {
          KillFrontierLocation(frontLoc);
        }
        bContinue = (!bAdded && (m_setFrontier.size() > 0));
      }

    } else if (numTilesMode == TAS_TILE_ADD_FULL_FRONT) {
      TASFrontierSet frontierOrig = m_setFrontier;
//MJP: should this start with a random frontier location??
      for (TASFrontierSet::iterator it = frontierOrig.begin(); it != frontierOrig.end(); it++) {
        if (AddTileToLocation(it->m_coords)) {
          bAdded = true;
        } else {
          KillFrontierLocation((TASFrontierLocation*)&*it);
        }
      }
//MJP: what we need to do is, for each frontier location compute the concentration c of the fitting tile type,
//     pick a random number r between 0 and 1, and if r <= c add a tile of that type  
    } else if (numTilesMode == TAS_TILE_TIMING_MODEL) {

      if (m_iTotalConc == 0) {
        CalculateTotalTileConcentrations();
      }

      int iNumAdded = 0;
      TASFrontierSet frontierOrig = m_setFrontier;
//MJP: should this start with a random frontier location??
//[MJP]  There is a potential bug due to the fact that the addition of one tile can potentially impact
//       the addition of further tiles during the same virtual time step by modifying the locations adjacent
//       to other frontier locations (even though both tile additions should be "simultaneous")
      for (TASFrontierSet::iterator it = frontierOrig.begin(); it != frontierOrig.end(); it++) {
        TASTileTypePtrVect* pTTVect = GetFittingTileTypes(it->m_coords);
        if (pTTVect) {
          int rand = g_pTASApp->RandInRange(m_iTotalConc);
          //  Figure out which of the tile types was selected
          int vectInd = 0;
          while ((vectInd < pTTVect->size()) && (rand >= pTTVect->at(vectInd)->m_concentration)) {
            rand = rand - pTTVect->at(vectInd)->m_concentration;
            vectInd++;
          }
          if (vectInd < pTTVect->size()) {
            if (AddTileToLocation(it->m_coords, vectInd)) {
              iNumAdded++;
              bAdded = true;
            } else {
              KillFrontierLocation((TASFrontierLocation*)&*it);
            }
          }
        }
      }
      wxString sNumAdded = wxString::Format(_T("On simulation step %d, added %d tiles with frontier size %d\n"), m_epoch, iNumAdded, frontierOrig.size());
      g_pTASApp->SetSimDebugText(sNumAdded, true, true);

/*
      TASFrontierSet frontierOrig = m_setFrontier;
      size_t frontierSize = frontierOrig.size();
      double midPoint = (double)frontierSize / 2.0;

      //RandomLib::Random r;
	  std::tr1::mt19937 engine(time(0));
      //RandomLib::NormalDistribution<double> normdist;
	  std::tr1::normal_distribution<double> distribution;
      //double dNumToAdd = normdist(r, midPoint, (midPoint/3));
	  typedef std::tr1::normal_distribution<double>::param_type MeanSigma;
	  double dNumToAdd = distribution(engine, MeanSigma(midPoint, (midPoint / 3.0)));
      int iNumToAdd = (int)std::ceil(dNumToAdd-0.5);

      while ((iNumToAdd < 0) || (iNumToAdd > frontierSize)) {
		  dNumToAdd = distribution(engine, MeanSigma(midPoint, (midPoint / 3.0)));
          iNumToAdd = (int)std::ceil(dNumToAdd-0.5);
      }
      std::set<int> done;

      wxString sNumAdding = wxString::Format(_T("On simulation step %d, adding %d tiles with frontier size %d\n"), m_epoch, iNumToAdd, frontierOrig.size());
      g_pTASApp->SetSimDebugText(sNumAdding, true, true);

      for (int i=0; i<iNumToAdd; i++) {
        int randInd = g_pTASApp->RandInRange(frontierOrig.size());
        if (done.size() == frontierOrig.size()) {
          break;                                                          //  ERROR!!!
        }
        while (done.find(randInd) != done.end()) {
          randInd = g_pTASApp->RandInRange(frontierOrig.size());
        }
        if (AddTileToLocation(frontierOrig.GetElement(randInd)->m_coords)) {
          bAdded = true;
        } else {
          KillFrontierLocation(frontierOrig.GetElement(randInd));
        }
      }*/
      if (iNumAdded == 0) {
        bAdded = true;  //  A pseudo-hack to claim we added a tile since we don't want to terminate this run
                        //  because of an unlucky random number generation, since we should only be here if
                        //  the frontier still has locations
      }
    }

    if (!bAdded) {
      size_t sizeFrontierTotal = m_setFrontier.size();
      wxString sNoAdd = wxString::Format(_T("Stopping execution -> Unable to add tiles to any of the %d frontier locations\n"), sizeFrontierTotal);
      g_pTASApp->SetSimOutputText(sNoAdd, true, true);

      TASFrontierLocation tfl;
      for (TASFrontierSet::iterator it = m_setFrontier.begin(); it != m_setFrontier.end(); it++) {
        sNoAdd = wxString::Format(_T("  No fitting tile type for frontier location (%d,%d,%d)\n"), it->m_coords.x, it->m_coords.y, it->m_coords.z);
        g_pTASApp->SetSimDebugText(sNoAdd, true);
      }
    }
  }

  //  Check if negative glues are being used and, if so, attempt to melt neighbors of added tiles
  //  (expanding the melting neighborhood as necessary
  if (m_tileset.m_bUsesNegativeGlues) {
    //  Create a list of locations with tiles which neighbor all locations where tiles were added during this epoch
    TileLocationIndex locsAttemptToMelt;
    size_t numAdded = GetNumTransitionsToTypeInEpoch(m_epoch, TAS_BEGIN_TILES);
    for (size_t na = 0; na < numAdded; na++) {
      TASLocationTransition* pLocTrans = GetTransitionToTypeInEpoch(m_epoch, na, TAS_BEGIN_TILES);

      for (int dir=TAS_NORTH; dir<=m_numNbrs; dir++) {
        TASCoords coordsNbr = m_nbrOffsets[dir] + pLocTrans->m_coords;
        TASTileType* nbrTType = GetTileType(coordsNbr);
        if (nbrTType) {
//MJP TODO: this could be optimized by only looking next to locations with negative glue interactions
          locsAttemptToMelt.SetLocation(Index(coordsNbr), m_tileset.IndexOf(nbrTType->m_sName), m_epoch);
        }
      }
    }
    //  Attempt to melt the tiles in each of those locations
    AttemptMelting(&locsAttemptToMelt);
  }

  if (m_tileset.m_bUsesNegativeGlues && (m_cutSets.size() > 0)) {
//MJP TODO: probably should randomly select one of the cuts, but since there already isn't a random
//          distribution of cuts calculated, skipping that for now.
    AddEpoch(true);
    for (std::set<TASCoords>::iterator it = m_cutSets[0].m_locs.begin(); it != m_cutSets[0].m_locs.end(); it++) {
      RemoveTile(*it);
    }
  }

  //  Check if temperature programming is being used and if the temperature now needs to be adjusted
  if (!bAdded && (m_vectTemps.size() > 0) && (m_currTemp < (m_vectTemps.size()-1))) {
    wxString sTempChange = wxString::Format(_T("Switching to next temperature value\n"));
    g_pTASApp->SetSimOutputText(sTempChange, true, true);
    m_currTemp++;
    m_epoch--;  //  Reset the epoch count since it will be incremented by SetTemp
    SetTemp(m_vectTemps[m_currTemp]);
    bAdded = true;
  }

  //  Check if staged assembly is being used and if the stage needs to be incremented
  if (!bAdded && m_bUseStages && (m_currStage < (m_vectStages.size()-1))) {
    //m_epoch--;
    AddEpoch();
    IncrementStage();
    bAdded = true;
  }

  return bAdded;
}

int TASTileSpace::GetIncidentGlueStrengthSum(const TASCoords& coords)
{
  int retSum = 0;
  TASTileType* pTT = NULL;
  
  //  Check for existence of north neighbor tile and get incident glue strength
  pTT = GetTileType(TASCoords(coords.x, coords.y + 1, coords.z));
  if (pTT != NULL) {
    retSum += m_tileset.GetPossibleGlueInteractionStrength(pTT->m_sides[TAS_SOUTH]);
  }
  //  Check for existence of south neighbor tile and get incident glue strength
  pTT = GetTileType(TASCoords(coords.x, coords.y - 1, coords.z));
  if (pTT != NULL) {
    retSum += m_tileset.GetPossibleGlueInteractionStrength(pTT->m_sides[TAS_NORTH]);
  }
  //  Check for existence of east neighbor tile and get incident glue strength
  pTT = GetTileType(TASCoords(coords.x + 1, coords.y, coords.z));
  if (pTT != NULL) {
    retSum += m_tileset.GetPossibleGlueInteractionStrength(pTT->m_sides[TAS_WEST]);
  }
  //  Check for existence of west neighbor tile and get incident glue strength
  pTT = GetTileType(TASCoords(coords.x - 1, coords.y,  coords.z));
  if (pTT != NULL) {
    retSum += m_tileset.GetPossibleGlueInteractionStrength(pTT->m_sides[TAS_EAST]);
  }
  if (m_dimNum > 2) {
    //  Check for existence of up neighbor tile and get incident glue strength
    pTT = GetTileType(TASCoords(coords.x, coords.y, coords.z + 1));
    if (pTT != NULL) {
      retSum += m_tileset.GetPossibleGlueInteractionStrength(pTT->m_sides[TAS_DOWN]);
    }
    //  Check for existence of down neighbor tile and get incident glue strength
    pTT = GetTileType(TASCoords(coords.x, coords.y, coords.z - 1));
    if (pTT != NULL) {
      retSum += m_tileset.GetPossibleGlueInteractionStrength(pTT->m_sides[TAS_UP]);
    }
  }

  return retSum;
}

TileLocationIndex::TileLocationIndex()
{
  m_growthFactor = 50000; //MJP: tune this
  m_pLocs = new std::vector<TileLocationIndexItem>;
  m_pLocs->resize(m_growthFactor, TileLocationIndexItem(TAS_TILE_NOT_FOUND, 0));
}

TileLocationIndex::~TileLocationIndex()
{
  if (m_pLocs) {
    delete m_pLocs;
  }
}

void TileLocationIndex::Clear()
{
  if (m_pLocs) {
    delete m_pLocs;
  }
  m_pLocs = new std::vector<TileLocationIndexItem>;
  m_pLocs->resize(m_growthFactor, TileLocationIndexItem(TAS_TILE_NOT_FOUND, 0));
}

//MJP: can't just blindly resize!  Will need to transfer values...
size_t TileLocationIndex::GetLocation(size_t ind)
{
  if ((m_pLocs->size() <= ind) || (ind == TAS_TILE_NOT_FOUND)) {
    return TAS_TILE_NOT_FOUND;
  } else {
    return m_pLocs->at(ind).m_tiletype;
  }
}

size_t TileLocationIndex::GetEpoch(size_t ind)
{
  if ((m_pLocs->size() <= ind) || (ind == TAS_TILE_NOT_FOUND)) {
    return TAS_TILE_NOT_FOUND;
  } else {
    return m_pLocs->at(ind).m_epoch;
  }
}

void TileLocationIndex::SetLocation(size_t ind, size_t tileTypeNum, size_t epoch)
{
  while (m_pLocs->size() <= ind) {
    m_pLocs->resize(m_pLocs->size() + m_growthFactor, TileLocationIndexItem(TAS_TILE_NOT_FOUND, 0));
  }
  m_pLocs->at(ind).m_tiletype = tileTypeNum;
  m_pLocs->at(ind).m_epoch    = epoch;
}

///////////////////////
//// MJP: can the EpochListList actually just be a list of tileTypeNums???
////    no, (maybe) because then you can't rewind?
////      - or maybe make rewinding update
////          - no jumping around (should be okay)
////          - what about frontier?
///////////////////////

TASTileSpace::TASTileSpace(int dimNum) : m_cutSets(this)
{
  ConstructorHelper(dimNum);
  ResetSpace();
}

TASTileSpace::~TASTileSpace()
{
  delete m_pLocationIndex;
}

size_t TASTileSpace::GetMaxEpoch()
{
  return m_listEpochCache.size();
}

void TASTileSpace::SetDims(const TASCoords& coordsMin, const TASCoords& coordsMax)
{
  m_coordsMin = coordsMin;
  m_coordsMax = coordsMax;

  m_coordsDims = TASCoords((m_coordsMax.x - m_coordsMin.x + 1),
                           (m_coordsMax.y - m_coordsMin.y + 1),
                           (m_coordsMax.z - m_coordsMin.z + 1));

  wxString dbgMsg = wxString::Format(_T("Setting tile space size to %s\n"), m_coordsDims.ToString().c_str());
  g_pTASApp->SetSimDebugText(dbgMsg, true);

  m_epoch = 0;

  delete m_pLocationIndex;
  m_pLocationIndex = new TileLocationIndex();

  m_listEpochCache.clear();
  m_setFrontier.clear();
}

void TASTileSpace::GrowSpace(const TASCoords& coords)
{
  TASCoords coordsDimsOld(m_coordsDims);
  TASCoords coordsMinsOld(m_coordsMin);
  TASCoords coordsMaxesOld(m_coordsMax);

  if (g_pContext->m_iTASDimMode == TAS_2D_MODE) {
    m_dimGrowthRate = TAS_SPACE_GROWTHINC_2D;
  } else {
    m_dimGrowthRate = TAS_SPACE_GROWTHINC_3D;
  }

  if (coords.x < m_coordsMin.x) {
    m_coordsMin.x = coords.x - m_dimGrowthRate;
  }
  if (coords.x > m_coordsMax.x) {
    m_coordsMax.x = coords.x + m_dimGrowthRate;
  }
  if (coords.y < m_coordsMin.y) {
    m_coordsMin.y = coords.y - m_dimGrowthRate;
  }
  if (coords.y > m_coordsMax.y) {
    m_coordsMax.y = coords.y + m_dimGrowthRate;
  }
  if (g_pContext->m_iTASDimMode == TAS_3D_MODE) {
    if (coords.z < m_coordsMin.z) {
      m_coordsMin.z = coords.z - m_dimGrowthRate;
    }
    if (coords.z > m_coordsMax.z) {
      m_coordsMax.z = coords.z + m_dimGrowthRate;
    }
  }

  if ((coordsMinsOld == m_coordsMin) && (coordsMaxesOld == m_coordsMax)) {
    return;
  }

  m_coordsDims = TASCoords((m_coordsMax.x - m_coordsMin.x + 1),
                             (m_coordsMax.y - m_coordsMin.y + 1),
                             (m_coordsMax.z - m_coordsMin.z + 1));

  wxString dbgMsg = wxString::Format(_T("Growing space %s -> %s for coords %s\n"), coordsDimsOld.ToString().c_str(), m_coordsDims.ToString().c_str(), ((TASCoords)coords).ToString().c_str());
  g_pTASApp->SetSimDebugText(dbgMsg, true);

  TileLocationIndex* pNewLocationIndex = new TileLocationIndex();
  for (size_t i=0; i<m_pLocationIndex->GetSize(); i++) {
    size_t loc = m_pLocationIndex->GetLocation(i);
    if (loc != TAS_TILE_NOT_FOUND) {
      TASCoords tc = CoordsFromIndexHelper(i, coordsDimsOld, coordsMinsOld);
      size_t indNew = Index(tc);
      size_t ep  = m_pLocationIndex->GetEpoch(i);
      pNewLocationIndex->SetLocation(indNew, loc, ep);
    }
  }
  delete m_pLocationIndex;
  m_pLocationIndex = pNewLocationIndex;
}

size_t TASTileSpace::Index(const TASCoords& coords)
{
  while (coords.x < m_coordsMin.x || coords.y < m_coordsMin.y || coords.z < m_coordsMin.z ||
         coords.x > m_coordsMax.x || coords.y > m_coordsMax.y || coords.z > m_coordsMax.z)
  {
    GrowSpace(coords);
  }
  
  if (m_dimNum < 3) {
    return (m_coordsDims.x * (coords.y - m_coordsMin.y)) + (coords.x - m_coordsMin.x);
  } else {
    return (m_coordsDims.y * m_coordsDims.x * (coords.z - m_coordsMin.z)) +
           (m_coordsDims.x * (coords.y - m_coordsMin.y)) +
           (coords.x - m_coordsMin.x);
  }
}

TASCoords TASTileSpace::CoordsFromIndex(size_t ind)
{
  return CoordsFromIndexHelper(ind, m_coordsDims, m_coordsMin);
}

TASCoords TASTileSpace::CoordsFromIndexHelper(size_t ind, const TASCoords& dims, const TASCoords& mins)
{
  TASCoords retCoords;

  if (m_dimNum < 3) {
    retCoords.x = (ind % dims.x) + mins.x;
    retCoords.y = ((ind - retCoords.x) / dims.x) + mins.y;
    retCoords.z = 0;
  } else {
    retCoords.x = (ind % dims.x) + mins.x;
    ind = (ind - retCoords.x) / dims.x;
    retCoords.y = (ind % dims.y) + mins.y;
    retCoords.z = ((ind - retCoords.y) / dims.y) + mins.z;
  }

  return retCoords;
}

void TASTileSpace::TruncateEpochCache(size_t maxEpoch)
{
  while (m_listEpochCache.size() > (maxEpoch+1)) {
    m_listEpochCache.pop_back();
  }
}

TASTileType* TASTileSpace::SafelyGetTileType(const TASCoords& coords)
{
  wxMutexLocker lock(*g_pMutexTileSpaceGuard);
  return GetTileType(coords);
}

TASTileType* TASTileSpace::GetTileType(const TASCoords& coords)
{
  if (!((TASCoords)coords).IsValid()) {
    return NULL;
  }

  size_t i = Index(coords);
  size_t ttn = m_pLocationIndex->GetLocation(i);
  if ((ttn == TAS_TILE_NOT_FOUND)     ||
      (ttn == TAS_TILE_FRONTIER)
      /*(m_pLocationIndex->GetEpoch(i) > epoch)*/)
  {
    return NULL;
  } else {
    return m_tileset[ttn];
  }
}

bool TASTileSpace::IsFrontierLocation(const TASCoords& coords)
{
  if (!((TASCoords)coords).IsValid()) {
    return false;
  }

  size_t i = Index(coords);
  size_t ttn = m_pLocationIndex->GetLocation(i);
  if ((ttn == TAS_TILE_FRONTIER) /* && (m_pLocationIndex->GetEpoch(i) <= epoch)*/) {
    return true;
  } else {
    return false;
  }
}

size_t TASTileSpace::GetEpochAdded(const TASCoords& coords)
{
  size_t i = Index(coords);
  return m_pLocationIndex->GetEpoch(i);
}

bool TASTileSpace::AddTile(const TASCoords& coords, const TASTileType* pTT)
{
  size_t indexLoc = Index(coords);
  size_t locType = m_pLocationIndex->GetLocation(indexLoc);

  if ((locType != TAS_TILE_NOT_FOUND) && (locType != TAS_TILE_FRONTIER)) {  //  A tile already exists here
    return false;
  }

  size_t ind = m_tileset.IndexOf(pTT->m_sName);
  if (ind == TAS_TILE_NOT_FOUND) {
    return false;
  }

  if (m_listEpochCache.size() <= (size_t)m_epoch) {
    AddEpoch();
  }

  TASLocationTransition trans(coords, locType, ind);
  m_listEpochCache[m_epoch].AddTransition(trans);

  m_pLocationIndex->SetLocation(indexLoc, ind, m_epoch);

  if (locType == TAS_TILE_FRONTIER) {
    m_setFrontier.RemoveElement(coords);
  }

  AddNeighborsToFrontierList(coords);

//MJP: figure out how to make sure this is efficiently utilized
//  std::sort(m_listFrontier.begin(), m_listFrontier.end());

  TruncateEpochCache(m_epoch);

  if (m_tileset.m_bUsesNegativeGlues) {
    m_cutSets.CheckCuts(coords);
  }

  return true;
}

bool TASTileSpace::RemoveTile(const TASCoords& coords)
{
  TASTileType* pTT = GetTileType(coords);
  if (!pTT) {      //  This location is already empty
    return true;
  }

  if ((size_t)m_epoch >= m_listEpochCache.size()) {
    AddEpoch();
  }

  size_t ind = m_tileset.IndexOf(pTT->m_sName);
  if (ind == TAS_TILE_NOT_FOUND) {
    return false;
  }

  TASLocationTransition trans(coords, ind, TAS_TILE_NOT_FOUND);
  m_listEpochCache[m_epoch].AddTransition(trans);

  m_pLocationIndex->SetLocation(Index(coords), TAS_TILE_NOT_FOUND, m_epoch);

  FixFrontierInNeighborhood(coords);

  TruncateEpochCache(m_epoch);
  return true;
}

bool TASTileSpace::SetTileTypes(wxString& sTileTypeDefns)
{
  m_tileset = sTileTypeDefns;
  ResetSpace();
  return true;
}

bool TASTileSpace::SetTileTypes(TASTileTypeSet& setIn)
{
  m_tileset = setIn;
  ResetSpace();
/*
  for (size_t i=0; i<m_listEpochCache.size(); i++) {
    if (m_listEpochCache[i].m_listTilesChanged.size() > 0) {
      for (TASTileLocationList::iterator iter=m_listEpochCache[i].m_listTilesChanged.begin(); 
                                            iter!=m_listEpochCache[i].m_listTilesChanged.end(); iter++)
      {
        TASTileType ttOld = setTileTypesOld[iter->m_tileTypeNum];
        size_t ind = m_tileset.IndexOf(ttOld);
        if (ind != TAS_TILE_NOT_FOUND) {
         iter->m_tileTypeNum = ind;
        } else {
  //MJP: if a tiletype has been deleted, should I try to replace with an equivalent?
          m_listEpochCache[i].m_listTilesChanged.erase(iter);
        }
      }
    }
  }
*/
  return true;
}

void TASTileSpace::AddEpoch(bool bDoInc)
{
  if (bDoInc) {
    m_epoch++;
  }

  if(m_listEpochCache.size() == m_epoch) {
	m_listEpochCache.push_back(TASTileEpoch(m_temp));
  } else {
	// (Michael) We will only be here if for some reason we increment m_epoch
	// more than once before getting to this function

	// Interesting note m_listEpochCache gets truncated most times it gets
	// called if we increase the size larger than m_epoch it cuts it off
	// wasting lots of time
	m_listEpochCache.resize(m_epoch+1);
	m_listEpochCache[m_epoch] = TASTileEpoch(m_temp);
  }
}

bool TASTileSpace::IncrementEpoch()
{
  wxMutexLocker lock(*g_pMutexTileSpaceGuard);

  if (m_epoch == 0) {   //  If we're leaving epoch 0, save the current assembly as the seed
    SetSeed();
    if (m_bUseStages) { //  Increment out of stage 0 if we're using staged assembly
      IncrementStage();
    }
  }

  m_epoch++;
  m_bEpochCached = false;

  if (IsEpochCached(m_epoch)) {    //  Epoch was cached so now just fix the m_pLocationIndex and frontier sets
    m_bEpochCached = true;

    m_temp = m_listEpochCache[m_epoch].m_temp;

    for (size_t i=0; i<GetNumTransitionsInEpoch(m_epoch); i++) {
      TASLocationTransition* pTrans = GetTransitionInEpoch(m_epoch, i);
      if (pTrans) {
        size_t indexLoc = Index(pTrans->m_coords);

        m_pLocationIndex->SetLocation(indexLoc, pTrans->m_typeEnd, m_epoch);

        //  If this location ends in a frontier, add it to the frontier set
        if (pTrans->m_typeEnd == TAS_TILE_FRONTIER) {
          m_setFrontier.insert(TASFrontierLocation(pTrans->m_coords));
        }

        //  If this location started the transition in a frontier, remove it
        if (pTrans->m_typeBegin == TAS_TILE_FRONTIER) {
          m_setFrontier.RemoveElement(pTrans->m_coords);
        }
      }
    }
  } else if (g_pContext->m_iTileAdditionsPerStep == TAS_TILE_ADD_FULL_FRONT) {
    if (!AddMaxTilesToFrontier()) {
      if (GetNumTransitionsInEpoch(m_epoch) == 0) {
        m_epoch--;
      }
      return false;
    }
  } else if (g_pContext->m_iTileAdditionsPerStep == TAS_TILE_TIMING_MODEL) {
    if (!UseTimingModelToAddTilesToFrontier()) {
      if (GetNumTransitionsInEpoch(m_epoch) == 0) {
        m_epoch--;
      }
      return false;
    }
  } else if (!AddTileToFrontier()) {
    if (GetNumTransitionsInEpoch(m_epoch) == 0) {
      m_epoch--;
    }
    return false;
  }

  for (size_t i=0; i<GetNumTransitionsInEpoch(m_epoch); i++) {
    TASLocationTransition* pTrans = GetTransitionInEpoch(m_epoch, i);
    if (TASIsTileType(pTrans->m_typeEnd) && !TASIsTileType(pTrans->m_typeBegin)) {
      m_numTiles++;
    } else if (TASIsTileType(pTrans->m_typeBegin) && !TASIsTileType(pTrans->m_typeEnd)) {
      m_numTiles--;
    }
  }

  return true;
}

bool TASTileSpace::DecrementEpoch()
{
  wxMutexLocker lock(*g_pMutexTileSpaceGuard);

  if (m_epoch > 0) {
    m_bEpochCached = true;

    m_temp = m_listEpochCache[m_epoch].m_temp;

    for (size_t i=0; i<GetNumTransitionsInEpoch(m_epoch); i++) {
      TASLocationTransition* pTrans = GetTransitionInEpoch(m_epoch, i);
      if (pTrans) {
        size_t indexLoc = Index(pTrans->m_coords);

        m_pLocationIndex->SetLocation(indexLoc, pTrans->m_typeBegin, m_epoch);

        //  If this location began in a frontier, add it to the frontier set
        if (pTrans->m_typeBegin == TAS_TILE_FRONTIER) {
          m_setFrontier.insert(TASFrontierLocation(pTrans->m_coords));
        }

        //  If this location ended the transition in a frontier, remove it
        if (pTrans->m_typeEnd == TAS_TILE_FRONTIER) {
          m_setFrontier.RemoveElement(pTrans->m_coords);
        }

        if (TASIsTileType(pTrans->m_typeEnd) && !TASIsTileType(pTrans->m_typeBegin)) {
          m_numTiles--;
        } else if (TASIsTileType(pTrans->m_typeBegin) && !TASIsTileType(pTrans->m_typeEnd)) {
          m_numTiles++;
        }
      }
    }

    m_epoch--;

    return true;
  } else {
    return false;
  }
}

int TASTileSpace::GetUnusedTileTypes(int epoch, TASTileTypeSet* pUnusedTypes)
{
  TASTileTypeSet usedTypes;

  for (size_t e=0; e<=(size_t)epoch; e++) {
    for (size_t i=0; i<GetNumTransitionsToTypeInEpoch(e, TAS_BEGIN_TILES); i++) {
      //  This will automatically fail if the tiletype already exists in the list
      TASLocationTransition* pTrans = GetTransitionToTypeInEpoch(e, i, TAS_BEGIN_TILES);
      usedTypes.push_back(*m_tileset[pTrans->m_typeEnd], false);
    }
    //MJP: not worrying about removed tiles because they were at least used at some point
  }

  for (size_t i=0; i<m_tileset.size(); i++) {
    if (usedTypes.IndexOf(m_tileset[i]->m_sName) == TAS_TILE_NOT_FOUND) {
      pUnusedTypes->push_back(*m_tileset[i], false);
    }
  }

  usedTypes.clear();
  return pUnusedTypes->size();
}

void TASTileSpace::OutputChangesPerEpoch()
{
  for (size_t i(0); i <= m_epoch; ++i) {
    OutputChangesInEpoch(i);
  }
}

void TASTileSpace::OutputChangesInEpoch(int epoch)
{
  wxString sMsg = wxString::Format(_T("== Epoch %d ==\nTemperature = %d\n"), epoch, m_listEpochCache[epoch].m_temp);
  g_pTASApp->SetSimDebugText(sMsg, true, false);

  size_t numTrans = GetNumTransitionsInEpoch(epoch);
  for (size_t i=0; i<numTrans; i++) {
    TASLocationTransition* pTrans = GetTransitionInEpoch(epoch, i);
    if (GetDimNum() == 2) {
      sMsg = wxString::Format(_T("(%d,%d)"), pTrans->m_coords.x, pTrans->m_coords.y);
    } else {
      sMsg = wxString::Format(_T("(%d,%d,%d)"), pTrans->m_coords.x, pTrans->m_coords.y, pTrans->m_coords.z);
    }
    sMsg += _T(" changed from ");
    if (TASIsTileType(pTrans->m_typeBegin)) {
      sMsg += wxString::Format(_T("tile type '%s'"), m_tileset.at(pTrans->m_typeBegin)->m_sName.c_str());
    } else if (pTrans->m_typeBegin == TAS_TILE_FRONTIER) {
      sMsg += _T("a frontier location");
    } else if (pTrans->m_typeBegin == TAS_TILE_NOT_FOUND) {
      sMsg += _T("an empty location");
    }
    sMsg += _T(" to ");
    if (TASIsTileType(pTrans->m_typeEnd)) {
      sMsg += wxString::Format(_T("tile type '%s'"), m_tileset.at(pTrans->m_typeEnd)->m_sName.c_str());
    } else if (pTrans->m_typeEnd == TAS_TILE_FRONTIER) {
      sMsg += _T("a frontier location");
    } else if (pTrans->m_typeEnd == TAS_TILE_NOT_FOUND) {
      sMsg += _T("an empty location");
    }
    sMsg += _T("\n");

    g_pTASApp->SetSimDebugText(sMsg, true, false);
  }
}

TASTileTypeVect::~TASTileTypeVect()
{
  clear();
}

void TASTileTypeVect::push_back(const TASTileType* pTT)
{
  std::vector<TASTileType*>::push_back(new TASTileType(*pTT));
}

void TASTileTypeVect::insert(TASTileTypeVect::iterator iter, const TASTileType* pTT)
{
  TASTileType* pNewTT = new TASTileType(*pTT);
  std::vector<TASTileType*>::insert(iter, pNewTT);
}

void TASTileTypeVect::erase(TASTileTypeVect::iterator iter)
{
  delete *iter;
  std::vector<TASTileType*>::erase(iter);
}

void TASTileTypeVect::clear()
{
  while (size() > 0) {
    delete at(size()-1);
    pop_back();
  }
}

void TASTileTypeSet::InvalidateBitmaps()
{
  for (size_t i=0; i<size(); i++) {
    at(i)->InvalidateBitmaps();
  }
}

bool TASTileTypeSet::SaveToFile(wxFFile* outFile)
{
  if (!outFile->IsOpened()) {
    return false;
  }

  TASTileTypeVect::iterator iter = m_vectTypes.begin();
  while (iter != m_vectTypes.end()) {
    outFile->Write((*iter)->Serialize());
    outFile->Write(wxString(_T("\n")));
    iter++;
  }

  outFile->Close();
  return true;
}

size_t TASTileTypeSet::LoadGlueFunctionDefinitions(wxString& sTTDefns, size_t index)
{
  size_t origIndex        = index;
  size_t len              = sTTDefns.Len() - index;
  size_t newIndex         = index;
  bool bHadFunctionDefns  = false;
  bool bInFunctionDefn    = false;

  m_mapGlueFunctions.clear();

  while ((newIndex < len) && (sTTDefns[newIndex] != '\n')) {
    newIndex++;
  }
  wxString sLine = sTTDefns.SubString(index, newIndex);
  newIndex++;
  index = newIndex;
  wxString sGlueName;
  TASGlueFunction glueFunction;
  while (sLine.length() > 0) {
    sLine.Trim().Trim(false);

    //  Parse current line and get value
    if (sLine.StartsWith(_T("TILENAME"))) {   //  If we find a tile type defintion, we're done
      if (bHadFunctionDefns) {
        wxString sMsg = wxString::Format(_T("Warning while loading glue function definitions.  Missing 'END GLUEFUNCTIONS' line.\n"));
        g_pTASApp->SetEditorOutputText(sMsg, true, true);
        return index;
      } else {
        return origIndex;
      }
    } else if (sLine.StartsWith(_T("BEGIN GLUEFUNCTIONS"))) {
      bHadFunctionDefns = true;
    } else if (sLine.StartsWith(_T("END GLUEFUNCTIONS"))) {
      if (!bHadFunctionDefns) {
          wxString sMsg = wxString::Format(_T("Error loading glue function definitions.  Missing 'BEGIN GLUEFUNCTIONS' line.\n"));
          g_pTASApp->SetEditorOutputText(sMsg, true, true);
      }
      return index;
    } else if (sLine.StartsWith(_T("["))) {
      bInFunctionDefn = true;
      glueFunction.clear();                 // Reset the strength function object
      glueFunction.SetDefaultStrength(0);
      sGlueName = sLine.SubString(1, sLine.Len()).Trim().Trim(false);
    } else if (sLine.StartsWith(_T("]"))) {
      bInFunctionDefn = false;
      m_mapGlueFunctions[sGlueName] = glueFunction;
    } else if (bInFunctionDefn) {
      int colonIndex = sLine.Find(':');
      if (colonIndex == wxNOT_FOUND) {                                  //  Most likely an error
        
      } else {
        wxString sStrength = sLine.SubString(0, colonIndex-1);
        long lStrength = 0;
        if (!sStrength.IsNumber() || !sStrength.ToLong(&lStrength)) {    //  An error!
          wxString sMsg = wxString::Format(_T("Error loading glue function definition for glue '%s'.  Invalid line '%s'\n"), sGlueName.c_str(), sLine.c_str());
          g_pTASApp->SetEditorOutputText(sMsg, true, true);
          return index;
        } else {
          wxString sGlueList = sLine.SubString(colonIndex+1, sLine.Len());
          wxArrayString arrStrGlues = wxStringTokenize(sGlueList, _T(" "));
          for (size_t i=0; i<arrStrGlues.GetCount(); i++) {
            wxString sCurrGlue = arrStrGlues[i];
            if (sCurrGlue.IsSameAs(TAS_DEFAULT_GLUE)) {
              glueFunction.SetDefaultStrength(lStrength);
            } else {
              glueFunction[sCurrGlue] = lStrength;
            }
          }
          if (lStrength < 0) {
            m_bUsesNegativeGlues = true;
          }
        }
      }
    }

    while ((newIndex < len) && (sTTDefns[newIndex] != '\n')) {
      newIndex++;
    }
    sLine = sTTDefns.SubString(index, newIndex);
    newIndex++;
    index = newIndex;
  }

  if (!bHadFunctionDefns) {
    index = origIndex;
  }
  return index;
}

TASTileTypeSet& TASTileTypeSet::operator=(const wxString& sTTDefns)
{
  clear();
  size_t index  = 0;
  m_bUsesNegativeGlues = false;   //MJP TODO: we may need to set this in places other than LoadGlueFunctionDefinitions

  // Load possible glue function definitions
  index = LoadGlueFunctionDefinitions((wxString&)sTTDefns, index);
  if (!m_mapGlueFunctions.empty()) {
    m_bUsesGlueFunctionMap = true;
  } else {
    m_bUsesGlueFunctionMap = false;
  }

  size_t len    = sTTDefns.Len();
  while (index < len) {
    TASTileType tt(sTTDefns, &index);
    if (tt.IsValid()) {
      push_back(tt, false);
    }
  }
  return *this;
}

TASTileTypeSetEx& TASTileTypeSetEx::operator=(const wxString& sTTDefns)
{
  TASTileTypeSet::operator =(sTTDefns);
  return *this;
}

TASTileTypeSet& TASTileTypeSet::operator=(TASTileTypeSet& tts)
{
  if (tts.m_bUsesGlueFunctionMap) {
    m_mapGlueFunctions = tts.m_mapGlueFunctions;
    m_bUsesGlueFunctionMap = true;
  } else {
    m_bUsesGlueFunctionMap = false;
  }
  m_bUsesNegativeGlues = tts.m_bUsesNegativeGlues;
  clear();
  for (size_t i=0; i<tts.size(); i++) {
    push_back(*tts[i], false);
  }
  return *this;
}

TASTileType* TASTileTypeSet::operator[](const size_t ind)
{
  return at(ind);
}

TASTileType* TASTileTypeSet::at(const size_t ind)
{
  if (ind >= m_vectTypes.size()) {
    return g_pEmptyTileType;
  }
  return m_vectTypes[ind];
}

TASTileType* TASTileTypeSet::operator[](const wxString& sName)
{
  return at(sName);
}

TASTileType* TASTileTypeSet::at(const wxString& sName)
{
  size_t ind = IndexOf(sName);
  if (ind == TAS_TILE_NOT_FOUND) {
    return g_pEmptyTileType;
  } else {
    return m_vectTypes[ind];
  }
}

bool TASTileTypeSet::push_back(const TASTileType& tt, bool bReIndex)
{
  if (IndexOf(tt.m_sName) != TAS_TILE_NOT_FOUND) {
    wxString sMsg = wxString::Format(_T("Error loading tile type named '%s' - a tile type with that name already exists!\n"), tt.m_sName.c_str());
    g_pTASApp->SetEditorOutputText(sMsg, true, true);

    return false;
  }

  m_vectTypes.push_back(&tt);
  m_mapNameIndex[tt.m_sName] = m_vectTypes.size() - 1;
  if (bReIndex) {
    ReIndex();
  } else {
    IndexTile(m_vectTypes[m_vectTypes.size()-1], m_vectTypes.size() - 1);
  }

  return true;
}

bool TASTileTypeSet::insert(const TASTileType& tt, size_t ind)
{
  if ((IndexOf(tt.m_sName) != TAS_TILE_NOT_FOUND) || (ind < 0) || (ind > size())) {
    wxString sMsg = wxString::Format(_T("Error inserting tile type named '%s' - a tile type with that name already exists or insertion is out of range!\n"), tt.m_sName.c_str());
    g_pTASApp->SetEditorOutputText(sMsg, true, true);

    return false;
  }

  size_t count = 0;
  TASTileTypeVect::iterator iter = m_vectTypes.begin();
  while ((count < ind) && iter != m_vectTypes.end()) {
    count++;
    iter++;
  }
  m_vectTypes.insert(iter, (TASTileType*)&tt);
  IndexTile(m_vectTypes[ind], ind);

  return true;
}

bool TASTileTypeSet::erase(size_t ind)
{
  if ((ind < 0) || (ind >= size())) {
    return false;
  }

  size_t count = 0;
  TASTileTypeVect::iterator iter = m_vectTypes.begin();
  while ((count < ind) && iter != m_vectTypes.end()) {
    count++;
    iter++;
  }
  m_vectTypes.erase(iter);
  ReIndex();

  return true;
}

void TASTileTypeSet::clear()
{
  m_vectTypes.clear();
  m_mapNameIndex.clear();
  m_mapGlueIndex.clear();
  m_mapTileSides.clear();
  ReIndex();
}

size_t TASTileTypeSet::size()
{
  return m_vectTypes.size();
}

size_t TASTileTypeSet::IndexOf(const TASTileType& tt)
{
  return IndexOf(tt.m_sName);
}

size_t TASTileTypeSet::IndexOf(const wxString& sName)
{
  std::map<wxString, size_t>::iterator iter = m_mapNameIndex.find(sName);
  if (iter == m_mapNameIndex.end()) {
    return TAS_TILE_NOT_FOUND;
  } else {
    return iter->second;
  }
}

void TASTileTypeSet::IndexTile(TASTileType* pTT, size_t num)
{
  if (IndexOf(pTT->m_sName) == TAS_TILE_NOT_FOUND) {
    m_mapNameIndex[pTT->m_sName] = num;
  }

  int numNbrs = (g_pContext->m_iTASDimMode == TAS_2D_MODE) ? TAS_WEST : TAS_DOWN;
  for (int j=TAS_NORTH; j<=numNbrs; j++) {
    wxString sGlueHash = wxString::Format(_T("%d:%s"), pTT->m_sides[j].m_iStrength, pTT->m_sides[j].m_sLabel.c_str());
    std::map<wxString, size_t>::iterator iter = m_mapGlueIndex.find(sGlueHash);
    if (iter == m_mapGlueIndex.end()) {
      m_mapGlueIndex[sGlueHash] = (num*6) + j;
    }
    pTT->m_sides[j].m_hashNum = m_mapGlueIndex[sGlueHash];
    std::map<size_t, TASTileTypePtrVect>::iterator iter2 = m_mmapGlues[j].find(m_mapGlueIndex[sGlueHash]);
    if (iter2 == m_mmapGlues[j].end()) {
      TASTileTypePtrVect v;
      v.push_back(pTT);
      m_mmapGlues[j][m_mapGlueIndex[sGlueHash]] = v;
    } else {
      m_mmapGlues[j][m_mapGlueIndex[sGlueHash]].push_back(pTT);
    }
  }
}

void TASTileTypeSetEx::IndexTile(TASTileType* pTT, size_t num)
{
  TASTileTypeSet::IndexTile(pTT, num);

  m_sArrNames.push_back(pTT->m_sName);

  if (m_sMapLabels.find(pTT->m_sLabel) != m_sMapLabels.end()) {
    m_sMapLabels[pTT->m_sLabel] = m_sMapLabels[pTT->m_sLabel] + 1;
  } else {
    m_sMapLabels[pTT->m_sLabel] = 1;
    m_sArrLabels.push_back(pTT->m_sLabel);
  }

  int numNbrs = (g_pContext->m_iTASDimMode == TAS_2D_MODE) ? TAS_WEST : TAS_DOWN;
  for (int i=TAS_NORTH; i<=numNbrs; i++) {
    if (m_sMapGlues.find(pTT->m_sides[i].m_sLabel) != m_sMapGlues.end()) {
      m_sMapGlues[pTT->m_sides[i].m_sLabel] = m_sMapGlues[pTT->m_sides[i].m_sLabel] + 1;
    } else {
      m_sMapGlues[pTT->m_sides[i].m_sLabel] = 1;
      m_sArrGlues.push_back(pTT->m_sides[i].m_sLabel);
    }
  }
}

void TASTileTypeSet::ReIndex()
{
  m_mapNameIndex.clear();
  m_mapGlueIndex.clear();
  m_mapTileSides.clear();
  for (int k=TAS_NORTH; k<TAS_DOWN; k++) {
    m_mmapGlues[k].clear();
  }

  for (size_t i=0; i<m_vectTypes.size(); i++) {
    IndexTile(m_vectTypes[i], i);
  }
}

void TASTileTypeSetEx::ReIndex()
{
  m_sArrNames.Clear();
  m_sArrLabels.Clear();
  m_sArrGlues.Clear();
  m_sMapLabels.clear();
  m_sMapGlues.clear();

  TASTileTypeSet::ReIndex();
}

TASTileType* TASTileTypeSet::GetEquivalentTileType(TASTileType* pTtype)
{
  int numNbrs = (g_pContext->m_iTASDimMode == TAS_2D_MODE) ? TAS_WEST : TAS_DOWN;
  for (int d=TAS_NORTH; d<=numNbrs; d++) {
    if (pTtype->m_sides[d].m_iStrength > 0) {
      std::vector<TASTileType*>* pVect = &m_mmapGlues[d][pTtype->m_sides[d].m_hashNum];
      for (size_t i=0; i<pVect->size(); i++) {
        if ((pVect->at(i)->m_sName != pTtype->m_sName) && (pVect->at(i)->IsEqual(pTtype, false))) {
          return pVect->at(i);
        }
      }
      return NULL;
    }
  }

  //  Should only get here for tiles with 0 strength glues on all sides
  for (size_t i=0; i<m_vectTypes.size(); i++) {
    if ((m_vectTypes[i]->m_sName != pTtype->m_sName) && (m_vectTypes[i]->IsEqual(pTtype, false))) {
      return m_vectTypes[i];
    }
  }
  return NULL;
}

void TASTileSpace::ConstructorHelper(int dimNum)
{
  m_epoch             = 0;
  m_numTiles          = 0;
  m_dimNum            = dimNum;
  m_numNbrs           = (m_dimNum == 2) ? TAS_WEST : TAS_DOWN;
  m_dimGrowthRate     = 0;//TAS_SPACE_GROWTHINC_2D;
  m_pLocationIndex    = NULL;
  m_temp              = 2;
  m_bUseStages        = false;
  m_iTotalConc        = 0;

  m_nbrOffsets[0]     = TASCoords( 0,  1,  0);
  m_nbrOffsets[1]     = TASCoords( 1,  0,  0);
  m_nbrOffsets[2]     = TASCoords( 0, -1,  0);
  m_nbrOffsets[3]     = TASCoords(-1,  0,  0);
  m_nbrOffsets[4]     = TASCoords( 0,  0,  1);
  m_nbrOffsets[5]     = TASCoords( 0,  0, -1);

  SetRandSeed(0, true);
}

TAS2HAMTileSpace::TAS2HAMTileSpace(int dimNum)
{
  ConstructorHelper(dimNum);

  //m_numRows             = 0;
  //m_numCols             = 0;
  //m_maxSupertile        = 0;
  //m_numSupertiles       = 0;
  m_numTilesInLargestST = 0;
  m_defaultPadding      = 2;
  m_bIncludeSingletons  = true;
  m_epochLastStats      = (size_t)-1;
  m_minVisSize          = 0;
  m_maxVisSize          = 0;
  m_minKeepSize         = 0;
  m_maxKeepSize         = 0;
  m_bOnlyLargestVis     = false;
  m_bOnlyLargestKeep    = false;

  //m_coordsMin   = TASCoords(-m_defaultPadding, -m_defaultPadding, 0);
  //m_coordsMax   = TASCoords(m_defaultPadding, m_defaultPadding, 0);
  //m_coordsDims  = TASCoords(2*m_defaultPadding, 2*m_defaultPadding, 1);

  CreateWorkerThreads();
  ResetSpace();
}

TAS2HAMTileSpace::TAS2HAMTileSpace(TASTileSpace* pTileSpace)
{
  ConstructorHelper(pTileSpace->GetDimNum());

  m_tileset         = pTileSpace->m_tileset;
  //TASFrontierSet        m_setFrontier;          //  Set of current frontier locations

  m_bEpochCached    = false;
  //TASSeed               m_seed;
  m_vectTemps       = pTileSpace->m_vectTemps;
  m_currTemp        = 0;
  m_bUseStages      = pTileSpace->m_bUseStages;
  //std::vector<std::set<wxString>> m_vectStages; //  If staged assembly is being used, this will be list of valid tile
                                                //  types for each stage
  m_currStage       = pTileSpace->m_currStage;
  //TileLocationIndex*    m_pLocationIndex;       //  Effectively an n-dimensional array to directly index all current space locations
  //TASCutSets            m_cutSets;              //  Set of cuts across the assembly that are below tau and can be discarded
  m_numNbrs         = pTileSpace->GetNumNbrs();
  m_temp            = pTileSpace->GetTemp();
  m_dimGrowthRate   = 2*m_defaultPadding;
  //TASTileEpochList        m_listEpochCache;

  m_dimNum              = pTileSpace->GetDimNum();

  m_numRows             = 0;
  m_numCols             = 0;
  m_maxSupertile        = 0;
  m_numSupertiles       = 0;
  m_numTilesInLargestST = 0;
  m_defaultPadding      = 2;
  m_epochLastStats      = (size_t)-1;

  m_bIncludeSingletons  = true;

  m_coordsMin           = TASCoords(-m_defaultPadding, -m_defaultPadding, 0);
  m_coordsMax           = TASCoords(m_defaultPadding, m_defaultPadding, 0);
  m_coordsDims          = TASCoords(2*m_defaultPadding, 2*m_defaultPadding, 1);

  m_minVisSize          = 0;
  m_maxVisSize          = 0;
  m_minKeepSize         = 0;
  m_maxKeepSize         = 0;
  m_bOnlyLargestVis     = false;
  m_bOnlyLargestKeep    = false;
  
  CreateWorkerThreads();
}

TAS2HAMTileSpace::~TAS2HAMTileSpace()
{
  DeleteWorkerThreads();

  if (m_pLocationIndex) {
    delete m_pLocationIndex;
  }

  for (size_t i=0; i<m_vectSTS.size(); i++) {
    delete m_vectSTS[i];
  }
}

bool TAS2HAMTileSpace::CreateWorkerThreads()
{
  wxThread::SetConcurrency(g_pContext->m_iNum2HAMThreads + 2);  //  Not sure what is optimal here

  m_currThreadMin = 0;
  m_currThreadMax = 0;

  for (int i=0; i<g_pContext->m_iNum2HAMThreads; i++) {
    TAS2HAMThread* pThread = new TAS2HAMThread(wxTHREAD_DETACHED, this, &m_semCounterBlock, &m_semMainBlock, &m_semFreeCount);
    if (pThread->Create() == wxTHREAD_NO_ERROR) {
      pThread->Run();
      m_vectWorkerThreads.push_back(pThread);
    } else {
      wxMessageBox(_T("ERROR:  Failed to create thread for simulation engine.  Exiting."), _T("Error Creating Thread"), wxICON_ERROR|wxOK);
      exit(1);
    }
  }

  return true;
}

bool TAS2HAMTileSpace::DeleteWorkerThreads()
{
  for (size_t i=0; i<m_vectWorkerThreads.size(); i++) {
    m_vectWorkerThreads[i]->SetShouldDie();
  }

  for (size_t i=0; i<m_vectWorkerThreads.size(); i++) {
    m_semCounterBlock.Post();
  }

  m_vectWorkerThreads.clear();

//  for (size_t i=0; i<m_vectWorkerThreads.size(); i++) {
//    if (m_vectWorkerThreads[i]->Delete() != /*(wxThread::ExitCode)-1*/wxTHREAD_NO_ERROR) {
//MJP: ERROR case
//bool bError = true;

//    }
//    delete m_vectWorkerThreads[i];
//  }

  return true;  // TODO: make this dependent on success
}

void TAS2HAMTileSpace::ResetSpace(bool WXUNUSED(bSaveSeed))
{
  for (size_t i=0; i<m_vectSTS.size(); i++) {
    delete m_vectSTS[i];
  }
  m_vectSTS.clear();
  m_setST.clear();

  m_numRows             = 0;
  m_numCols             = 0;
  m_maxSupertile        = 0;
  m_numSupertiles       = 0;
  m_numTilesInLargestST = 0;
  m_currThreadMin       = 0;
  m_currThreadMax       = 0;
  m_bEpochCached        = false;

  m_list2HAMEpochCache.clear();
  TAS2HAMEpochCacheEntry cacheEntry;

  if (m_bIncludeSingletons) {
    for (size_t i=0; i < m_tileset.size(); i++) {
      TASSupertile* pST = new TASSupertile(m_tileset[i], &m_tileset);
      size_t dim = (pST->m_maxDim > m_maxSupertile) ? pST->m_maxDim : m_maxSupertile;
      TASSupertileSpace* pSTS = new TASSupertileSpace(pST, dim + m_defaultPadding);
      AddSupertileSpace(pSTS, 0);
      cacheEntry.m_numTilesInLargestST = 1;
      cacheEntry.m_maxSupertile = 1;
    }
  }
  m_list2HAMEpochCache.push_back(cacheEntry);
  CalculateStats(true);
}

void TAS2HAMTileSpace::AddSupertileSpace(TASSupertileSpace* pSTS, size_t epochAdded)
{
  wxMutexLocker lock(*g_pMutexTileSpaceGuard);

  m_vectSTS.push_back(pSTS);
  m_numSupertiles = m_vectSTS.size();
  pSTS->m_pST->m_supertileNum = m_numSupertiles - 1;
  pSTS->m_pST->m_epochAdded = epochAdded;

  if (pSTS->m_pST->m_maxDim > m_maxSupertile) {
    m_maxSupertile = pSTS->m_pST->m_maxDim;
  }
  if (pSTS->m_pST->m_numTiles > m_numTilesInLargestST) {
    m_numTilesInLargestST = pSTS->m_pST->m_numTiles;
  }
  m_numTiles = m_vectSTS.size();
  m_setST.insert(pSTS->m_pST->m_sUID);
//  CalculateStats();  overkill!
}

void TAS2HAMTileSpace::RemoveSupertileSpace(TASSupertileSpace* pSTS)
{
  wxMutexLocker lock(*g_pMutexTileSpaceGuard);
//MJP TODO: should clean up memory!!

  if ((pSTS->m_pST->m_count == 1) || (pSTS->m_pST->m_count == TAS_2HAM_INFINITE_COUNT)) {
    std::set<wxString>::iterator it2 = m_setST.find(pSTS->m_pST->m_sUID);
    if (it2 != m_setST.end()) {
      m_setST.erase(it2);
    }

    std::vector<TASSupertileSpace*>::iterator it = std::find(m_vectSTS.begin(), m_vectSTS.end(), pSTS);
    if (it != m_vectSTS.end()) {
      m_vectSTS.erase(it);
    }

    m_numTiles = m_vectSTS.size();
  } else {
//MJP: Note that this doesn't remove the st space but simply decrements the count...
    pSTS->m_pST->m_count--;
  }
}

TASSupertileSpace* TAS2HAMTileSpace::SafelyGetSupertileSpace(const TASCoords& coords)
{
  //wxLogDebug(wxString::Format(wxT("Trying to lock g_pMutexTileSpaceGuard in TAS2HAMTileSpace::SafelyGetSupertileSpace for thread %d"), wxThread::GetCurrentId()));
  wxMutexLocker lock(*g_pMutexTileSpaceGuard);
  //wxLogDebug(wxString::Format(wxT("Locked g_pMutexTileSpaceGuard in TAS2HAMTileSpace::SafelyGetSupertileSpace for thread %d"), wxThread::GetCurrentId()));
  
  //wxLogDebug(wxString::Format(wxT("Unlocking g_pMutexTileSpaceGuard in TAS2HAMTileSpace::SafelyGetSupertileSpace for thread %d"), wxThread::GetCurrentId()));
  return GetSupertileSpace(coords);
}

TASSupertileSpace* TAS2HAMTileSpace::GetSupertileSpace(const TASCoords& coords)
{
  //  Get supertileSpace row and column to figure out which supertileSpace
  size_t supertileColumn = coords.x / (m_maxVisSupertile + m_defaultPadding);
  size_t supertileRow    = coords.y / (m_maxVisSupertile + m_defaultPadding);
  if ((supertileColumn >= m_numCols) || (supertileRow >= m_numRows)) {
    return NULL;
  }

  size_t supertileNum = (m_numCols * supertileRow) + supertileColumn;
  if (supertileNum >= m_vectVisSTS.size()) {
    return NULL;
  }

  return m_vectVisSTS[supertileNum];
}

TASSupertile* TAS2HAMTileSpace::SafelyGetSupertile(const TASCoords& coords)
{
  wxMutexLocker lock(*g_pMutexTileSpaceGuard);
  return GetSupertile(coords);
}

TASSupertile* TAS2HAMTileSpace::GetSupertile(const TASCoords& coords)
{
  TASSupertileSpace* pSTS = GetSupertileSpace(coords);
  if (pSTS) {
    return pSTS->m_pST;
  }
  return NULL;
}

TASTileType* TAS2HAMTileSpace::SafelyGetTileType(const TASCoords& coords)
{
  wxMutexLocker lock(*g_pMutexTileSpaceGuard);
  return GetTileType(coords);
}

TASTileType* TAS2HAMTileSpace::GetTileType(const TASCoords& coords)
{
  TASSupertileSpace* pSTS = GetSupertileSpace(coords);
  if (pSTS) {
    TASCoords localCoords(coords.x % (m_maxVisSupertile + m_defaultPadding),
                          coords.y % (m_maxVisSupertile + m_defaultPadding), 1);
    return pSTS->GetTileType(localCoords);
  }
  return NULL;
}

size_t TAS2HAMTileSpace::SafelyGetNumTilesInLargestSupertile()
{
  wxMutexLocker lock(*g_pMutexTileSpaceGuard);
  return m_numTilesInLargestST;
}

bool TAS2HAMTileSpace::IsDuplicate(TASSupertile* pST)
{
  return (m_setST.find(pST->m_sUID) != m_setST.end());
/*
  for (size_t i=0; i<m_vectSTS.size(); i++) {
    if (m_vectSTS[i]->m_pST->IsEqual(pST)) {
      return true;
    }
  }
  return false;
*/
}

std::vector<int> TwoHAMHackSTDeleteList;

bool TAS2HAMTileSpace::IncrementEpoch()
{
  if (!IsEpochCached(m_epoch+1)) {  
  //wxString sFoo;
  //int counter = 0;
    for (size_t i(0); i < m_vectWorkerThreads.size(); ++i) {
      m_semFreeCount.Post();
    }

    int chunkSize = 10;
    bool bAddedNew = false;
    m_numSupertilesBefore = m_vectSTS.size();
    m_maxNewSupertile = m_maxSupertile;

    g_pTASApp->Set2HAMStageProgress(0);

  //sFoo += wxString::Format("Covering range 0->%d: ", m_numSupertilesBefore);

    if (g_pContext->m_bUse2HAMHack) {
      TwoHAMHackSTDeleteList.clear();
      TwoHAMHackSTDeleteList.assign(m_numSupertilesBefore, TAS_2HAMHACK_INIT);
    }

    for (size_t i=0; i<m_numSupertilesBefore; i++) {

      //  Break off a group of supertiles to be checked by an available thread
      m_currThreadMin = i;
      size_t currMax = ((i + chunkSize) > (m_numSupertilesBefore-1)) ? (m_numSupertilesBefore-1) : (i + chunkSize);
      m_currThreadMax = currMax;
      i = currMax;
      
  //sFoo += wxString::Format("[%d->%d] ", m_currThreadMin, m_currThreadMax);
      m_semFreeCount.Wait();
      m_semCounterBlock.Post();
      m_semMainBlock.Wait();

      g_pTASApp->Set2HAMStageProgress((i * 100) / m_numSupertilesBefore);
  //  Should check to see if we need to pause/stop and record where we are so we can continue
  //if (m_pSimFrame->GetSimDirection() == STOP)
    }

    //  Make sure that all threads have completed
    for (size_t i=0; i<m_vectWorkerThreads.size(); i++) {
      m_semFreeCount.Wait();
    }

    TAS2HAMEpochCacheEntry cacheEntry(m_maxSupertile, m_numTilesInLargestST);
    m_list2HAMEpochCache.push_back(cacheEntry);

    if (g_pContext->m_bUse2HAMHack) {
      //  Delete unwanted tiles if we're using the 2HAM hack
      size_t numDeleted = 0;
      for (size_t i=0; i<TwoHAMHackSTDeleteList.size(); i++) {
        if (TwoHAMHackSTDeleteList[i] == TAS_2HAMHACK_DELETE) {
          RemoveSupertileSpace(m_vectSTS[i - numDeleted]);
          numDeleted++;
        }
      }
      wxString sDeletingSupertiles = wxString::Format(wxT("Deleted %d supertiles [using 2HAM hack]\n"), numDeleted);
      g_pTASApp->SetSimOutputText(sDeletingSupertiles, true, true);

      //  Fix the minToCheck values now
      for (size_t y=0; y < m_vectSTS.size(); y++) {
        m_vectSTS[y]->m_pST->m_minToCheck = m_vectSTS[y]->m_pST->m_minToCheck - numDeleted;
      }
    }

    if ((g_pContext->m_refreshRate > 0) && ((m_epoch % g_pContext->m_refreshRate) == 0)) {
      CalculateStats(true, m_epoch+1);
    }

    g_pTASApp->Set2HAMStageProgress(-1);
    
    if (m_numSupertilesBefore == m_vectSTS.size()) {
      wxString sNoNewSupertiles = wxString::Format(_T("No new supertiles can be created.\n"));
      g_pTASApp->SetSimOutputText(sNoNewSupertiles, true, true);
    } else {
      int numAdded = m_vectSTS.size() - m_numSupertilesBefore;
      wxString sNoNewSupertiles = wxString::Format(_T("Added %d supertiles.  The largest supertile contains %d tiles.\n"), numAdded, m_numTilesInLargestST);
      g_pTASApp->SetSimOutputText(sNoNewSupertiles, true, true);
      m_epoch++;
      bAddedNew = true;
    }
    return bAddedNew;
  } else {

    m_maxSupertile = m_list2HAMEpochCache[m_epoch+1].m_maxSupertile;
    m_numTilesInLargestST = m_list2HAMEpochCache[m_epoch+1].m_numTilesInLargestST;

    if ((g_pContext->m_refreshRate > 0) && ((m_epoch % g_pContext->m_refreshRate) == 0)) {
      CalculateStats(true, m_epoch+1);
    }
    m_epoch++;

    return true;
  }

//sFoo += wxString::Format(", %d combinations were tried\n", counter);
//g_pTASApp->SetSimDebugText(sFoo, true, true);
}

bool TAS2HAMTileSpace::DecrementEpoch()
{
  m_bEpochCached = true;
  if (m_epoch > 0) {
    m_epoch--;

    m_maxSupertile = m_list2HAMEpochCache[m_epoch].m_maxSupertile;
    m_numTilesInLargestST = m_list2HAMEpochCache[m_epoch].m_numTilesInLargestST;

    if ((g_pContext->m_refreshRate > 0) && ((m_epoch % g_pContext->m_refreshRate) == 0)) {
      CalculateStats(true);
    }
    return true;
  } else {
    return false;
  }
}

void TAS2HAMTileSpace::CalculateStats(bool bForce, size_t epochIn)
{
  wxMutexLocker lock(*g_pMutexTileSpaceGuard);

  size_t epoch = m_epoch;
  if (epochIn != (size_t)-1) {
    epoch = epochIn;
  }

  if (bForce || (m_epochLastStats != epoch)) {
    m_vectStats.clear();
    m_vectStats.assign(m_numTilesInLargestST+1, 0);
    std::vector<std::vector<TASSupertileSpace*> > vectTemp;
    std::vector<TASSupertileSpace*> vsts;
    vectTemp.assign(m_numTilesInLargestST+1, vsts);
    for (size_t j=0; j<m_vectSTS.size(); j++) {
      if (m_vectSTS[j]->m_pST->m_epochAdded <= epoch) {
        m_vectStats[m_vectSTS[j]->m_pST->m_numTiles]++;
        vectTemp[m_vectSTS[j]->m_pST->m_numTiles].push_back(m_vectSTS[j]);
      }
    }

    //  Now set the visible collection of supertiles
    m_vectVisSTS.clear();
    if (m_bOnlyLargestVis) {
      for (size_t i=0; i<vectTemp[vectTemp.size()-1].size(); i++) {
        m_vectVisSTS.push_back(vectTemp[vectTemp.size()-1].at(i));
      }
    } else {
      size_t upperBound = (m_maxVisSize == 0) ? m_numTilesInLargestST : m_maxVisSize;
      for (size_t i=m_minVisSize; i<=upperBound; i++) {
        for (size_t j=0; j<vectTemp[i].size(); j++) {
          if (vectTemp[i].at(j)->m_pST->m_epochAdded <= epoch) {
            m_vectVisSTS.push_back(vectTemp[i].at(j));
          }
        }
      }
    }
    m_numVisSupertiles = m_vectVisSTS.size();
    if (m_numVisSupertiles == 0) {
      m_maxVisSupertile = 0;
    } else {
      m_maxVisSupertile = m_vectVisSTS[m_numVisSupertiles-1]->m_pST->m_numTiles;
    }

    m_epochLastStats = m_epoch;
  }

  m_numRows = ceil(sqrt((double)m_numVisSupertiles));
  m_numCols = m_numRows;// + (m_numVisSupertiles - (m_numRows * m_numRows));

  int maxX = (m_maxVisSupertile + m_defaultPadding) * m_numCols;
  int maxY = (m_maxVisSupertile + m_defaultPadding) * m_numRows;
  m_coordsMin = TASCoords(0, 0, 0);
  m_coordsMax = TASCoords(maxX, maxY, 1);

  m_coordsDims = TASCoords((m_coordsMax.x - m_coordsMin.x)+1, (m_coordsMax.y - m_coordsMin.y)+1, 1);
}

TASTileType* TASSupertileSpace::GetTileType(const TASCoords& coords)
{
  // (Michael) m_offsetX is unsigned while coords.x is signed...
  // does this affect behavior?
  if ((coords.x < m_offsetX) || (coords.y < m_offsetY)) {
    return NULL;
  } else {
    size_t adjX = coords.x - m_offsetX;
    size_t adjY = coords.y - m_offsetY;
    if ((adjX >= m_pST->m_vectSTC.size()) || (adjY >= m_pST->m_vectSTC[adjX].size())) {
      return NULL;
    } else {
      return m_pST->m_vectSTC[adjX][adjY];
    }
  }
}

TASSupertileSpace::~TASSupertileSpace()
{
  if (m_pST) {
    delete m_pST;
  }
}

void TASSupertileSpace::CalculateOffsets()
{
  size_t stWidth  = m_pST->m_vectSTC.size();
  size_t stHeight = m_pST->m_vectSTC[0].size();

  m_offsetX = (m_dims - stWidth) / 2;
  m_offsetY = (m_dims - stHeight) / 2;
}

TASSupertile::TASSupertile(TASTileType* pTT, TASTileTypeSet* pTS, size_t count)
{
  m_pTS         = pTS;
  m_count       = count;
  m_minToCheck  = 0;
  m_numTiles    = 1;
  m_epochAdded  = 0;

  TASSupertileColumn stc;
  stc.push_back(pTT);
  m_vectSTC.push_back(stc);
  CalculateDims();
  GenerateUID();
}

TASSupertile::TASSupertile(TASSupertile* pST, TASTileTypeSet* pTS, size_t count)
{
  m_pTS         = pTS;
  m_count       = count;
  m_minToCheck  = pST->m_minToCheck;
  m_numTiles    = pST->m_numTiles;
  m_epochAdded  = pST->m_epochAdded;
//MJP TODO: implement this
  //CalculateDims();
  GenerateUID();
}

TASSupertile::TASSupertile(TASSeed* pSeed, TASTileTypeSet* pTS, size_t count)
{
  m_pTS         = pTS;
  m_count       = count;
  m_minToCheck  = 0;
  m_numTiles    = 0;
  m_epochAdded  = 0;

  if (pSeed->size() > 0) {
    int xMin = INT_MAX;
    int xMax = INT_MIN;
    int yMin = INT_MAX;
    int yMax = INT_MIN;
    for (size_t i=0; i<pSeed->size(); i++) {
      //  First find greatest and least coordinates
      if (pSeed->at(i).m_coords.x < xMin) xMin = pSeed->at(i).m_coords.x;
      if (pSeed->at(i).m_coords.x > xMax) xMax = pSeed->at(i).m_coords.x;
      if (pSeed->at(i).m_coords.y < yMin) yMin = pSeed->at(i).m_coords.y;
      if (pSeed->at(i).m_coords.y > yMax) yMax = pSeed->at(i).m_coords.y;
    }
    int numCols = (xMax - xMin) + 1;
    for (int j=0; j<numCols; j++) {
      TASSupertileColumn stc((yMax-yMin)+1, NULL);
      for (size_t i=0; i<pSeed->size(); i++) {
        if (pSeed->at(i).m_coords.x == (j+xMin)) {
          stc[pSeed->at(i).m_coords.y - yMin] = m_pTS->at(pSeed->at(i).m_sName);
          m_numTiles++;
        }
      }
      m_vectSTC.push_back(stc);
    }
  }
  CalculateDims();
  GenerateUID();
}

TASSupertile::TASSupertile(TASSupertile* pST1, TASSupertile* pST2, TASCoords& offsetST2, TASTileTypeSet* pTS, size_t count)
{
  m_pTS         = pTS;
  m_count       = count;
  m_minToCheck  = 0;
  m_numTiles    = pST1->m_numTiles + pST2->m_numTiles;

  TASSupertile* pFirst = pST1;
  TASSupertile* pSecond = pST2;
  int xBias = offsetST2.x;
  int yFirst = 0;
  int ySecond = 0;
  if (offsetST2.x < 0) {
    pFirst = pST2;
    pSecond = pST1;
    xBias = -offsetST2.x;
    if (offsetST2.y < 0) {
      ySecond = -offsetST2.y;
    } else {
      yFirst = offsetST2.y;
    }
  } else {
    if (offsetST2.y < 0) {
      yFirst = -offsetST2.y;
    } else {
      ySecond = offsetST2.y;
    }
  }
  int maxHeight = pFirst->m_height + yFirst;
  if ((pSecond->m_height + ySecond) > maxHeight) {
    maxHeight = pSecond->m_height + ySecond;
  }

  for (size_t i=0; i<pFirst->m_vectSTC.size(); i++) {
    //  Copy column from first supertile
    m_vectSTC.push_back(TASSupertileColumn(yFirst, NULL));
    m_vectSTC[i].insert(m_vectSTC[i].end(), pFirst->m_vectSTC[i].begin(), pFirst->m_vectSTC[i].end());
    if ((i >= xBias) && (pSecond->m_vectSTC.size() > (i-xBias))) {
      while (m_vectSTC[i].size() <= (pSecond->m_vectSTC[i-xBias].size()+ySecond-1)) m_vectSTC[i].push_back(NULL);
      for (size_t j=0; j<pSecond->m_vectSTC[i-xBias].size(); j++) {
        if (pSecond->m_vectSTC[i-xBias][j] != NULL) {
          m_vectSTC[i][j+ySecond] = pSecond->m_vectSTC[i-xBias][j];
        }
      }
    }
    while (m_vectSTC[i].size() < maxHeight) {
      m_vectSTC[i].push_back(NULL);
    }
  }

  int colDone = pFirst->m_vectSTC.size() - xBias;
  if (pSecond->m_vectSTC.size() > colDone) {
    for (size_t i=0; i<(pSecond->m_vectSTC.size()-colDone); i++) {
      m_vectSTC.push_back(TASSupertileColumn(ySecond, NULL));
      m_vectSTC[pFirst->m_vectSTC.size()+i].insert(m_vectSTC[pFirst->m_vectSTC.size()+i].end(), pSecond->m_vectSTC[i+colDone].begin(), pSecond->m_vectSTC[i+colDone].end());
      while (m_vectSTC[pFirst->m_vectSTC.size()+i].size() < maxHeight) {
        m_vectSTC[pFirst->m_vectSTC.size()+i].push_back(NULL);
      }
    }
  }

  CalculateDims();
  GenerateUID();
}

bool TASSupertile::IsEqual(TASSupertile* pST)
{
  return (pST->m_sUID == m_sUID);
/*
  if ((m_pTS != pST->m_pTS) || (m_vectSTC.size() != pST->m_vectSTC.size())) {
    return false;
  }
  for (size_t i=0; i<m_vectSTC.size(); i++) {
    if (m_vectSTC[i].size() != pST->m_vectSTC[i].size()) {
      return false;
    }
    for (size_t j=0; j<m_vectSTC[i].size(); j++) {
      if (m_vectSTC[i][j] != pST->m_vectSTC[i][j]) {
        return false;
      }
    }
  }
  return true;
*/
}

void TASSupertile::GenerateUID()
{
  PrintToString(&m_sUID);
}
void TASSupertile::PrintToString(wxString* pStr)
{
  if (!m_sUID.IsEmpty()) {
    *pStr = m_sUID;
  } else {
    for (size_t i=0; i<m_vectSTC.size(); i++) {
      for (size_t j=0; j<m_vectSTC[i].size(); j++) {
        if (m_vectSTC[i][j] != NULL) {
	  wxString sRepr;
	  sRepr << m_vectSTC[i][j]->m_sName.c_str() << _T(" ") << i << _T(" ") << j << _T(" ") << 0 << _T("\n");
          pStr->Append(sRepr);
        }
      }
    }
  }
}
void TASSupertile::CalculateDims()
{
  m_width =  m_vectSTC.size();
  m_height = m_vectSTC[0].size();
  m_maxDim = (m_width > m_height) ? m_width : m_height;
}

bool TASSupertile::ComboAttempted(TASSupertile* WXUNUSED(pST)) {
//MJP TODO: make this thread safe!!!
  /*if ((m_comboSet.find(pST) != m_comboSet.end()) ||
      (pST->m_comboSet.find(pST) != pST->m_comboSet.end())) {
      return true;
  }*/
  return false;
}

bool TASSupertile::GetCombinations(TASSupertile* pST, std::vector<TASSupertile*>* pVectRet)
{
  pVectRet->clear();

  for (size_t i=0; i<=(m_width+pST->m_width); i++) {
    for (size_t j=0; j<=(m_height+pST->m_height); j++) {
      TASSupertile* pNewST = CanCombine(i, j, pST);
      if (pNewST != NULL) {
        pVectRet->push_back(pNewST);
      }
    }
  }

  //  Record the fact that we've attempted this combination
  if (m_count != TAS_2HAM_INFINITE_COUNT) {
    m_comboSet.insert(pST);
//  pST->m_comboSet.insert(this);
  }

  return (pVectRet->size() > 0);
}

TASSupertile* TASSupertile::CanCombine(size_t xTrans, size_t yTrans, TASSupertile* pST)
{
/*wxString s1;
wxString s2;
this->PrintToString(&s1);
pST->PrintToString(&s2);
s1.Replace("\n"," ");
s2.Replace("\n"," ");
if ((s1 == "C0-C1-9 0 1 0 C0-C1-8 1 1 0 C0-C1-6 2 0 0 C0-C1-7 2 1 0 ") && (s2 == "C0-C1-12 0 0 0 C0-C1-13 0 1 0 C0-C1-11 1 0 0 C0-C1-10 1 1 0 ")) {
  bool blah = true;
}*/
//wxString sFoo = wxString::Format(wxT(" Trans %d %d:"), xTrans, yTrans);

  //  Check if this supertile is translated to either of the four corners where
  //  it could not possibly attach
  if (((xTrans == 0) && (yTrans == 0)) || 
      ((xTrans == 0) && (yTrans == (m_height + pST->m_height))) ||
      ((xTrans == (m_width + pST->m_width)) && (yTrans == 0))   ||
      ((xTrans == (m_width + pST->m_width)) && (yTrans == (m_height + pST->m_height)))) {
//sFoo += " (corner), ";
//g_pTASApp->SetSimDebugText(sFoo, true);
    return NULL;
  }

  //  Create a representation of the space occupied by both supertiles and their 'interfaces'
  int left   = ((xTrans < m_width) ? m_width : xTrans) - 1;
  int right  = (xTrans < pST->m_width) ? (m_width + xTrans) : (m_width + pST->m_width);
  int bottom = ((yTrans < m_height) ? m_height : yTrans) - 1;
  int top    = (yTrans < pST->m_height) ? (m_height + yTrans) : (m_height + pST->m_height);
  wxRect rectInterface(wxPoint(left, bottom), wxPoint(right, top));
  std::vector<std::vector<int> > map(rectInterface.width, std::vector<int>(rectInterface.height, 0));

  //  Fill in the map with a 1 if there is a tile from this supertile or a 2 if there is a tile
  //  from the input supertile in the corresponding location.  Also, detect collisions.
  for (int x=left; x<=right; x++) {
    for (int y=bottom; y<=top; y++) {
      bool bInThisSupertile = false;
      //  If a tile of this supertile exists here, mark that in the map
      int xAdj = x - xTrans;
      int yAdj = y - yTrans;
      if ((xAdj >= 0) && (xAdj < m_vectSTC.size()) &&
          (yAdj >= 0) && (yAdj < m_vectSTC[xAdj].size()) &&
          (m_vectSTC[xAdj][yAdj] != NULL)) {
        bInThisSupertile = true;
        map[x-left][y-bottom] = 1;
      }
      //  If a tile of the input supertile exists here mark that in the map (unless
      //  both supertiles have a tile here, in which case bail out.
      xAdj = x - m_width;
      yAdj = y - m_height;
      if ((xAdj >= 0) && (xAdj < pST->m_vectSTC.size()) &&
          (yAdj >= 0) && (yAdj < pST->m_vectSTC[xAdj].size()) &&
          (pST->m_vectSTC[xAdj][yAdj] != NULL)) {
        if (bInThisSupertile) {
//sFoo += wxT(" (Collision!), ");
//g_pTASApp->SetSimDebugText(sFoo, true);
          return NULL;  //  Collision!!
        } else {
          map[x-left][y-bottom] = 2;
        }
      }
    }
  }

  //  Loop over the map and for every location occupied by a tile from this supertile,
  //  check to see if it can bind to the input supertile, keeping a running total of the
  //  binding strength.
  int bindingStrength = 0;
  for (size_t i=0; i<map.size(); i++) {
    for (size_t j=0; j<map[i].size(); j++) {
      if (map[i][j] == 1) {
        for (int d=TAS_NORTH; d<=TAS_WEST; d++) {   // MJP: Currently hardcoded for 2-D (but so is the rest of the code)
          int xDir = i + ((d == TAS_WEST) ? -1 : (d == TAS_EAST) ? 1 : 0);
          int yDir = j + ((d == TAS_SOUTH) ? -1 : (d == TAS_NORTH) ? 1 : 0);

          if ((xDir >= 0) && (yDir >= 0) && (xDir < rectInterface.width) && (yDir < rectInterface.height) && (map[xDir][yDir] == 2)) {
            int xAdj1 = i + left - xTrans;
            int yAdj1 = j + bottom - yTrans;
            int xAdj2 = xDir + left - m_width;
            int yAdj2 = yDir + bottom - m_height;
//int savedBindingStrength = bindingStrength;
            bindingStrength += m_pTS->GetGlueInteractionStrength(m_vectSTC[xAdj1][yAdj1]->m_sides[d], 
                                                                 pST->m_vectSTC[xAdj2][yAdj2]->m_sides[TASOppositeDirection[d]]);
//if (savedBindingStrength != bindingStrength) {
//sFoo += wxString::Format(wxT("(glues %s,%s), "), m_vectSTC[xAdj1][yAdj1]->m_sides[d].m_sLabel.c_str(), pST->m_vectSTC[xAdj2][yAdj2]->m_sides[TASOppositeDirection[d]].m_sLabel.c_str());
//}
            if (bindingStrength >= g_pTileSpace->GetTemp()) {
              TASCoords offset((xTrans-m_width), (yTrans-m_height), 0);
              TASSupertile* pNewST = new TASSupertile(pST, this, offset, m_pTS);
//wxString sBlah;
//pNewST->PrintToString(&sBlah);
//sFoo += wxString::Format(wxT(" BOUND to make %s!! "), sBlah.c_str());
//g_pTASApp->SetSimDebugText(sFoo, true);
              return pNewST;
            }
          }
        }
      }
    }
  }

//sFoo += " (no bind), ";
//g_pTASApp->SetSimDebugText(sFoo, true);
  return NULL;
}

void TASkTAMTileSpace::initializeStochasticSimulator()
{
	double attach = totalConcentration / exp(gmc);
	// TODO : (Michael) Fix this hack... it assumes a max bonding str of 8
	double detach = 1.0 / exp(8 * gse);

	// The min value is dropped to 98% so that it is possible to
	// select the min even if there are rounding errors and max generally fall
	// into the largest group so it doesn't have to be shifted
	double min = std::min(attach, detach);
	min = std::min(min, 1.0) * .98;

	double max = std::max(attach, detach);
	max = std::max(max, 1.0);

	sim = new CompositionAndRejection<KTamEvent>(min, max);
}

// (Michael) What happens if tiles don't have concentrations?
void TASkTAMTileSpace::initializeTileChooser()
{
	totalConcentration = 0.0;
	int max(0);
	int min(std::abs(static_cast<int>(~0)));

	// Find max concentration and min concentration for all tiles
	for(int i(0); i < m_tileset.size(); ++i) {
		TASTileType* tile = m_tileset.at(i);
		int concentration = tile->m_concentration;
		max = std::max(max, concentration);
		min = std::min(min, concentration);
	}

	tileChooser = new CompositionAndRejection<TileConcentrationEvent>(min, max);

	for(int i(0); i < m_tileset.size(); ++i) {
		TASTileType* tile = m_tileset.at(i);
		totalConcentration += static_cast<double>(tile->m_concentration) / static_cast<double>(max);
		tileChooser->AddEvent(TileConcentrationEvent(tile, tile->m_concentration));
	}
}

// TODO : (Michael) Create an Error Container

bool TASkTAMTileSpace::IncrementEpoch() {
	if(m_epoch == 0) {
		// (Michael) Must call initializeTileChooser first as it sets
		// the totalConcentration variable
		// Initialize Simulators
		initializeTileChooser();
		initializeStochasticSimulator();

		SetSeed();

		// Look around seed and add our first attachment events
		for(std::vector<TASSeedEntry>::const_iterator it = m_seed.begin(); it != m_seed.end(); ++it) {
			TASCoords me = it->m_coords;

			for(int i = 0; i < 4; ++i) {
				TASCoords position = me + m_nbrOffsets[i];
				// Be sure we are not in the seed then get ready for creating events
				if(!IsInSeed(position)) {
					sim->AddEvent(KTamEvent::CreateAttachmentEvent(position,  totalConcentration, gmc));
				}
			}
		}
	}

	++m_epoch;

	KTamEvent event = sim->SampleEvent();
	TASCoords me = event.getCoords();

	if(!me.IsValid()) {
		// TODO : (Michael) Print more information about the event that got us here.
		wxString sMsg = _T("We have invalid coordinates.\n");
		g_pTASApp->SetSimDebugText(sMsg, true, false);
	}

	// TODO : (Michael) Get an event handler from a factory then have the event handler do
	//					the correct thing.
	//	EX  : KTamEventHandler handler = KTamEventHandlerFactory::instance(event.getEvent());
	//		  handler.handle(...);
	if(event.getEvent() == KTamEvent::Attachment) {
		sim->EraseEvent(event);
		AddTile(me, tileChooser->SampleEvent().getTile());
		m_numTiles++;
        
        // TODO : (Michael) Check if tile is causing an error and increment errorCount
        if(ErrorAt(me)) {
			errors.insert(me);
            ++errorCount;
        }

		sim->AddEvent(KTamEvent::CreateDetachmentEvent(me, GetBondStrength(me), gse));
	} else if(event.getEvent() == KTamEvent::Detachment) {
		// We add attachment event at position and remove tile
		sim->EraseEvent(event);
		RemoveTile(me);
		m_numTiles--;
        
        // TODO : (Michael) If error is stored then remove error
		if(errors.find(me) != errors.end()) {
			errors.erase(me);
            --errorCount;
        }
        
		sim->AddEvent(KTamEvent::CreateAttachmentEvent(me, totalConcentration, gmc));
	} else if(event.getEvent() == KTamEvent::Error) {
		// Something went terribly wrong... we pulled out an Error event which
		// should have never been inserted
		return false;
	}


	// Check around added tile and fix events/errors
	// TODO : (Michael) Fix for 3-d
	for(int i = TAS_NORTH; i <= TAS_WEST; ++i) {
        TASCoords coord = me + m_nbrOffsets[i];
        
		// Fixes any errors around the tile
		if(errors.find(me) != errors.end()) {
			// We had an error here before
			if(!ErrorAt(me)) {
				--errorCount;
				errors.erase(me);
			}
		} else {
			// We did not have an error here before
			if(ErrorAt(me)) {
				++errorCount;
				errors.insert(me);
			}
		}
        
		UpdateTileEvents(coord);
	}

	// TODO : (Michael) Name the constant... fhatk maybe
	totalTime += sim->TimeLapsed(12000000.0);

	return true;
}

// TODO : (Michael) Find out the proper way to get errors
bool TASkTAMTileSpace::ErrorAt(const TASCoords& coords) {
	// Ensure that we have a tile
	TASTileType* me = GetTileType(coords);
	if(me == NULL) {
		return false;
	}

    return HasMismatchedGlue(coords);
}

// Check all sides of a position to see if a neighbor exists
bool TASkTAMTileSpace::HasNeighbor(const TASCoords& coords) {
	// TODO : (Michael) fix for 3-D
	for(int d = TAS_NORTH; d <= TAS_WEST; ++d) {
		// If a tile exists
		if(GetTileType(coords + m_nbrOffsets[d]) != NULL) {
			return true;
		}
	}

	return false;
}

void TASkTAMTileSpace::UpdateTileEvents(const TASCoords& coords) {
	bool eventExists = sim->ContainsEvent(KTamEvent::CreateEmptyEvent(coords));

	if(GetTileType(coords) == NULL) {
		// If there is no possibility for attachment
		if(!HasNeighbor(coords)) {
			// if we have an event remove it
			if(eventExists) {
				sim->EraseEvent(KTamEvent::CreateEmptyEvent(coords));
			}
			// if we don't have an event we are done
		} else {
			// if we have an event we are done
			// if we don't have an event add event
			if(!eventExists) {
				sim->AddEvent(KTamEvent::CreateAttachmentEvent(coords, totalConcentration, gmc));
			}
		}
	} else if(!IsInSeed(coords)) {
		// This means there is a tile there and its not the seed so there
		// should be a detachment event
		if(eventExists) {
			sim->ModifyEvent(KTamEvent::CreateDetachmentEvent(coords, GetBondStrength(coords), gse));
		} else {
			// This is an Error so correct and log
			sim->AddEvent(KTamEvent::CreateDetachmentEvent(coords, GetBondStrength(coords), gse));

			// This assumes that we take care of attachment and detachment outside of this function
			wxString sMsg = wxString::Format(_T("We lost a detachment event at (%d, %d) but it is now fixed.\n"), coords.x, coords.y);
			g_pTASApp->SetSimDebugText(sMsg, true, false);
		}
	}
}

// TODO : (Michael) Fix for different types of glues( ie diagonal etc..)
int TASkTAMTileSpace::GetBondStrength(const TASCoords& coords) {
	TASTileType* me = GetTileType(coords);
	int strength = 0;
	if(me != NULL) {
		// TODO : (Michael) fix for 3-D
		for(int i = TAS_NORTH; i <= TAS_WEST; ++i) {
			TASTileType* other = GetTileType(m_nbrOffsets[i] + coords);
			if(other != NULL) {
				TASTileSide& meSide = me->m_sides[i];
				TASTileSide& otherSide = other->m_sides[TASOppositeDirection[i]];
				if(meSide == otherSide) {
					strength += meSide.m_iStrength;
				}
			}
		}
	}
	return strength;
}

// Checks if any glue on the tile at the coords has a mismatched glue with its neighbors
bool TASkTAMTileSpace::HasMismatchedGlue(const TASCoords& coords) {
	TASTileType* me = GetTileType(coords);

	if(me != NULL) {
		// TODO : (Michael) fix for 3-D
		for(int i = TAS_NORTH; i <= TAS_WEST; ++i) {
			TASTileType* other = GetTileType(m_nbrOffsets[i] + coords);
			if(other != NULL) {
				TASTileSide& meSide = me->m_sides[i];
				TASTileSide& otherSide = other->m_sides[TASOppositeDirection[i]];
				if(meSide != otherSide) {
					return true;
				}
			}
		}
	}

	return false;
}
