/////////////////////////////////////////////////////////////////////////////
// Name:        TASXMLSettings.h
// Purpose:     Defines the class which handles app settings in xml format
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2012 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "wxprec.h"
#include <wx/xml/xml.h>
#include "TASXmlSettings.h"


TASXmlSetting::TASXmlSetting(const TASXmlSetting& in)
{
  m_sName         = in.m_sName;
  m_type          = in.m_type;
  m_sVal          = in.m_sVal;
  m_vectChildren  = in.m_vectChildren;
}

bool TASXmlSetting::GetValue(wxString* pStr)
{
  if (m_type == TAS_SETTING_TYPE_STRING) {
    *pStr = m_sVal;
    return true;
  } else {
    return false;
  }
}

bool TASXmlSetting::GetValue(long* pLong)
{
  if (m_type == TAS_SETTING_TYPE_LONG) {
    return m_sVal.ToLong(pLong);
  } else {
    return false;
  }
}

bool TASXmlSetting::GetValue(double* pDouble)
{
  if (m_type == TAS_SETTING_TYPE_DOUBLE) {
    return m_sVal.ToDouble(pDouble);
  } else {
    return false;
  }
}

bool TASXmlSetting::GetValue(bool* pBool)
{
  if (m_type == TAS_SETTING_TYPE_BOOL) {
    *pBool = (m_sVal == _T("1"));
    return true;
  } else {
    return false;
  }
}

bool TASXmlSetting::GetValues(TASXmlSettingVect* pValues)
{
  if (m_type == TAS_SETTING_TYPE_GROUP) {
    pValues = &m_vectChildren;
    return true;
  } else {
    return false;
  }
}

bool TASXmlSetting::LoadFromXml(wxXmlNode* pNode)
{
  bool bRet = false;
  m_sName = pNode->GetName();
  wxXmlNode* child = pNode->GetChildren();

  while (child) {
    wxString sName = child->GetName();
    TASXmlSetting setting(sName);

    wxXmlNode* grandchild = child->GetChildren();
    if (grandchild) {
      if (child->HasAttribute(TAS_SETTING_ATTR_TYPE)) {
        wxString sGrandVal = grandchild->GetContent();
        wxString sAttr = child->GetAttribute(TAS_SETTING_ATTR_TYPE);
        if (sAttr == TAS_SETTING_TYPE_STRING_STR) {
          setting.SetValue(sGrandVal);
          bRet = true;
        } else if (sAttr == TAS_SETTING_TYPE_LONG_STR) {
          long l = 0;
          if (sGrandVal.ToLong(&l)) {
            setting.SetValue(l);
            bRet = true;
          }
        } else if (sAttr == TAS_SETTING_TYPE_DOUBLE_STR) {
          double d = 0;
          if (sGrandVal.ToDouble(&d)) {
            setting.SetValue(d);
            bRet = true;
          }
        } else if (sAttr == TAS_SETTING_TYPE_BOOL_STR) {
          bool b = (sGrandVal == _T("1"));
          setting.SetValue(b);
          bRet = true;
        }
      } else {
        setting.m_type = TAS_SETTING_TYPE_GROUP;
        bRet = setting.LoadFromXml(child);
      }
    }
    m_vectChildren.push_back(setting);
    child = child->GetNext();
  }
  return bRet;
}

bool TASXmlSetting::DumpToXml(wxXmlNode* pNode)
{
  pNode->SetName(m_sName);
  if ((m_type == TAS_SETTING_TYPE_GROUP) || (m_type == TAS_SETTING_TYPE_UNSET)) {
    for (size_t i=0; i<m_vectChildren.size(); i++) {
      wxXmlNode* node = new wxXmlNode(wxXML_ELEMENT_NODE, m_vectChildren[i].GetName());
      m_vectChildren[i].DumpToXml(node);
      pNode->AddChild(node);
    }
  } else {
    pNode->AddAttribute(TAS_SETTING_ATTR_TYPE, TASXmlSettingTypes[m_type]);
    wxXmlNode* child = new wxXmlNode(wxXML_TEXT_NODE, m_sVal);
    child->SetContent(m_sVal);
    pNode->AddChild(child);
  }
  return true;
}

bool TASXmlSettings::ReadFromFile(wxString& sPath)
{
  m_fn = sPath;
  m_settings.Clear();

  if (m_fn.IsFileReadable()) {
    wxXmlDocument doc;
    if (doc.Load(m_fn.GetFullPath())) {
      wxString sName = doc.GetRoot()->GetName();
      m_sName = sName;
      return m_settings.LoadFromXml(doc.GetRoot());
    }
  }
  return false;
}

bool TASXmlSettings::WriteToFile(wxString& sPath)
{
  wxXmlDocument doc;
  wxXmlNode* node = new wxXmlNode(wxXML_ELEMENT_NODE, m_sName);
  doc.SetRoot(node);
  m_settings.DumpToXml(doc.GetRoot());
  bool bRet = doc.Save(sPath);

  return bRet;
}

void TASXmlSettings::AddSetting(wxString& sName, TASXmlSettingType type, wxString& sVal)
{
  TASXmlSetting sett(sName);
  if (type == TAS_SETTING_TYPE_STRING) {
    sett.SetValue(sVal);
  } else if (type == TAS_SETTING_TYPE_LONG) {
    sett.SetLongValueFromString(sVal);
  } else if (type == TAS_SETTING_TYPE_DOUBLE) {
    sett.SetDoubleValueFromString(sVal);
  } else if (type == TAS_SETTING_TYPE_BOOL) {
    sett.SetBoolValueFromString(sVal);
  }
  m_settings.AddChild(sett);
}

void TASXmlSettings::AddSetting(wxString& sName, long lVal)
{
  TASXmlSetting sett(sName, lVal);
  m_settings.AddChild(sett);
}

void TASXmlSettings::AddSetting(wxString& sName, double dVal)
{
  TASXmlSetting sett(sName, dVal);
  m_settings.AddChild(sett);
}

void TASXmlSettings::AddSetting(wxString& sName, bool bVal)
{
  TASXmlSetting sett(sName, bVal);
  m_settings.AddChild(sett);
}

void TASXmlSetting::SetValue(TASXmlSettingType type, wxString& str)
{
  if (type == TAS_SETTING_TYPE_STRING) {
    SetValue(str);
  } else if (type == TAS_SETTING_TYPE_LONG) {
    SetLongValueFromString(str);
  } else if (type == TAS_SETTING_TYPE_DOUBLE) {
    SetDoubleValueFromString(str);
  } else if (type == TAS_SETTING_TYPE_BOOL) {
    SetBoolValueFromString(str);
  }
}

void TASXmlSettings::AddChild(std::vector<wxString>* pNameList, TASXmlSetting& sett)
{
  TASXmlSetting* pSett = &m_settings;
  for (size_t i=0; i<pNameList->size(); i++) {
    bool bFound = false;
    for (size_t j=0; j<pSett->m_vectChildren.size(); j++) {
      if (pSett->m_vectChildren[j].GetName() == pNameList->at(i)) {
        pSett = &pSett->m_vectChildren[j];
        bFound = true;
      }
    }
    if (!bFound) {
      TASXmlSetting newSett(pNameList->at(i));
      if (i == (pNameList->size()-1)) {
        newSett.AddChild(sett);
        pSett->m_vectChildren.push_back(newSett);
      } else {
        pSett->m_vectChildren.push_back(newSett);
        pSett = &pSett->m_vectChildren[pSett->m_vectChildren.size()-1];
      }
    } else if (i == (pNameList->size()-1)) {
      pSett->AddChild(sett);
    }
  }
}

void TASXmlSettings::AddSetting(std::vector<wxString>* pNameList, TASXmlSettingType sType, wxString& sVal)
{
  TASXmlSetting* pSett = &m_settings;
  for (size_t i=0; i<pNameList->size(); i++) {
    bool bFound = false;
    for (size_t j=0; j<pSett->m_vectChildren.size(); j++) {
      if (pSett->m_vectChildren[j].GetName() == pNameList->at(i)) {
        pSett = &pSett->m_vectChildren[j];
        bFound = true;
      }
    }
    if (!bFound) {
      TASXmlSetting newSett(pNameList->at(i));
      if (i == (pNameList->size()-1)) {
        newSett.SetValue(sType, sVal);
        pSett->m_vectChildren.push_back(newSett);
      } else {
        pSett->m_vectChildren.push_back(newSett);
        pSett = &pSett->m_vectChildren[pSett->m_vectChildren.size()-1];
      }
    }
  }
}

void TASXmlSettings::AddSetting(std::vector<wxString>* pNameList, long lVal)
{
  TASXmlSetting sett;
  sett.SetValue(lVal);
  AddSetting(pNameList, TAS_SETTING_TYPE_LONG, sett.m_sVal);
}

void TASXmlSettings::AddSetting(std::vector<wxString>* pNameList, double dVal)
{
  TASXmlSetting sett;
  sett.SetValue(dVal);
  AddSetting(pNameList, TAS_SETTING_TYPE_DOUBLE, sett.m_sVal);
}

void TASXmlSettings::AddSetting(std::vector<wxString>* pNameList, bool bVal)
{
  TASXmlSetting sett;
  sett.SetValue(bVal);
  AddSetting(pNameList, TAS_SETTING_TYPE_BOOL, sett.m_sVal);
}

bool TASXmlSettings::GetSetting(wxString sName, wxString* psVal)
{
  for (size_t i=0; i<m_settings.m_vectChildren.size(); i++) {
    if (m_settings.m_vectChildren[i].GetName() == sName) {
      return m_settings.m_vectChildren[i].GetValue(psVal);
    }
  }
  return false;
}

bool TASXmlSettings::GetSetting(wxString sName, long* plVal)
{
  for (size_t i=0; i<m_settings.m_vectChildren.size(); i++) {
    if (m_settings.m_vectChildren[i].GetName() == sName) {
      return m_settings.m_vectChildren[i].GetValue(plVal);
    }
  }
  return false;
}

bool TASXmlSettings::GetSetting(wxString sName, double* pdVal)
{
  for (size_t i=0; i<m_settings.m_vectChildren.size(); i++) {
    if (m_settings.m_vectChildren[i].GetName() == sName) {
      return m_settings.m_vectChildren[i].GetValue(pdVal);
    }
  }
  return false;
}

bool TASXmlSettings::GetSetting(wxString sName, bool* pbVal)
{
  for (size_t i=0; i<m_settings.m_vectChildren.size(); i++) {
    if (m_settings.m_vectChildren[i].GetName() == sName) {
      return m_settings.m_vectChildren[i].GetValue(pbVal);
    }
  }
  return false;
}

bool TASXmlSettings::GetSetting(wxString sName, TASXmlSetting** pSetting)
{
  for (size_t i=0; i<m_settings.m_vectChildren.size(); i++) {
    if (m_settings.m_vectChildren[i].GetName() == sName) {
      *pSetting = &m_settings.m_vectChildren[i];
      return true;
    }
  }
  return false;
}

bool TASXmlSettings::GetSetting(std::vector<wxString>* pNameList, TASXmlSetting** pSetting)
{
  TASXmlSetting* pSett = &m_settings;
  for (size_t i=0; i<pNameList->size(); i++) {
    bool bFound = false;
    for (size_t j=0; j<pSett->m_vectChildren.size(); j++) {
      if (pSett->m_vectChildren[j].GetName() == pNameList->at(i)) {
        if (i == (pNameList->size()-1)) {
          *pSetting = &pSett->m_vectChildren[j];
          return true;
        } else {
          pSett = &pSett->m_vectChildren[j];
          bFound = true;
          break;
        }
      }
    }
    if (!bFound) return false;
  }
  return false;
}

bool TASXmlSettings::GetSetting(std::vector<wxString>* pNameList, wxString* psVal)
{
  TASXmlSetting* pSett = NULL;
  if (GetSetting(pNameList, &pSett)) {
    return pSett->GetValue(psVal);
  }
  return false;
}

bool TASXmlSettings::GetSetting(std::vector<wxString>* pNameList, long* plVal)
{
  TASXmlSetting* pSett = NULL;
  if (GetSetting(pNameList, &pSett)) {
    return pSett->GetValue(plVal);
  }
  return false;
}

bool TASXmlSettings::GetSetting(std::vector<wxString>* pNameList, double* pdVal)
{
  TASXmlSetting* pSett = NULL;
  if (GetSetting(pNameList, &pSett)) {
    return pSett->GetValue(pdVal);
  }
  return false;
}

bool TASXmlSettings::GetSetting(std::vector<wxString>* pNameList, bool* pbVal)
{
  TASXmlSetting* pSett = NULL;
  if (GetSetting(pNameList, &pSett)) {
    return pSett->GetValue(pbVal);
  }
  return false;
}
