/////////////////////////////////////////////////////////////////////////////
// Name:        TASXMLSettings.h
// Purpose:     Declares the class which handles app settings in xml format
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2012 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef TASXMLSettings_H
#define TASXMLSettings_H

#include <vector>
#include <wx/xml/xml.h>
#include <wx/filename.h>

#define TAS_SETTING_TYPE_UNSET  0
#define TAS_SETTING_TYPE_STRING 1
#define TAS_SETTING_TYPE_LONG   2
#define TAS_SETTING_TYPE_DOUBLE 3
#define TAS_SETTING_TYPE_BOOL   4
#define TAS_SETTING_TYPE_GROUP  5
typedef int TASXmlSettingType;

#define TAS_SETTING_TYPE_UNSET_STR  _T("")
#define TAS_SETTING_TYPE_STRING_STR _T("string")
#define TAS_SETTING_TYPE_LONG_STR   _T("long")
#define TAS_SETTING_TYPE_DOUBLE_STR _T("double")
#define TAS_SETTING_TYPE_BOOL_STR   _T("bool")
#define TAS_SETTING_TYPE_GROUP_STR  _T("group")
static wxString TASXmlSettingTypes[] = {TAS_SETTING_TYPE_UNSET_STR, TAS_SETTING_TYPE_STRING_STR, TAS_SETTING_TYPE_LONG_STR,
                                        TAS_SETTING_TYPE_DOUBLE_STR, TAS_SETTING_TYPE_BOOL_STR, TAS_SETTING_TYPE_GROUP_STR};

#define TAS_SETTING_ATTR_TYPE   _T("type")

class TASXmlSetting;
typedef std::vector<TASXmlSetting> TASXmlSettingVect;

class TASXmlSetting {
public:

  TASXmlSetting() { m_type = TAS_SETTING_TYPE_UNSET; };
  TASXmlSetting(wxString& sName) { m_type = TAS_SETTING_TYPE_UNSET; m_sName = sName; };
  TASXmlSetting(const TASXmlSetting& in);
  TASXmlSetting(wxString& sName, wxString sVal) { SetName(sName); SetValue(sVal); };
  TASXmlSetting(wxString& sName, long lVal) { SetName(sName); SetValue(lVal); };
  TASXmlSetting(wxString& sName, double dVal) { SetName(sName); SetValue(dVal); };
  TASXmlSetting(wxString& sName, bool bVal) { SetName(sName); SetValue(bVal); };

  void Clear() { m_type = TAS_SETTING_TYPE_UNSET; m_sVal = _T(""); m_vectChildren.clear(); };

  wxString& GetName() { return m_sName; };
  void SetName(wxString& sName) { m_sName = sName; };

  //TODO: these should return a bool based on success
  void SetValue(wxString& str) { Clear(); m_type = TAS_SETTING_TYPE_STRING; m_sVal = str; };
  void SetValue(long l) { Clear(); m_type = TAS_SETTING_TYPE_LONG; m_sVal = wxString::Format(_T("%ld"),l); };
  void SetLongValueFromString(wxString& str) { Clear(); m_type = TAS_SETTING_TYPE_LONG; m_sVal = str; };
  void SetValue(double d) { Clear(); m_type = TAS_SETTING_TYPE_DOUBLE; m_sVal = m_sVal.FromDouble(d); };
  void SetDoubleValueFromString(wxString& str) { Clear(); m_type = TAS_SETTING_TYPE_DOUBLE; m_sVal = str; };
  void SetValue(bool b) { Clear(); m_type = TAS_SETTING_TYPE_BOOL; m_sVal = (b ? _T("1") : _T("0")); };
  void SetBoolValueFromString(wxString& str) { Clear(); m_type = TAS_SETTING_TYPE_BOOL; m_sVal = str; };
  void SetValue(TASXmlSettingVect* pChildren) { Clear(); m_type = TAS_SETTING_TYPE_GROUP; m_vectChildren = *pChildren; };
  void SetValue(TASXmlSettingType type, wxString& str);

  bool GetValue(wxString* pStr);
  bool GetValue(long* pLong);
  bool GetValue(double* pDouble);
  bool GetValue(bool* pBool);
  bool GetValues(TASXmlSettingVect* pValues);

  void AddChild(TASXmlSetting& sett) { m_vectChildren.push_back(sett); };

  TASXmlSettingType GetType() { return m_type; };

  bool LoadFromXml(wxXmlNode* pNode);
  bool DumpToXml(wxXmlNode* pNode);
  
  wxString m_sName;
  TASXmlSettingType m_type;
  wxString m_sVal;
  TASXmlSettingVect m_vectChildren;
};

class TASXmlSettings {
public:
  TASXmlSettings() { };

  void SetName(wxString& sName) { m_sName = sName; m_settings.SetName(sName); };

  void AddSetting(wxString& sName, TASXmlSettingType sType, wxString& sVal);
  void AddSetting(wxString& sName, long lVal);
  void AddSetting(wxString& sName, double dVal);
  void AddSetting(wxString& sName, bool bVal);

  void AddSetting(std::vector<wxString>* pNameList, TASXmlSettingType sType, wxString& sVal);
  void AddSetting(std::vector<wxString>* pNameList, long lVal);
  void AddSetting(std::vector<wxString>* pNameList, double dVal);
  void AddSetting(std::vector<wxString>* pNameList, bool bVal);

  //  MJP: expand this to allow for paths (so node isn't added to root)
  void AddChild(TASXmlSetting& sett) { m_settings.AddChild(sett); };
  void AddChild(std::vector<wxString>* pNameList, TASXmlSetting& sett);

  bool GetSetting(wxString sName, wxString* psVal);
  bool GetSetting(wxString sName, long* plVal);
  bool GetSetting(wxString sName, double* pdVal);
  bool GetSetting(wxString sName, bool* pbVal);
  bool GetSetting(wxString sName, TASXmlSetting** pSetting);
  bool GetSetting(std::vector<wxString>* pNameList, wxString* psVal);
  bool GetSetting(std::vector<wxString>* pNameList, long* plVal);
  bool GetSetting(std::vector<wxString>* pNameList, double* pdVal);
  bool GetSetting(std::vector<wxString>* pNameList, bool* pbVal);
  bool GetSetting(std::vector<wxString>* pNameList, TASXmlSetting** pSetting);

  //bool HasSetting(wxString sName);
  //bool HasSetting(std::vector<wxString>* pNameList);

  bool ReadFromFile(wxString& sPath);
  bool WriteToFile(wxString& sPath);
  
  wxString      m_sName;
  wxFileName    m_fn;
  TASXmlSetting m_settings;
};

#endif // TASXMLSettings_H